﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Adds a color point attribute to all frames of an animation using values of a specified attribute and a palette module.
    /// </summary>
    [ModuleInfo("Oldřich Petřík", "Adds a color point attribute to all frames of an animation using values of a specified attribute and a palette module.", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("InAnimation", "Input dynamic mesh")]
    [PortInfo("OutAnimation", "Colored animation")]
    [PortInfo("InColor", "Input for color value - should be connected to a color palette module which transforms the value at the OutValue port into a color")]
    [PortInfo("OutValue", "Output value - should be connected to a palette module which transforms the value into a color")]
    public class AnimationColorMapping : Module
    {
        private UniformDataArray _res = null;

        private ColorRGBA[] _palette;
        private bool[] _done;

        private string _inAttrName = "";
        private string _outAttrName = "ColorRGBA";
        private int _sampling = 1000;

        /// <summary>Name of the input attribute.</summary>
        public string InputAttribute
        {
            get { return _inAttrName; }
            set { _inAttrName = value; }
        }

        /// <summary>Name of the output attribute.</summary>
        public string OutputAttribute
        {
            get { return _outAttrName; }
            set { _outAttrName = value; }
        }

        /// <summary>Sampling density of the color values.</summary>
        public int SamplingDensity
        {
            get { return _sampling; }
            set { _sampling = value; }
        }

        /// <summary>
        /// AnimationColorMapping constructor.
        /// </summary>
        public AnimationColorMapping()
        {
            AddInPort("InColor", typeof(ColorRGBA), true, false);
            AddOutPort("OutValue", typeof(Scalar), false);

            AddInPort("InAnimation", typeof(UniformDataArray));
            AddOutPort("OutAnimation", typeof(UniformDataArray));
        }

        private void ResetDone()
        {
            if (_done == null || _done.Length != _sampling)
            {
                _done = new bool[_sampling];
                _palette = new ColorRGBA[_sampling];
            }
            for (int i = 0; i < _sampling; i++) _done[i] = false;
        }

        private ColorRGBA GetColor(int index)
        {
            if (!_done[index])
            {
                Scalar val;
                val.Value = (double)index / (_sampling - 1);
                SetOutput("OutValue", val, false);
                UpdateInput("InColor");
                _palette[index] = (ColorRGBA)GetInput("InColor");
                _done[index] = true;
            }
            return _palette[index];
        }

        /// <inheritdoc />
        public override void Execute()
        {
            if (_res != null && IsDataSame("InAnimation") && IsAllUserSettingsSame())
            {
                SetOutput("OutAnimation", _res, true);
            }
            else
            {
                ResetDone();

                UniformDataArray anim = (UniformDataArray)GetInput("InAnimation");
                UnstrGrid[] frames = (UnstrGrid[])anim.Array;
                UnstrGrid frame;
                UniformDataArray attr;
                Scalar[] aAttr;
                ColorRGBA[] cAttr;

                int N = frames.Length;
                double frac = 0.5 / N;
                double min = double.MaxValue;
                double max = double.MinValue;
                double mul, off;
                int idx;

                _res = new UniformDataArray(typeof(UnstrGrid), N);

                for (int f = 0, i = 1; f < N; f++, i++)
                {
                    frame = frames[f];

                    if (_inAttrName.Length > 0) attr = frame.GetPointAttrs(_inAttrName);
                    else attr = frame.GetPointAttrs(typeof(Scalar))[0];

                    aAttr = (Scalar[])attr.Array;
                    for (int j = 0; j < aAttr.Length; j++)
                    {
                        min = Math.Min(min, aAttr[j].Value);
                        max = Math.Max(max, aAttr[j].Value);
                    }

                    ProgressInfo(i * frac);
                }

                mul = (_sampling - 1) / (max - min);
                off = (_sampling - 1) * min;

                for (int f = 0, i = N + 1; f < N; f++, i++)
                {
                    frame = frames[f];

                    if (_inAttrName.Length > 0) attr = frame.GetPointAttrs(_inAttrName);
                    else attr = frame.GetPointAttrs(typeof(Scalar))[0];

                    aAttr = (Scalar[])attr.Array;
                    attr = new UniformDataArray(typeof(ColorRGBA), aAttr.Length);
                    cAttr = (ColorRGBA[])attr.Array;

                    for (int j = 0; j < aAttr.Length; j++)
                    {
                        idx = (int)Math.Round(aAttr[j].Value * mul - off);
                        cAttr[j] = GetColor(idx);
                    }

                    ProgressInfo(i * frac);

                    frame = frame.ShallowCopy();
                    frame.AddPointAttrs(_outAttrName, attr);
                    _res[f] = frame;
                }

                SetOutput("OutAnimation", _res, false);
            }
        }
    }
}
