using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// KG error measuring module, uses per-frame avregaes and Frobenius norm.
    /// </summary>
    [ModuleInfo("Libor Va", "KG error measuring module, per-frame avregaes, Frobenius norm", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("AnimationA", "First input animation")]
    [PortInfo("AnimationB", "Second input animation")]
    [PortInfo("KGError", "KG Error value")]
    class AnimationComparerKG:Module
    {

        public AnimationComparerKG()
        {
            AddInPort("AnimationA", typeof(UniformDataArray));
            AddInPort("AnimationB", typeof(UniformDataArray));
            AddOutPort("KGError", typeof(Scalar));            
        }

        Scalar resultMSE;                

        protected bool paramsChanged = true;

        double[,] means = null;

        private double[] GetPartialSums(object index)
        {
            int i = (int)index;
            double[] s = new double[2];
            UnstrGrid frame1 = (UnstrGrid)input1[i];
            UnstrGrid frame2 = (UnstrGrid)input2[i];

            UniformDataArray points1 = frame1.Points;
            UniformDataArray points2 = frame2.Points;
            for (int j = 0; j < pointCount; j++)
            {
                Point3D p1 = (Point3D)points1[j];
                Point3D p2 = (Point3D)points2[j];
                if (double.IsInfinity(p2.X) || double.IsInfinity(p2.Y) || double.IsInfinity(p2.Z))
                {
                    WriteDebug("Found infinity");
                }

                s[0] += (p1.X - p2.X) * (p1.X - p2.X);
                s[0] += (p1.Y - p2.Y) * (p1.Y - p2.Y);
                s[0] += (p1.Z - p2.Z) * (p1.Z - p2.Z);
                s[1] += (p1.X - means[i, 0]) * (p1.X - means[i, 0]);
                s[1] += (p1.Y - means[i, 1]) * (p1.Y - means[i, 1]);
                s[1] += (p1.Z - means[i, 2]) * (p1.Z - means[i, 2]);                
            }
            return (s);
        }

        UniformDataArray input1;
        UniformDataArray input2;
        int pointCount;

        public override void Execute()
        {            
            if (!IsAllDataSame()||(paramsChanged))
            {
                input1 = (UniformDataArray)GetInput("AnimationA");
                input2 = (UniformDataArray)GetInput("AnimationB");

                pointCount = ((UnstrGrid)input1[0]).Points.Count;
                int frameCount = input1.Count;
                
                //double result = 0;
                if (!IsDataSame("AnimationA") || means == null)
                {
                    means = new double[frameCount, 3];
                    for (int i = 0; i < pointCount; i++)
                    {
                        for (int j = 0; j < frameCount; j++)
                        {
                            UnstrGrid mesh = (UnstrGrid)input1[j];
                            Point3D point = (Point3D)mesh.Points[i];
                            means[j, 0] += point.X;
                            means[j, 1] += point.Y;
                            means[j, 2] += point.Z;
                        }
                    }
                    for (int i = 0; i < frameCount; i++)
                    {
                        means[i, 0] /= pointCount;
                        means[i, 1] /= pointCount;
                        means[i, 2] /= pointCount;
                    }
                }

                double s1 = 0;
                double s2 = 0;

                object[] indices = new object[frameCount];
                for (int i = 0; i < frameCount; i++)
                {
                    indices[i] = (object)i;
                }
                ThreadedExecution<object, double[]> threadExecution = new ThreadedExecution<object, double[]>(indices, GetPartialSums, this.ProgressInfo);
                double[][] sums = threadExecution.Execute();
                for (int i = 0; i < frameCount; i++)
                {
                    s1 += sums[i][0];
                    s2 += sums[i][1];
                }

                if (s2 == 0) s2 = 1;
                resultMSE = new Scalar((Math.Sqrt(s1)/Math.Sqrt(s2))*100);

                SetOutput("KGError", resultMSE, false);                

                WriteLog("KG Error", resultMSE.Value);

                paramsChanged = false;
            }
            else
            {
                SetOutput("KGError", resultMSE, true);                
            }
        }
    }
}
