using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Module comparing two dynamic meshes using the STED algorithm.
    /// </summary>
    [ModuleInfo("Libor Va", "Module comparing two dynamic meshes using the STED algorithm.", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("AnimationA", "First input mesh")]
    [PortInfo("AnimationB", "Second input mesh")]
    [PortInfo("SpatialNormalizeConstant", "Normalization constant. Should be set to the distance of two most distant points in the first frame of the animation. The Distance module from Visualization can be used to evaluate it.")]
    [PortInfo("Error", "Resulting STED value")]
    class AnimationComparerSTED : Module
    {
        public AnimationComparerSTED()
        {
            AddInPort("AnimationA", typeof(UniformDataArray));
            AddInPort("AnimationB", typeof(UniformDataArray));
            AddInPort("SpatialNormalizeConstant", typeof(Scalar), false);
            AddOutPort("Error", typeof(Scalar));
        }        

        //double frameDistance = 0.1;
        double frameDistance = 0.0003;

        public double FrameDistance
        {
            get { return frameDistance; }
            set { frameDistance = value; }
        }        

        //bool paramsChanged = true;

        bool abs = true;

        public bool Abs
        {
            get { return abs; }
            set
            {
                abs = value;
                //paramsChanged = true;
            }
        }

        private int width = 1;

        public int Width
        {
            get { return width; }
            set { width = value; }
        }

        int wsize = 5;

        public int SpeedWindowSize
        {
            get { return wsize; }
            set { wsize = value; }
        }

        double weightingConstant = 0.00009144;

        public double WeightingConstant
        {
            get { return weightingConstant; }
            set { weightingConstant = value; }
        }

        bool recomputeIncidence = false;

        public bool RecomputeIncidence
        {
            get { return recomputeIncidence; }
            set { recomputeIncidence = value; }
        }


        Line[] edgeIncidence;
        List<Line> edges;        
        List<int>[] vertexEdgeIncidence;
        List<int>[] vertexVertexIncidence;
        List<int>[] nbhoods;

        public override void Execute()
        {
            if (IsAllDataSame() && IsAllUserSettingsSame())
            {
                SetOutput("Error", new Scalar(lastResult));
            }
            else
            {

                UniformDataArray original = (UniformDataArray)GetInput("AnimationA");
                UniformDataArray distorted = (UniformDataArray)GetInput("AnimationB");

                int frameCount = original.Count;
                UnstrGrid first = (UnstrGrid)original[0];
                int pointCount = first.Points.Count;

                //double maxDiag = first.BoundingBox.MaxDiag;            

                double normConstant = 0;
                if (GetInPort("SpatialNormalizeConstant").Link != null)
                {
                    normConstant = (double)((Scalar)GetInput("SpatialNormalizeConstant"));
                }
                else
                {
                    normConstant = first.BoundingBox.MaxDiag;
                }

                double invMaxDiag = 1 / normConstant;

                UniformDataArray triangles = first.GetCells(typeof(Triangle))[0];
                int count = 0;

                if ((edgeIncidence == null) || recomputeIncidence)
                {
                    edgeIncidence = new Line[triangles.Count * 3];
                    edges = new List<Line>();
                    vertexEdgeIncidence = new List<int>[pointCount];
                    vertexVertexIncidence = new List<int>[pointCount];
                    nbhoods = new List<int>[pointCount];
                    for (int i = 0; i < vertexEdgeIncidence.Length; i++)
                    {
                        vertexEdgeIncidence[i] = new List<int>();
                        vertexVertexIncidence[i] = new List<int>();
                        nbhoods[i] = new List<int>();
                    }

                    foreach (Triangle t in triangles)
                    {
                        int i1, i2, i3;
                        Line l = new Line(t.V1, t.V2, false);
                        if (!edges.Contains(l))
                        {
                            i1 = edges.Count;
                            edges.Add(l);
                        }
                        else
                        {
                            i1 = edges.IndexOf(l);
                        }
                        l = new Line(t.V2, t.V3, false);
                        if (!edges.Contains(l))
                        {
                            i2 = edges.Count;
                            edges.Add(l);
                        }
                        else
                        {
                            i2 = edges.IndexOf(l);
                        }
                        l = new Line(t.V3, t.V1, false);
                        if (!edges.Contains(l))
                        {
                            i3 = edges.Count;
                            edges.Add(l);
                        }
                        else
                        {
                            i3 = edges.IndexOf(l);
                        }
                        /*edgeIncidence[count * 3] = new Line(i1, i2);
                        edgeIncidence[count * 3+1] = new Line(i1, i3);
                        edgeIncidence[count * 3 + 2] = new Line(i2, i3);*/
                        if (!vertexEdgeIncidence[t.V1].Contains(i1))
                            vertexEdgeIncidence[t.V1].Add(i1);
                        if (!vertexEdgeIncidence[t.V1].Contains(i3))
                            vertexEdgeIncidence[t.V1].Add(i3);
                        if (!vertexEdgeIncidence[t.V2].Contains(i1))
                            vertexEdgeIncidence[t.V2].Add(i1);
                        if (!vertexEdgeIncidence[t.V2].Contains(i2))
                            vertexEdgeIncidence[t.V2].Add(i2);
                        if (!vertexEdgeIncidence[t.V3].Contains(i2))
                            vertexEdgeIncidence[t.V3].Add(i2);
                        if (!vertexEdgeIncidence[t.V3].Contains(i3))
                            vertexEdgeIncidence[t.V3].Add(i3);

                        if (!vertexVertexIncidence[t.V1].Contains(t.V2))
                        {
                            vertexVertexIncidence[t.V1].Add(t.V2);
                        }
                        if (!vertexVertexIncidence[t.V2].Contains(t.V1))
                        {
                            vertexVertexIncidence[t.V2].Add(t.V1);
                        }
                        if (!vertexVertexIncidence[t.V1].Contains(t.V3))
                        {
                            vertexVertexIncidence[t.V1].Add(t.V3);
                        }
                        if (!vertexVertexIncidence[t.V3].Contains(t.V1))
                        {
                            vertexVertexIncidence[t.V3].Add(t.V1);
                        }
                        if (!vertexVertexIncidence[t.V2].Contains(t.V3))
                        {
                            vertexVertexIncidence[t.V2].Add(t.V3);
                        }
                        if (!vertexVertexIncidence[t.V3].Contains(t.V2))
                        {
                            vertexVertexIncidence[t.V3].Add(t.V2);
                        }

                        count++;
                        ProgressInfo((double)count / triangles.Count, 0, 3);
                    }

                    for (int i = 0; i < nbhoods.Length; i++)
                    {
                        List<int> NV = new List<int>();
                        NV.Add(i);

                        for (int j = 0; j < width; j++)
                        {
                            List<int> toAdd = new List<int>();
                            foreach (int v in NV)
                            {
                                foreach (int n in vertexVertexIncidence[v])
                                    toAdd.Add(n);
                            }
                            foreach (int n in toAdd)
                            {
                                if (!NV.Contains(n))
                                    NV.Add(n);
                            }
                        }

                        foreach (int n in NV)
                        {
                            foreach (int l in vertexEdgeIncidence[n])
                            {
                                if (!nbhoods[i].Contains(l))
                                    nbhoods[i].Add(l);
                            }
                        }
                    }
                }

                /*for (int i = 0; i < nbhoods.Length; i++)
                {
                    //nbhoods[i] = new List<int>();

                    foreach (int l in vertexEdgeIncidence[i])
                        nbhoods[i].Add(l);
                    foreach (int n in vertexVertexIncidence[i])
                    {
                        foreach (int l in vertexEdgeIncidence[n])
                        {
                            if (!nbhoods[i].Contains(l))
                                nbhoods[i].Add(l);
                        }
                    }
                }*/

                double dt = frameDistance;

                double spatSum = 0;

                double lengthSum = 0;

                double zeroLengthPenalty = 0;

                for (int f = 0; f < frameCount; f++)
                {
                    double[] eds;
                    double[] els = new double[edges.Count];
                    eds = edgeDifferences(original, distorted, edges, ref lengthSum, zeroLengthPenalty, f, els);                    

                    for (int i = 0; i < pointCount; i++)
                    {
                        double sum = 0;
                        double suml = 0;
                        int cnt = 0;
                        foreach (int e in nbhoods[i])
                        {
                            //sum += eds[e];
                            sum += eds[e];//*els[e];
                            suml += els[e];
                            cnt++;
                        }
                        //double avg = sum / cnt;
                        double avg = sum / suml;
                        if (!double.IsNaN(avg))
                        {
                            sum = 0;
                            //foreach (int e in vertexEdgeIncidence[i])
                            foreach (int e in nbhoods[i])
                            {
                                //sum += (eds[e] - avg) * (eds[e] - avg);
                                if (els[e] > double.Epsilon)
                                {
                                    sum += (eds[e] / els[e] - avg) * (eds[e] / els[e] - avg) * els[e];
                                }
                                else
                                {
                                    //WriteDebug("Hergot");
                                }
                                //sum += Math.Abs(eds[e] - avg);
                            }
                            double stdev = Math.Sqrt(sum / suml);
                            //double stdev = Math.Sqrt(sum / cnt);
                            //double stdev = sum / cnt;
                            double avglength = suml / cnt;
                            spatSum += stdev * avglength;
                            //spatSum += stdev;
                            //spatSum += sum / cnt;
                            if (double.IsNaN(spatSum))
                            {
                                WriteDebug("Hergot");
                            }
                        }
                    }

                    ProgressInfo((double)f / frameCount, 1);
                }

                double tempSum = 0;
                double l1sum = 0;
                double l2sum = 0;
                double difsum = 0;
                double dividersum = 0;

                double[,] speeds = new double[frameCount, pointCount];
                int[,] counts = new int[frameCount, pointCount];
                for (int f = 1; f < frameCount; f++)
                {
                    UnstrGrid frameA = (UnstrGrid)original[f];
                    UnstrGrid frameB = (UnstrGrid)distorted[f];
                    UnstrGrid pframeA = (UnstrGrid)original[f - 1];
                    UnstrGrid pframeB = (UnstrGrid)distorted[f - 1];
                    for (int p = 0; p < pointCount; p++)
                    {
                        Point3D p1A = (Point3D)pframeA.Points[p];
                        Point3D p1B = (Point3D)pframeB.Points[p];
                        Point3D p2A = (Point3D)frameA.Points[p];
                        Point3D p2B = (Point3D)frameB.Points[p];
                        double dx1 = (p1A.X - p2A.X) * invMaxDiag;
                        double dy1 = (p1A.Y - p2A.Y) * invMaxDiag;
                        double dz1 = (p1A.Z - p2A.Z) * invMaxDiag;
                        double l1 = Math.Sqrt(dx1 * dx1 + dy1 * dy1 + dz1 * dz1 + dt * dt);
                        int wstart = Math.Max(0, f - wsize);
                        int wend = Math.Min(frameCount, f + wsize);
                        for (int lf = wstart; lf < wend; lf++)
                        {
                            speeds[lf, p] += l1;
                            counts[lf, p]++;
                        }
                    }
                }

                double[,] difs = new double[frameCount, pointCount];

                for (int f = 1; f < frameCount; f++)
                {
                    UnstrGrid frameA = (UnstrGrid)original[f];
                    UnstrGrid frameB = (UnstrGrid)distorted[f];
                    UnstrGrid pframeA = (UnstrGrid)original[f - 1];
                    UnstrGrid pframeB = (UnstrGrid)distorted[f - 1];
                    for (int p = 0; p < pointCount; p++)
                    {
                        Point3D p1A = (Point3D)pframeA.Points[p];
                        Point3D p1B = (Point3D)pframeB.Points[p];
                        Point3D p2A = (Point3D)frameA.Points[p];
                        Point3D p2B = (Point3D)frameB.Points[p];
                        double dx1 = (p1A.X - p2A.X) * invMaxDiag;
                        double dy1 = (p1A.Y - p2A.Y) * invMaxDiag;
                        double dz1 = (p1A.Z - p2A.Z) * invMaxDiag;
                        double l1 = Math.Sqrt(dx1 * dx1 + dy1 * dy1 + dz1 * dz1 + dt * dt);
                        l1sum += l1;
                        double dx2 = (p1B.X - p2B.X) * invMaxDiag;
                        double dy2 = (p1B.Y - p2B.Y) * invMaxDiag;
                        double dz2 = (p1B.Z - p2B.Z) * invMaxDiag;
                        double l2 = Math.Sqrt(dx2 * dx2 + dy2 * dy2 + dz2 * dz2 + dt * dt);
                        l2sum += l2;

                        double dif = l2 - l1;

                        difsum += Math.Abs(dif);
                        
                        double divider = speeds[f, p] / counts[f, p];                        
                        dividersum += divider;
                        dif = dif / divider;
                    

                        if (abs)
                        {
                            tempSum += Math.Abs(dif);
                        }
                        {
                            tempSum += dif * dif;
                        }
                        //tempSum = Math.Abs(dif);
                    }
                    ProgressInfo((double)f / frameCount, 2);
                }



                spatSum = spatSum / (pointCount * frameCount);
                //spatSum = spatSum / (lengthSum);
                tempSum = tempSum / (pointCount * frameCount);

                double overall = Math.Sqrt(spatSum * spatSum + weightingConstant * weightingConstant * tempSum * tempSum);

                WriteLog("Spatial error", spatSum);
                WriteLog("Temporal error", tempSum);
                WriteLog("Overal error", overall);
                //WriteLog("l1 avg", l1sum / (pointCount * frameCount));
                //WriteLog("l2 avg", l2sum / (pointCount * frameCount));
                //WriteLog("dif avg", difsum / (pointCount * frameCount));
                //WriteLog("div avg", dividersum / (pointCount * frameCount));
                SetOutput("Error", new Scalar(overall));
                lastResult = overall;
            }
        }

        double lastResult = 0;

        private double[] edgeDifferences(UniformDataArray original, UniformDataArray distorted, List<Line> edges, ref double lengthSum, double zeroLengthPenalty, int f, double[] els)
        {
            // edge differences
            double[] eds;
            eds = new double[edges.Count];
            UnstrGrid frameA = (UnstrGrid)original[f];
            UnstrGrid frameB = (UnstrGrid)distorted[f];
            double invFDiagonal = frameA.BoundingBox.MaxDiag;
            int pos = 0;
            foreach (Line l in edges)
            {
                Point3D v1 = (Point3D)frameA.Points[l.V2] - (Point3D)frameA.Points[l.V1];
                Point3D v2 = (Point3D)frameB.Points[l.V2] - (Point3D)frameB.Points[l.V1];
                els[pos] = v1.Abs();
                double dif = (els[pos] - v2.Abs());
                
                if (abs)
                    lengthSum += v1.Abs();
                else
                    lengthSum += v1.DotProduct(v1);
                
                double divider = invFDiagonal;
                
                if (v1.Abs() > 0)
                    dif = dif / divider;
                else
                    dif = zeroLengthPenalty;


                eds[pos] = dif;                
                pos++;
            }
            return eds;
        }        
    }
}
