﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// For each point of a dynamic mesh, this module calculates the distance between this point in the original mesh and the same point in its distorted version.
    /// The distances are attached to the distorted mesh as a point attribute.
    /// </summary>
    [ModuleInfo("Oldřich Petřík",
        "For each point of a dynamic mesh, this module calculates the distance between this point in the original mesh and the same point in its distorted version.\n" +
        "The distances are attached to the distorted mesh as a point attribute.", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("Original", "Original version of the input mesh.")]
    [PortInfo("Distorted", "Distorted version of the input mesh.")]
    [PortInfo("Painted", "Distorted version of the input mesh with the distances added as a point parameter.")]
    class AnimationDistancePainter:Module
    {
        public AnimationDistancePainter()
        {
            AddInPort("Original", typeof(UniformDataArray));
            AddInPort("Distorted", typeof(UniformDataArray));
            AddOutPort("Painted", typeof(UniformDataArray));
        }

        UniformDataArray result = null;

        public override void Execute()
        {
            if (IsAllDataSame() && IsAllUserSettingsSame())
            {
                SetOutput("Painted", result, true);
            }
            else
            {
                UniformDataArray original = (UniformDataArray)GetInput("Original");
                UniformDataArray distorted = (UniformDataArray)GetInput("Distorted");

                int frameCount = original.Count;
                int pointCount = ((UnstrGrid)original[0]).Points.Count;

                result = new UniformDataArray(distorted.ElemType, frameCount);

                for (int f = 0; f < frameCount; f++)
                {
                    UniformDataArray distance = new UniformDataArray(typeof(Scalar), pointCount);
                    UnstrGrid origMesh = (UnstrGrid)original[f];
                    UnstrGrid distMesh = (UnstrGrid)distorted[f];
                    for (int p = 0; p < pointCount; p++)
                    {
                        Point3D op = (Point3D)origMesh.Points[p];
                        Point3D dp = (Point3D)distMesh.Points[p];
                        Point3D dif = op - dp;
                        distance[p] = new Scalar(dif.Abs());
                    }
                    result[f] = ((UnstrGrid)distorted[f]).ShallowCopy();
                    ((UnstrGrid)result[f]).AddPointAttrs("distortion", distance);
                }

                SetOutput("Painted", result);
            }
        }
    }
}
