using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.ComponentModel;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Module for controlled distortion of dynamic meshes. Parameters of the distortion are set by module properties.
    /// </summary>
    [ModuleInfo("Libor Va", "Module for controlled distortion of dynamic meshes. Parameters of the distortion are set by module properties.", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("Input animation", "Input original animation")]
    [PortInfo("Output animation", "Distorted animation")]
    class AnimationDistorter:Module
    {        
        public AnimationDistorter()
        {
            AddInPort("Input animation", typeof(UniformDataArray));
            AddOutPort("Output animation", typeof(UniformDataArray));
        }

        public enum DistortionType { Gauss, GaussConst, GaussConstLength, GaussTraj, GaussTrajLocal, GaussTemp, SinSpat, SinTemp, SinSpatTraj, SinTempTraj, SinTempTrajLocal };

        private DistortionType distortion;

        private bool paramsChanged = false;

        /// <summary>
        /// Hujaja
        /// </summary>
        [Description("Specifies the type of distortion")]
        public DistortionType Distortion
        {
            get { return distortion; }
            set { distortion = value;
            paramsChanged = true;
        }
        }

        double amount = 1;

        [Description("Specifies the amount of distortion")]
        public double Amount
        {
            get { return amount; }
            set { amount = value;
            paramsChanged = true;
        }
        }

        double freq = 1;

        [Description("Specifies the frequency for sine type of distortion")]
        public double Freq
        {
            get { return freq; }
            set
            {
                freq = value;
                paramsChanged = true;
            }
        }

        static Random r = new Random();

        public static double Gauss(double amount)
        {
            double x, y, r2;

            do
            {
                /* choose x,y in uniform square (-1,-1) to (+1,+1) */

                x = -1 + 2 * r.NextDouble();
                y = -1 + 2 * r.NextDouble();

                /* see if it is in the unit circle */
                r2 = x * x + y * y;
            }
            while (r2 > 1.0 || r2 == 0);

            /* Box-Muller transform */
            return amount * y * Math.Sqrt(-2.0 * Math.Log(r2) / r2);
        }        

        private double[] edgeLengths(UniformDataArray data)
        {
            UnstrGrid first = (UnstrGrid)data[0];
            int[] counts = new int[first.Points.Count];
            double[] lengths = new double[first.Points.Count];
            List<Line> edges = new List<Line>();
            UniformDataArray triangles = first.GetCells(typeof(Triangle))[0];
            foreach (Triangle t in triangles)
            {
                Line l = new Line(t.V1, t.V2, false);
                ProcessLine(data, counts, lengths, edges, l);
                l = new Line(t.V1, t.V3, false);
                ProcessLine(data, counts, lengths, edges, l);
                l = new Line(t.V3, t.V2, false);
                ProcessLine(data, counts, lengths, edges, l);
            }
            for (int i = 0; i < lengths.Length; i++)
            {
                lengths[i] /= counts[i];
            }
            return (lengths);
        }

        private static void ProcessLine(UniformDataArray data, int[] counts, double[] lengths, List<Line> edges, Line l)
        {
            if (!edges.Contains(l))
            {
                double len = 0;
                for (int f = 0; f < data.Count; f++)
                {
                    UnstrGrid frame = (UnstrGrid)data[f];
                    Point3D p1 = (Point3D)frame.Points[l.V1];
                    Point3D p2 = (Point3D)frame.Points[l.V2];
                    len += (p1 - p2).Abs();
                }
                counts[l.V1]++;
                lengths[l.V1] += len;
                counts[l.V2]++;
                lengths[l.V2] += len;
                edges.Add(l);
            }            
        }

        private double[,] movement(int windowRadius, UniformDataArray data)
        {
            UnstrGrid first = (UnstrGrid) data[0];

            double[,] result = new double[data.Count, first.Points.Count];
            for (int f = 0;f<data.Count;f++)
                for (int p = 0; p < first.Points.Count; p++)
                {
                    int start = Math.Max(0, f - windowRadius);
                    int end = Math.Min(data.Count, f + windowRadius);
                    for (int lf = start+1; lf < end; lf++)
                    {
                        Point3D p1 = (Point3D)(((UnstrGrid)data[lf]).Points[p]);
                        Point3D p2 = (Point3D)(((UnstrGrid)data[lf-1]).Points[p]);
                        Point3D dif = p2 - p1;
                        result[f, p] += dif.Abs();
                    }
                    result[f, p] /= (end - start);
                }
            return (result);
        }

        public override void Execute()
        {
            if (paramsChanged || (!IsAllDataSame()))
            {
                UniformDataArray input = (UniformDataArray)GetInput("Input animation");
                int frameCount = input.Count;
                UniformDataArray output = new UniformDataArray(input.ElemType, input.Count);// (UniformDataArray)input.DeepCopy();
                for (int i = 0;i<frameCount;i++)
                {
                    output[i] = input[i].DeepCopy();
                    ((UnstrGrid)output[i]).Points = new UniformDataArray(typeof(Point3D), ((UnstrGrid)input[0]).Points.Count);// (UniformDataArray)((UnstrGrid)input[i]).Points.DeepCopy();
                }

                switch (distortion)
                {
                    case DistortionType.Gauss:
                        for (int f = 0;f<frameCount;f++)
                        {                            
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            for (int p = 0;p<frameOut.Points.Count;p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];                            
                                point.X += Gauss(amount);
                                point.Y += Gauss(amount);
                                point.Z += Gauss(amount);
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.GaussTemp:
                        for (int f = 0; f < frameCount; f++)
                        {
                            Point3D shift = new Point3D(Gauss(amount), Gauss(amount), Gauss(amount));
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];
                                point += shift;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.GaussConst:
                        UnstrGrid first = (UnstrGrid)output[0];
                        double[] distortions = new double[first.Points.Count*3];
                        for (int i = 0;i<distortions.Length;i++)
                        {
                            distortions[i] = Gauss(amount);
                        }
                        for (int f = 0;f<frameCount;f++)
                        {                            
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            for (int p = 0;p<frameOut.Points.Count;p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];                            
                                point.X += distortions[3*p];
                                point.Y += distortions[3*p+1];
                                point.Z += distortions[3*p+2];
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.GaussConstLength:
                        first = (UnstrGrid)output[0];
                        distortions = new double[first.Points.Count * 3];
                        for (int i = 0; i < distortions.Length; i++)
                        {
                            distortions[i] = Gauss(amount);
                        }
                        double[] eLengths = edgeLengths(input);
                        double sum = 0;
                        foreach (double l in eLengths)
                            sum += l;
                        double invavg = eLengths.Length / sum;
                        for (int f = 0; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];
                                point.X += distortions[3 * p] * eLengths[p] * invavg;
                                point.Y += distortions[3 * p + 1] * eLengths[p] * invavg;
                                point.Z += distortions[3 * p + 2] * eLengths[p] * invavg;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.GaussTraj:
                        first = (UnstrGrid)output[0];
                        distortions = new double[first.Points.Count*3];
                        double[] lengths = new double[first.Points.Count];
                        for (int f = 1;f<frameCount;f++)
                        {                            
                            UnstrGrid frameOut = (UnstrGrid)output[f-1];                         
                            UnstrGrid frameOut2 = (UnstrGrid)output[f];
                            for (int p = 0;p<frameOut.Points.Count;p++)
                            {
                                Point3D p1 = (Point3D)frameOut.Points[p];
                                Point3D p2 = (Point3D)frameOut2.Points[p];
                                Point3D dif = p2 - p1;
                                lengths[p] += dif.Abs();
                            }
                        }
                        sum = 0;
                        foreach (double l in lengths)
                            sum += l;
                        invavg = lengths.Length / sum;
                        for (int i = 0;i<lengths.Length;i++)
                        {
                            distortions[3 * i] = Gauss(amount) * lengths[i] * invavg;
                            distortions[3 * i + 1] = Gauss(amount) * lengths[i] * invavg;
                            distortions[3 * i + 2] = Gauss(amount) * lengths[i] * invavg;
                        }
                        for (int f = 0;f<frameCount;f++)
                        {                            
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            for (int p = 0;p<frameOut.Points.Count;p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];                            
                                point.X += distortions[3*p];
                                point.Y += distortions[3*p+1];
                                point.Z += distortions[3*p+2];
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.GaussTrajLocal:
                        double[,] locaLengths = movement(15, input);
                        sum = 0;
                        foreach (double l in locaLengths)
                            sum += l;
                        invavg = (locaLengths.GetLength(0) * locaLengths.GetLength(1)) / sum;
                        for (int f = 0; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];
                                point.X += Gauss(amount) * locaLengths[f, p] * invavg;
                                point.Y += Gauss(amount) * locaLengths[f, p] * invavg;
                                point.Z += Gauss(amount) * locaLengths[f, p] * invavg;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.SinSpat:
                        for (int f = 0;f<frameCount;f++)
                        {                            
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            BoundingBox bb = frameIn.BoundingBox;
                            double dx = freq*2*Math.PI/(bb.MaxPoint[0] - bb.MinPoint[0]);
                            double dy = freq*2*Math.PI/(bb.MaxPoint[1] - bb.MinPoint[1]);
                            double dz = freq * 2 * Math.PI / (bb.MaxPoint[2] - bb.MinPoint[2]);
                            for (int p = 0;p<frameOut.Points.Count;p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];                            
                                point.X += Math.Sin((point.X-bb.MinPoint[0])*dx)*amount;
                                point.Y += Math.Sin((point.Y-bb.MinPoint[1])*dy)*amount;
                                point.Z += Math.Sin((point.Z-bb.MinPoint[2])*dz)*amount;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.SinSpatTraj:
                        /*first = (UnstrGrid)output[0];
                        //distortions = new double[first.Points.Count * 3];
                        lengths = new double[first.Points.Count];
                        for (int f = 1; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f - 1];
                            UnstrGrid frameOut2 = (UnstrGrid)output[f];
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D p1 = (Point3D)frameOut.Points[p];
                                Point3D p2 = (Point3D)frameOut.Points[p];
                                Point3D dif = p2 - p1;
                                lengths[p] += dif.Abs();
                            }
                        }
                        sum = 0;
                        foreach (double l in lengths)
                            sum += l;
                        invavg = lengths.Length / sum;*/
                        locaLengths = movement(15, input);
                        sum = 0;
                        foreach (double l in locaLengths)
                            sum += l;
                        invavg = (locaLengths.GetLength(0) * locaLengths.GetLength(1)) / sum;
                        for (int f = 0; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            BoundingBox bb = frameIn.BoundingBox;
                            double dx = freq * 2 * Math.PI / (bb.MaxPoint[0] - bb.MinPoint[0]);
                            double dy = freq * 2 * Math.PI / (bb.MaxPoint[1] - bb.MinPoint[1]);
                            double dz = freq * 2 * Math.PI / (bb.MaxPoint[2] - bb.MinPoint[2]);
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];
                                point.X += Math.Sin((point.X - bb.MinPoint[0]) * dx) * amount * locaLengths[f, p]*invavg;
                                point.Y += Math.Sin((point.Y - bb.MinPoint[1]) * dy) * amount * locaLengths[f, p] * invavg;
                                point.Z += Math.Sin((point.Z - bb.MinPoint[2]) * dz) * amount * locaLengths[f, p] * invavg;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.SinTemp:
                        for (int f = 0; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            double shift = Math.Sin((freq * f * Math.PI * 2) / frameCount)*amount;
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];
                                point.X += shift;
                                point.Y += shift;
                                point.Z += shift;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.SinTempTraj:
                        first = (UnstrGrid)output[0];
                        //distortions = new double[first.Points.Count * 3];
                        lengths = new double[first.Points.Count];
                        for (int f = 1; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f - 1];
                            UnstrGrid frameOut2 = (UnstrGrid)output[f];
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D p1 = (Point3D)frameOut.Points[p];
                                Point3D p2 = (Point3D)frameOut.Points[p];
                                Point3D dif = p2 - p1;
                                lengths[p] += dif.Abs();
                            }
                        }
                        sum = 0;
                        foreach (double l in lengths)
                            sum += l;
                        invavg = lengths.Length / sum;
                        for (int f = 0; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            double shift = Math.Sin((freq * f * Math.PI * 2) / frameCount)*amount;
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];
                                point.X += shift * lengths[p]*invavg;
                                point.Y += shift * lengths[p]*invavg;
                                point.Z += shift * lengths[p]*invavg;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                    case DistortionType.SinTempTrajLocal:
                        locaLengths = movement(15, input);
                        sum = 0;
                        foreach (double l in locaLengths)
                            sum += l;
                        invavg = (locaLengths.GetLength(0)*locaLengths.GetLength(1))/sum;
                        for (int f = 0; f < frameCount; f++)
                        {
                            UnstrGrid frameOut = (UnstrGrid)output[f];
                            UnstrGrid frameIn = (UnstrGrid)input[f];
                            double shift = Math.Sin((freq * f * Math.PI * 2) / frameCount) * amount;
                            for (int p = 0; p < frameOut.Points.Count; p++)
                            {
                                Point3D point = (Point3D)frameIn.Points[p];
                                point.X += shift * locaLengths[f,p] * invavg;
                                point.Y += shift * locaLengths[f,p] * invavg;
                                point.Z += shift * locaLengths[f,p] * invavg;
                                frameOut.Points[p] = point;
                            }
                        }
                        break;

                }

                SetOutput("Output animation", output, false);
                WriteLog("Distortion", distortion);
                WriteLog("Amount", amount);
            }
        }
    }
}
