﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// For each point of a dynamic mesh, this module calculates square distance between this point in the original mesh and the same point in its distorted version.
    /// The distances are attached to the distorted mesh as a point attribute.
    /// </summary>
    [ModuleInfo("Oldřich Petřík",
        "For each point of a dynamic mesh, this module calculates square distance between this point in the original mesh and the same point in its distorted version.\n" +
        "The distances are attached to the distorted mesh as a point attribute.", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("Original", "Original version of the input mesh.")]
    [PortInfo("Distorted", "Distorted version of the input mesh.")]
    [PortInfo("Painted", "Distorted version of the input mesh with square distances added as a point parameter.")]
    public class AnimationSqrDistPainter : Module
    {
        private string _attrName = "SqrError";

        UniformDataArray result = null;

        /// <summary>
        /// Name of the resulting attribute.
        /// </summary>
        public string AttributeName
        {
            get { return _attrName; }
            set { _attrName = value; }
        }

        /// <summary>
        /// AnimationSqrDistPainter constructor.
        /// </summary>
        public AnimationSqrDistPainter()
        {
            AddInPort("Original", typeof(UniformDataArray));
            AddInPort("Distorted", typeof(UniformDataArray));
            AddOutPort("Painted", typeof(UniformDataArray));
        }

        ///<inheritdoc />
        public override void Execute()
        {
            if (IsAllDataSame() && IsAllUserSettingsSame())
            {
                SetOutput("Painted", result, true);
            }
            else
            {
                UniformDataArray original = (UniformDataArray)GetInput("Original");
                UniformDataArray distorted = (UniformDataArray)GetInput("Distorted");

                int frameCount = original.Count;
                int pointCount = ((UnstrGrid)original[0]).Points.Count;
                double val, max;
                int maxFrame, maxPoint;

                result = new UniformDataArray(distorted.ElemType, frameCount);
                max = double.MinValue;
                maxFrame = -1;
                maxPoint = -1;

                for (int f = 0; f < frameCount; f++)
                {
                    UniformDataArray distance = new UniformDataArray(typeof(Scalar), pointCount);
                    UnstrGrid origMesh = (UnstrGrid)original[f];
                    UnstrGrid distMesh = (UnstrGrid)distorted[f];
                    for (int p = 0; p < pointCount; p++)
                    {
                        Point3D op = (Point3D)origMesh.Points[p];
                        Point3D dp = (Point3D)distMesh.Points[p];
                        val = op.SquareDistanceTo(dp);

                        distance[p] = (Scalar)val;
                        if (val > max)
                        {
                            max = val;
                            maxFrame = f;
                            maxPoint = p;
                        }
                    }
                    result[f] = ((UnstrGrid)distorted[f]).ShallowCopy();
                    ((UnstrGrid)result[f]).AddPointAttrs(_attrName, distance);
                }

                Console.WriteLine("{0} has finished:", Name);
                Console.WriteLine("  maximum={0} (frame {1}, point {2})", max, maxFrame, maxPoint);
                Console.WriteLine("  diagonal={0}", ((UnstrGrid)(original[0])).BoundingBox.MaxDiag);

                SetOutput("Painted", result, false);
            }
        }
    }
}
