using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Strings;
using Zcu.Mve.Visualization;
using System.Windows.Forms.Design;
using System.ComponentModel;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// File browser dialog for .wrl (VRML) files.
    /// </summary>
    public class FileNameEditorVRML : FileNameEditor
    {
        /// <summary>
        /// Dialog initialization method.
        /// </summary>
        /// <param name="openFileDialog">dialog reference</param>
        protected override void InitializeDialog(System.Windows.Forms.OpenFileDialog openFileDialog)
        {
            openFileDialog.Filter = "VRML files (*.wrl)|*.wrl|All files (*.*)|*.*";
        }
    }

    /// <summary>
    /// Loads an animation from a VRML 2.0 file
    /// </summary>
    [ModuleInfo("Libor Va, Oldich Petk", "Loads an animation from a VRML 2.0", CategoryType.Loader, CategoryData.MeshAnimated, IconName = "vrml.ico")]
    [PortInfo("Animation", "The loaded animation.")]
    [PortInfo("FileName", "Name of the file to be loaded. Should include the .wrl extension. When not connected, then the filename is controlled by the FileName property.")]
    class AnimationVRMLLoader : Module
    {
        public AnimationVRMLLoader()
        {
            AddInPort("FileName", typeof(MveString), false);
            AddOutPort("Animation", typeof(UniformDataArray));
        }

        string fileName = "animation.wrl";

        [Editor(typeof(FileNameEditorVRML), typeof(System.Drawing.Design.UITypeEditor))]
        [Description("Name of the file to be loaded. Should include the .wrl extension. Only used when the FileName port is not connected.")]
        public string FileName
        {
            get { return fileName; }
            set
            {
                fileName = value;
                paramsChanged = true;
            }
        }

        UniformDataArray result;

        bool paramsChanged = true;

        bool reload = false;

        public bool Reload
        {
            get { return reload; }
            set { reload = value; }
        }

        private StringBuilder ReadFile(string fileName, out int pointPos, out int normalPos)
        {
            using (StreamReader sr = new StreamReader(fileName))
            {
                Stream stream = sr.BaseStream;
                long size = stream.Length;
                long pos = 0;
                long last = 0;
                long step = 1024 * 1024 * 8;
                string line;
                StringBuilder file = new StringBuilder();
                int p;
                pointPos = -1;
                normalPos = -1;

                while (true)
                {
                    line = sr.ReadLine();
                    if (line == null) break;
                    line = line.Trim();

                    if (line.Length == 0 || line[0] == '#') continue;

                    if (pointPos < 0)
                    {
                        p = line.IndexOf("CoordinateInterpolator");
                        if (p >= 0) pointPos = file.Length + p + 22;
                    }

                    if (normalPos < 0)
                    {
                        p = line.IndexOf("NormalInterpolator");
                        if (p >= 0) normalPos = file.Length + p + 18;
                    }

                    file.Append(line);
                    file.Append(' ');

                    pos = stream.Position;
                    if (pos - last > step)
                    {
                        ProgressInfo((double)pos / size);
                        last = pos;
                    }
                }

                if (pointPos < 0) throw new MveException(this, "Vertex definition not found.");
                return file;
            }
        }

        private string[] GetValues(string range)
        {
            char[] sep = { ' ', ',', '\t', '\n', '\r' };
            return range.Split(sep, StringSplitOptions.RemoveEmptyEntries);
        }

        private int GetPointCount(ref string file)
        {
            Regex pat = new Regex(@"(?<=\s+point\s+\[\s*)[^\]]*", RegexOptions.IgnoreCase);
            Match mat = pat.Match(file);
            string[] values = GetValues(mat.Value);
            if (values.Length % 3 != 0) throw new MveException(this, "Invalid number of point coordinates.");
            return values.Length / 3;
        }

        private UniformDataArray GetTriangles(ref string file)
        {
            Regex pat = new Regex(@"(?<=\s+coordIndex\s+\[\s*)[^\]]*", RegexOptions.IgnoreCase);
            Match mat = pat.Match(file);
            string[] values = GetValues(mat.Value);
            List<Triangle> res = new List<Triangle>();
            int j = 0;
            int[] v = new int[4];

            for (int i = 0; i < values.Length; i++)
            {
                v[j] = int.Parse(values[i]);
                if (v[j] < 0)
                {
                    if (j == 3) res.Add(new Triangle(v[0], v[1], v[2]));
                    else throw new MveException(this, "Invalid face type. Only triangle faces are supported.");
                    j = 0;
                }
                else j++;
            }
            if (j != 0) throw new MveException(this, "Face definition ended improperly.");

            return UniformDataArray.FromList(res);
        }

        private UniformDataArray[] GetPoints(ref string file, int pointCount, int pointPos)
        {
            Regex pat = new Regex(@"(?<=\s+key\s+\[\s*)[^\]]*");
            Match mat = pat.Match(file, pointPos);
            int frameCount = GetValues(mat.Value).Length;

            pat = new Regex(@"(?<=\s+keyValue\s+\[\s*)[^\]]*");
            mat = pat.Match(file, mat.Index + mat.Length);
            string[] values = GetValues(mat.Value);
            if (values.Length != 3 * frameCount * pointCount) throw new MveException(this, "Invalid number of vertex definitions.");

            UniformDataArray[] res = new UniformDataArray[frameCount];
            UniformDataArray frame;
            double p1, p2, p3;
            for (int f = 0, i = 0; f < frameCount; f++)
            {
                frame = new UniformDataArray(typeof(Point3D), pointCount);
                for (int p = 0; p < pointCount; p++)
                {
                    p1 = double.Parse(values[i++], Globals.nfi);
                    p2 = double.Parse(values[i++], Globals.nfi);
                    p3 = double.Parse(values[i++], Globals.nfi);
                    frame[p] = new Point3D(p1, p2, p3);
                }
                res[f] = frame;
                if (f % 20 == 0)
                {
                    ProgressInfo((f + 1) / (double)frameCount);
                }
            }
            return res;
        }

        private UniformDataArray[] GetNormals(ref string file, int frameCount, int pointCount, int normalPos)
        {
            Regex pat = new Regex(@"(?<=\s+key\s+\[\s*)[^\]]*");
            Match mat = pat.Match(file, normalPos);
            if (frameCount != GetValues(mat.Value).Length) throw new MveException(this, "Number of frames inconsistent between vertices and normals.");

            pat = new Regex(@"(?<=keyValue\s+\[\s*)[^\]]*");
            mat = pat.Match(file, mat.Index + mat.Length);
            string[] values = GetValues(mat.Value);

            if (values.Length != 3 * frameCount * pointCount) throw new MveException(this, "Invalid number of normal definitions.");
            UniformDataArray[] res = new UniformDataArray[frameCount];
            UniformDataArray frame;
            double n1, n2, n3;
            for (int f = 0, i = 0; f < frameCount; f++)
            {
                frame = new UniformDataArray(typeof(Normal3D), pointCount);
                for (int p = 0; p < pointCount; p++)
                {
                    n1 = double.Parse(values[i++], Globals.nfi);
                    n2 = double.Parse(values[i++], Globals.nfi);
                    n3 = double.Parse(values[i++], Globals.nfi);
                    frame[p] = new Normal3D(n1, n2, n3);
                }
                res[f] = frame;
                if (f % 20 == 0)
                {
                    ProgressInfo((f + 1) / (double)frameCount);
                }
            }
            return res;
        }

        public override void Execute()
        {
            if (GetInPort("FileName").Link != null)
            {
                if (!IsAllDataSame())
                {
                    FileName = ((MveString)GetInput("FileName")).Value;
                }
            }

            WriteLog("Dataset", fileName);

            if (paramsChanged || reload)
            {
                ProgressInfo(0, 0, 3, "Loading file...");
                int pointPos, normalPos;
                string file = ReadFile(fileName, out pointPos, out normalPos).ToString();

                int pointCount = GetPointCount(ref file);
                UniformDataArray cells = GetTriangles(ref file);

                ProgressInfo(0, 1, 3, "Reading Points...");
                UniformDataArray[] points = GetPoints(ref file, pointCount, pointPos);
                int frameCount = points.Length;

                UniformDataArray[] normals = null;
                if (normalPos >= 0)
                {
                    ProgressInfo(0, 2, 3, "Reading Normals...");
                    normals = GetNormals(ref file, frameCount, pointCount, normalPos);
                }
                result = new UniformDataArray(typeof(TriangleMesh), frameCount);
                TriangleMesh frame;
                for (int f = 0; f < frameCount; f++)
                {
                    frame = new TriangleMesh();
                    frame.Points = points[f];
                    frame.AddCells("Triangles", cells);
                    if (normals != null) frame.AddPointAttrs("Normal3D", normals[f]);
                    result[f] = frame;
                }

                ProgressInfo(1, 2, 3, "Done");

                SetOutput("Animation", result);
                paramsChanged = false;
            }
            else
            {
                SetOutput("Animation", result, true);
            }
        }
    }
}
