using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using Zcu.Mve.Strings;
using System.IO;
using System.ComponentModel;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Saves an animation into a VRML 2.0 file
    /// </summary>
    [ModuleInfo("Libor Va", "Saves an animation into a VRML 2.0", CategoryType.Saver, CategoryData.MeshAnimated, IconName = "vrml.ico")]
    [PortInfo("Animation", "Animation to be saved.")]
    [PortInfo("FileName", "Name of the file to be saved. Should include the .wrl extension. When not connected, then the filename is controlled by the FileName property.")]
    class AnimationVRMLSaver:Module
    {
        string fileName = "animation.wrl";

        [Description("Name of the file to be saved. Should include the .wrl extension. Only used when the FileName port is not connected.")]
        [Editor(typeof(FileNameEditorVRML), typeof(System.Drawing.Design.UITypeEditor))]
        public string FileName
        {
            get { return fileName; }
            set { fileName = value; }
        }

        //private bool paramsChanged = true;

        public AnimationVRMLSaver()
        {
            AddInPort("Animation", typeof(UniformDataArray));
            AddInPort("FileName", typeof(MveString), false);
        }

        private bool saveNormals = false;

        [Description("Determines whether or not to save normals with the animation. When true, then there should be a Normal3D type point attribute assigned to every frame, containing the vertex normals.")]
        public bool SaveNormals
        {
          get { return saveNormals; }
          set { saveNormals = value; }
        }

        private double creaseAngle = 1.5;

        public double CreaseAngle
        {
            get { return creaseAngle; }
            set { creaseAngle = value; }
        }

        private bool optimize = true;

        public bool Optimize
        {
            get { return optimize; }
            set { optimize = value; }
        }

        private int digitCount = 4;

        public int DigitCount
        {
            get { return digitCount; }
            set { digitCount = value; }
        }


        public override void Execute()
        {

            string fs = "r";
            if (GetInPort("FileName").Link != null)
            {
                fileName = ((MveString)GetInput("FileName")).Value;
            }

            UniformDataArray input = (UniformDataArray)GetInput("Animation");

            UniformDataArray points;
            double span = 1;
            double multiplier = Math.Pow(10, digitCount);
            double xmin, ymin, zmin;
            xmin = ymin = zmin = double.MaxValue;
            double xmax, ymax, zmax;
            xmax = ymax = zmax = double.MinValue;

        
            for (int i = 0; i < input.Count; i++)
            {
                UnstrGrid frame = (UnstrGrid)input[i];
                points = frame.Points;
                for (int j = 0; j < points.Count; j++)
                {
                    Point3D p = (Point3D)points[j];
                    if (p.X > xmax)
                        xmax = p.X;
                    if (p.Y > ymax)
                        ymax = p.Y;
                    if (p.Z > zmax)
                        zmax = p.Z;
                    if (p.X < xmin)
                        xmin = p.X;
                    if (p.Y < ymin)
                        ymin = p.Y;
                    if (p.Z < zmin)
                        zmin = p.Z;
                }
            }
            span = xmax - xmin;
            if ((ymax - ymin) > span)
                span = ymax - ymin;
            if ((zmax - zmin) > span)
                span = zmax - zmin;


            using (StreamWriter sw = new StreamWriter(fileName, false))
            {
                sw.WriteLine(@"#VRML V2.0 utf8
# Produced by Libor VASA, lvasa@kiv.zcu.cz");

                double xpos, ypos, zpos;
                if (optimize)
                {
                    xpos = (((xmax - xmin) / 2) / span) * multiplier;
                    ypos = (((ymax - ymin) / 2) / span) * multiplier;
                    zpos = (((zmax - zmin) / 2) / span) * multiplier + multiplier;
                }
                else
                {
                    xpos = (xmin + xmax) / 2;
                    ypos = (ymin + ymax) / 2 - span * 0.1;
                    zpos = (zmin + zmax) / 2 + span;
                }

                sw.WriteLine(@"Viewpoint {
	fieldOfView 0.39
	position " + xpos.ToString("0.00", Globals.nfi) + " " + ypos.ToString("0.00", Globals.nfi) + " " + zpos.ToString("0.00", Globals.nfi) + @"
	orientation 1 0 0 -0.2
	jump TRUE
	description ""v1""
}");


                sw.WriteLine(@"
Group {
	children [
		Shape {
			appearance Appearance {
				material Material {
					diffuseColor 1.0 0.83 0.0
					ambientIntensity 0.3333
					specularColor 0.8 0.8 0.8
					shininess 0.5
					transparency 0
				}
			}
			geometry IndexedFaceSet {
				solid FALSE
                creaseAngle " + creaseAngle.ToString(Globals.nfi) + @"
				coord DEF co Coordinate {
					point [");

                UnstrGrid first = (UnstrGrid)input[0];
                points = first.Points;
                int pointCount = points.Count;
                for (int i = 0; i < pointCount; i++)
                {
                    Point3D p = (Point3D)points[i];
                    if (optimize)
                    {
                        int x = (int)(((p.X - xmin) / span) * multiplier);
                        int y = (int)(((p.Y - ymin) / span) * multiplier);
                        int z = (int)(((p.Z - zmin) / span) * multiplier);
                        sw.WriteLine(x.ToString() + " " + y.ToString() + " " + z.ToString() + ",");
                    }
                    else
                        sw.WriteLine(p.X.ToString(fs, Globals.nfi) + " " + p.Y.ToString(fs, Globals.nfi) + " " + p.Z.ToString(fs, Globals.nfi) + ",");
                }
                sw.WriteLine(@"]
}
coordIndex [");
                UniformDataArray triangles = first.GetCells(typeof(Triangle))[0];
                foreach (Triangle t in triangles)
                {
                    sw.WriteLine(t.V1 + "," + t.V2 + "," + t.V3 + ",-1,");
                }
                sw.WriteLine(@"]
				texCoordIndex [ 
				]");
                if (saveNormals)
                {
                    UniformDataArray normals = first.GetPointAttrs(typeof(Normal3D))[0];
                    sw.WriteLine(@"normal DEF no Normal {
					vector [");

                    foreach (Normal3D n in normals)
                    {
                        if (optimize)
                        {
                            int x = (int)(n.Nx * 50);
                            int y = (int)(n.Ny * 50);
                            int z = (int)(n.Nz * 50);
                            sw.WriteLine(x.ToString() + " " + y.ToString() + " " + z.ToString() + ",");
                        }
                        else
                            sw.WriteLine(n.Nx.ToString(fs, Globals.nfi) + " " + n.Ny.ToString(fs, Globals.nfi) + " " + n.Nz.ToString(fs, Globals.nfi) + ",");
                    }
                    sw.WriteLine(@"]
				}");
                }
                sw.WriteLine(@" 
			}
		}
		DEF ci CoordinateInterpolator {
			key [");
                for (int i = 0; i < input.Count; i++)
                {
                    sw.WriteLine(((double)i / (input.Count - 1)).ToString("0.0000", Globals.nfi));
                }

                sw.WriteLine(@"]
			keyValue [");

                for (int f = 0; f < input.Count; f++)
                {
                    UnstrGrid mesh = (UnstrGrid)input[f];
                    points = mesh.Points;
                    for (int i = 0; i < pointCount; i++)
                    {
                        Point3D p = (Point3D)points[i];
                        if (optimize)
                        {
                            int x = (int)(((p.X - xmin) / span) * multiplier);
                            int y = (int)(((p.Y - ymin) / span) * multiplier);
                            int z = (int)(((p.Z - zmin) / span) * multiplier);
                            sw.WriteLine(x.ToString() + " " + y.ToString() + " " + z.ToString() + ",");
                        }
                        else
                            sw.WriteLine(p.X.ToString(fs, Globals.nfi) + " " + p.Y.ToString(fs, Globals.nfi) + " " + p.Z.ToString(fs, Globals.nfi) + ",");
                    }
                }

                sw.Write(@"]
		}");

                if (saveNormals)
                {
                    sw.WriteLine(@"DEF ni NormalInterpolator {
			key [");
                    for (int i = 0; i < input.Count; i++)
                    {
                        sw.WriteLine(((double)i / (input.Count - 1)).ToString("0.0000", Globals.nfi));
                    }

                    sw.WriteLine(@"]
			keyValue [");

                    for (int f = 0; f < input.Count; f++)
                    {
                        UnstrGrid mesh = (UnstrGrid)input[f];
                        UniformDataArray normals = mesh.GetPointAttrs(typeof(Normal3D))[0];
                        for (int i = 0; i < pointCount; i++)
                        {
                            Normal3D n = (Normal3D)normals[i];
                            if (optimize)
                            {
                                int x = (int)(n.Nx * 50);
                                int y = (int)(n.Ny * 50);
                                int z = (int)(n.Nz * 50);
                                sw.WriteLine(x.ToString() + " " + y.ToString() + " " + z.ToString() + ",");
                            }
                            else
                                sw.WriteLine(n.Nx.ToString(fs, Globals.nfi) + " " + n.Ny.ToString(fs, Globals.nfi) + " " + n.Nz.ToString(fs, Globals.nfi) + ",");
                        }
                    }

                    sw.Write(@"]
		}");
                }

                sw.WriteLine(@"DEF ts TimeSensor {
			cycleInterval ");
                sw.WriteLine((input.Count / 25.0).ToString(Globals.nfi));
                sw.WriteLine(@"loop TRUE
		}
	]
}
ROUTE ts.fraction_changed TO ci.set_fraction
ROUTE ci.value_changed TO co.point");
                if (saveNormals)
                {
                    sw.WriteLine(@"ROUTE ts.fraction_changed TO ni.set_fraction
ROUTE ni.value_changed TO no.vector");

                }
                sw.Close();
            }
        }
    }
}
