﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Strings;
using Zcu.Mve.Compression;
using Zcu.Mve.Numerics;
using System.IO;
using System.ComponentModel;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Compression module performing the coddyac compression of an animated mesh. Outputs a simulated result of compression with original vertex indices.
    /// </summary>
    [ModuleInfo("Libor Váša", "Compression module performing the coddyac compression of an animated mesh. Outputs a simulated result of compression with original vertex indices.", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("AnimationIn", "Input animation, each frame should have the same connectivity")]
    [PortInfo("AnimationOut", "Simulated decompression result with original vertex indices (can be directly compared with original")]
    [PortInfo("bpfv", "Resulting data rate int bit per frame and vertex")]
    class CompressorCoddyac:Module
    {
        public CompressorCoddyac()
        {
            AddInPort("AnimationIn", typeof(UniformDataArray));
            AddOutPort("AnimationOut", typeof(UniformDataArray));
            AddOutPort("bpfv", typeof(Scalar));
        }

        string fileName = "data.dat";

        /// <summary>
        /// Name of the file to write to.
        /// </summary>
        [Description("Name of the file to write to.")]
        public string FileName
        {
            get { return fileName; }
            set { fileName = value; }
        }

        UniformDataArray result = null;
        Scalar bpfv = new Scalar(0);
        double[][] trajectories;
        PCA pca;
        double avgEdgeLength = 0;

        int baseSize = 50;

        public int BaseSize
        {
            get { return baseSize; }
            set { baseSize = value; }
        }

        double basisQ = 20;

        public double BasisQ
        {
            get { return basisQ; }
            set { basisQ = value; }
        }

        double quantizationBits = 5;

        public double QuantizationBits
        {
            get { return quantizationBits; }
            set { quantizationBits = value; }
        }

        public override void Execute()
        {
            if (IsAllDataSame() && (IsAllUserSettingsSame()))
            {
                SetOutput("AnimationOut", result, true);
                SetOutput("bpfv", bpfv);
            }
            else
            {
                // output stream for most of the data
                FileStream outStream = null;
                try
                {
                    outStream = new FileStream(fileName, FileMode.Create);
                    BinaryWriter writer = new BinaryWriter(outStream);
                    trajectories = null;

                    UniformDataArray input = (UniformDataArray)GetInput("AnimationIn");
                    int meshCount = input.Count;
                    int pointCount = ((TriangleMesh)input[0]).Points.Count;
                    TriangleMesh firstFrame = (TriangleMesh)input[0];

                    // write basic data
                    writer.Write(pointCount);

                    // store trajectories into an array
                    initTrajectories(input, meshCount, pointCount);

                    // trajectory space PCA
                    if (!IsDataSame("AnimationIn"))
                    {
                        pca = new PCA(PCA.PCAMethod.direct, true);
                        pca.Data = trajectories;
                        pca.ComputeBasis();
                        pca.BackupBasis();
                    }
                    else
                    {
                        pca.RestoreBasis();
                    }

                    UniformDataArray triangles = ((TriangleMesh)input[0]).GetCells(typeof(Triangle))[0];
                    if (!IsDataSame("AnimationIn"))
                        avgEdgeLength = getAvgEdgeLength(triangles, input);

                    // prepare quantization
                    double caseCount = Math.Pow(2, quantizationBits);
                    double delta = 0;
                    delta = avgEdgeLength / caseCount;
                    quantizer = new UniformQuantizer(delta);
                    writer.Write(delta);

                    // compute feature vectors for trajectories                
                    ThreadedExecution<double[], double[]> threadExecution2 = new ThreadedExecution<double[], double[]>(trajectories, this.GetCoefs, this.ProgressInfo);
                    double[][] rempResult = threadExecution2.Execute();
                    this.trajectories = rempResult;

                    // encode the PCA basis
                    double basisCaseCount = Math.Pow(2, basisQ);
                    double encodedBasisSize = pca.EncodeBasisCobra(outStream, baseSize, 1 / basisCaseCount) * 8;

                    // traverse mesh, encode feature vectors
                    residuals = new List<int>[baseSize];
                    for (int i = 0; i < baseSize; i++)
                        residuals[i] = new List<int>();
                    firstTriangles = new List<int>();

                    MemoryStream topoStream = new MemoryStream();
                    EdgebreakerCompressor topoCompressor = new EdgebreakerCompressor();
                    topoCompressor.OutputStream = topoStream;
                    topoCompressor.firstTriangleEventHandler += new EventHandler(this.encodeTriangle);
                    topoCompressor.newVectorEventHandler += new EventHandler(this.predictAndEncode);
                    topoCompressor.Run(triangles, pointCount);

                    WriteLog("Encoded topology[B]", topoStream.Length);

                    // encode residuals into the stream
                    ArithCoder.encode(firstTriangles.ToArray(), outStream);
                    for (int i = 0; i < baseSize; i++)
                    {
                        ArithCoder.encode(residuals[i].ToArray(), outStream);
                    }

                    // write topology into the out stream
                    topoStream.Position = 0;
                    topoStream.WriteTo(outStream);

                    // reconstruct the trajectories
                    for (int i = 0; i < trajectories.Length; i++)
                    {
                        trajectories[i] = pca.Decode(trajectories[i]);
                    }

                    // reconstruct the result
                    if (!IsDataSame("AnimationIn"))
                    {
                        result = new UniformDataArray(typeof(TriangleMesh), input.Count);
                        for (int i = 0; i < meshCount; i++)
                        {
                            TriangleMesh tm = new TriangleMesh();
                            tm.AddCells("Triangles", triangles);
                            UniformDataArray newPoints = new UniformDataArray(typeof(Point3D), pointCount);
                            tm.Points = newPoints;
                            result[i] = tm;
                        }
                    }
                    for (int i = 0; i < meshCount; i++)
                    {
                        UniformDataArray points = ((UnstrGrid)result[i]).Points;
                        for (int j = 0; j < pointCount; j++)
                        {
                            Point3D p = new Point3D(trajectories[j][i * 3],
                                trajectories[j][i * 3 + 1],
                                trajectories[j][i * 3 + 2]);
                            points[j] = p;
                        }
                    }

                    //set the outputs
                    int fv = pointCount * meshCount;
                    int outStreamLength = (int)outStream.Position;
                    double rate = (double)outStreamLength / fv;
                    SetOutput("AnimationOut", result, false);
                    SetOutput("bpfv", new Scalar((double)(8*outStream.Position) / (pointCount * input.Count)));
                }
                finally
                {
                    if (outStream!=null)
                        outStream.Close();
                }
            }
        }

        private double[] GetCoefs(double[] data)
        {
            return (this.pca.GetCoefs(data, baseSize));
        }

        List<int> firstTriangles;
        List<int>[] residuals;
        IQuantizer quantizer;

        private void encodeTriangle(object source, EventArgs args)
        {
            int v1, v2, v3;
            TopologyEventData data = (TopologyEventData)args;
            v1 = data.LeftCornerVertexIndex;
            v2 = data.RightCornerVertexIndex;
            v3 = data.CornerVertexIndex;
            for (int i = 0; i < baseSize; i++)
            {
                int quantized = quantizer.Quantize(trajectories[v1][i]);
                firstTriangles.Add(quantized);
                trajectories[v1][i] = quantizer.Restore(quantized);

                quantized = quantizer.Quantize(trajectories[v2][i]);
                firstTriangles.Add(quantized);
                trajectories[v2][i] = quantizer.Restore(quantized);

                quantized = quantizer.Quantize(trajectories[v3][i]);
                firstTriangles.Add(quantized);
                trajectories[v3][i] = quantizer.Restore(quantized);
            }
        }

        private void predictAndEncode(object sender, EventArgs args)
        {
            int v1, v2, v3, v4;
            TopologyEventData data = (TopologyEventData)args;
            v2 = data.LeftCornerVertexIndex;
            v3 = data.RightCornerVertexIndex;
            v4 = data.CornerVertexIndex;
            v1 = data.OpositeCornerVertexIndex;
            for (int i = 0; i < baseSize; i++)
            {
                double prediction = trajectories[v2][i] + trajectories[v3][i] - trajectories[v1][i];
                double residual = trajectories[v4][i] - prediction;
                int quantized = quantizer.Quantize(residual);
                residuals[i].Add(quantized);
                trajectories[v4][i] = prediction + quantizer.Restore(quantized);
            }
        }

        private void initTrajectories(UniformDataArray input, int meshCount, int pointCount)
        {
            trajectories = new double[pointCount][];
            for (int i = 0; i < pointCount; i++)
            {
                trajectories[i] = new double[meshCount * 3];
                for (int j = 0; j < meshCount; j++)
                {
                    TriangleMesh mesh = (TriangleMesh)input[j];
                    Point3D p = (Point3D)mesh.Points[i];
                    trajectories[i][j * 3] = p.X;
                    trajectories[i][j * 3 + 1] = p.Y;
                    trajectories[i][j * 3 + 2] = p.Z;
                }
            }
        }

        private double getAvgEdgeLength(UnstrGrid frame)
        {
            double sum = 0;
            foreach (Line l in edges.Keys)
            {
                Point3D p1 = (Point3D)frame.Points[l.V1];
                Point3D p2 = (Point3D)frame.Points[l.V2];
                sum += (p1 - p2).Abs();
            }
            return (sum / edges.Count);
        }

        Dictionary<Line, object> edges;

        private double getAvgEdgeLength(UniformDataArray triangles, UniformDataArray input)
        {
            edges = new Dictionary<Line, object>();
            foreach (Triangle t in triangles)
            {
                Line l = new Line(t.V1, t.V2, false);
                if (!(edges.ContainsKey(l)))
                    edges.Add(l, null);
                l = new Line(t.V2, t.V3, false);
                if (!(edges.ContainsKey(l)))
                    edges.Add(l, null);
                l = new Line(t.V1, t.V3, false);
                if (!(edges.ContainsKey(l)))
                    edges.Add(l, null);
            }
            double sum = 0;
            ThreadedExecution<UnstrGrid, double> threadExecution = new ThreadedExecution<UnstrGrid, double>((UnstrGrid[])input.Array, getAvgEdgeLength, this.ProgressInfo);
            double[] res = threadExecution.Execute();
            foreach (double d in res)
                sum += d;
            edges = null;
            GC.Collect();
            return (sum / input.Count);
        }        
    }
}
