﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.IO;
using System.Windows.Forms.Design;
using System.ComponentModel;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Decompression module for data compressed by the CompressorCoddyac module. Outputs a dynamic mesh of shared connectivity. Data is loaded from a file specified by the FileName property.
    /// </summary>
    [ModuleInfo("Libor Váša", "Decompression module for data compressed by the CompressorCoddyac module. Outputs a dynamic mesh of shared connectivity. Data is loaded from a file specified by the FileName property.", CategoryType.Loader, CategoryData.MeshAnimated)]
    [PortInfo("Animation", "Resulting animation. It is likely to have different vertex indices, and therefore it cannot be directly compared with the original. Instead, compare with an animation processed by the DummyCompressor module, which performs renumbering of vertices.")]
    class DecompressorCoddyac:Module
    {
        public DecompressorCoddyac()
        {
            AddOutPort("Animation", typeof(UniformDataArray));
        }
        
        string fileName = "animation.dat";

        /// <summary>
        /// Name of the file to load the data from.
        /// </summary>
        [Description("Name of the file to load the data from.")]
        [Editor(typeof(FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
        public string FileName
        {
            get { return fileName; }
            set { fileName = value; }
        }

        /// <summary>
        /// Adds a new triangle to the animation. Called by the edgebreaker traversal engine.
        /// </summary>
        /// <param name="sender">Not used</param>
        /// <param name="args">TopologyEventData instance specifying the indices of the new triangle.</param>
        private void OnNewTriangle(object sender, EventArgs args)
        {
            int v1, v2, v3;
            TopologyEventData data = (TopologyEventData)args;
            v1 = data.LeftCornerVertexIndex;
            v2 = data.RightCornerVertexIndex;
            v3 = data.CornerVertexIndex;
            for (int i = 0; i < baseSize; i++)
            {
                int quantized = encodedFirstTriangles[firstTrianglePointer];
                firstTrianglePointer++;
                trajectories[v1][i] = quantizer.Restore(quantized);

                quantized = encodedFirstTriangles[firstTrianglePointer];
                firstTrianglePointer++;
                trajectories[v2][i] = quantizer.Restore(quantized);

                quantized = encodedFirstTriangles[firstTrianglePointer];
                firstTrianglePointer++;
                trajectories[v3][i] = quantizer.Restore(quantized);
            }
        }

        /// <summary>
        /// Adds a new vertex predicted by paralellogram rule. Called by the EdgeBreaker traversal engine
        /// </summary>
        /// <param name="sender">Not used</param>
        /// <param name="args">TopologyEventData instance specifying the indices of neighbouring vertices and the index of the new vertex.</param>
        private void OnPredict(object sender, EventArgs args)
        {
            int v1, v2, v3, v4;
            TopologyEventData data = (TopologyEventData)args;
            v2 = data.LeftCornerVertexIndex;
            v3 = data.RightCornerVertexIndex;
            v4 = data.CornerVertexIndex;
            v1 = data.OpositeCornerVertexIndex;
            for (int i = 0; i < baseSize; i++)
            {
                double prediction = trajectories[v2][i] + trajectories[v3][i] - trajectories[v1][i];                
                int quantized = encoded[i][encodedPointer];                
                trajectories[v4][i] = prediction + quantizer.Restore(quantized);
            }
            encodedPointer++;
        }

        double[][] trajectories;
        int[][] encoded;
        int[] encodedFirstTriangles;
        int encodedPointer, firstTrianglePointer;
        int baseSize;
        IQuantizer quantizer;

        UniformDataArray result = null;

        public override void Execute()
        {
            if (IsAllUserSettingsSame() && (result != null))
            {
                SetOutput("Animation", result, true);
            }
            else
            {
                // initialize streams, read global values
                FileStream fs = null;
                try
                {
                    fs = new FileStream(fileName, FileMode.Open);
                    BinaryReader br = new BinaryReader(fs);
                    int pointCount = br.ReadInt32();
                    double delta = br.ReadDouble();
                    quantizer = new UniformQuantizer(delta);

                    // reconstruct the PCA basis
                    PCA pca = new PCA(fs, false);

                    // infer values from PCA basis
                    int frameCount = pca.basis[0].Length / 3;
                    baseSize = pca.basis.Length;

                    // initialize data structures for feature vectors
                    trajectories = new double[pointCount][];
                    for (int i = 0; i < pointCount; i++)
                    {
                        trajectories[i] = new double[baseSize];
                    }
                    encoded = new int[baseSize][];
                    encodedPointer = 0;
                    firstTrianglePointer = 0;

                    // decode quantized residuals
                    encodedFirstTriangles = ArithCoder.decode(fs);
                    for (int i = 0; i < baseSize; i++)
                    {
                        encoded[i] = ArithCoder.decode(fs);
                    }

                    // traverse the mesh
                    EdgebreakerDecompressor decompressor = new EdgebreakerDecompressor();
                    decompressor.InputStream = fs;
                    decompressor.firstTriangleEventHandler += new EventHandler(this.OnNewTriangle);
                    decompressor.newVectorEventHandler += new EventHandler(this.OnPredict);
                    UniformDataArray triangles = decompressor.Run();

                    // reconstruct coordinates from feature vectors
                    for (int i = 0; i < pointCount; i++)
                    {
                        trajectories[i] = pca.Decode(trajectories[i]);
                    }

                    // prepare the data structures for the output
                    result = new UniformDataArray(typeof(TriangleMesh), frameCount);
                    for (int f = 0; f < frameCount; f++)
                    {
                        TriangleMesh frame = new TriangleMesh();
                        frame.Points = new UniformDataArray(typeof(Point3D), pointCount);
                        for (int p = 0; p < pointCount; p++)
                        {
                            frame.Points[p] = new Point3D(trajectories[p][3 * f], trajectories[p][3 * f + 1], trajectories[p][3 * f + 2]);
                        }
                        // add reference to the shared connectivity
                        frame.AddCells("triangles", triangles);
                        result[f] = frame;
                    }

                    SetOutput("Animation", result, false);
                }
                finally
                {
                    fs.Close();
                }
            }
        }
    }
}
