﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.IO;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Dummy compression module, preserves geometry, changes vertex indices in order to be able to compare the result against a compressed mesh.
    /// </summary>
    [ModuleInfo("Libor Váša", "Dummy compression module, preserves geometry, changes vertex indices in order to be able to compare the result against a compressed mesh.", CategoryType.Filter, CategoryData.MeshAnimated)]
    [PortInfo("InAnimation", "input animation")]
    [PortInfo("OutAnimation", "Processed animation, unchanged geometry, updated vertex indices")]
    class DummyCompressor:Module
    {
        public DummyCompressor()
        {
            AddInPort("InAnimation", typeof(UniformDataArray));
            AddOutPort("OutAnimation", typeof(UniformDataArray));
        }

        List<int> indices;
        int currentIndex;
        UniformDataArray result, input;

        public override void Execute()
        {
            if (IsAllDataSame() && IsAllUserSettingsSame() && (result != null))
            {
                SetOutput("OutAnimation", result, true);
            }
            else
            {
                input = (UniformDataArray)GetInput("InAnimation");
                UnstrGrid firstFrame = (UnstrGrid)input[0];
                UniformDataArray triangles = firstFrame.GetCells(typeof(Triangle))[0];
                int pointCount = firstFrame.Points.Count;

                // traverse the topology, save vertex indices
                indices = new List<int>();

                using (MemoryStream topoStream = new MemoryStream())
                {
                    EdgebreakerCompressor compressor = new EdgebreakerCompressor();
                    compressor.OutputStream = topoStream;
                    compressor.firstTriangleEventHandler += new EventHandler(this.newTriangleCompress);
                    compressor.newVectorEventHandler += new EventHandler(this.newVertexCompress);
                    compressor.Run(triangles, pointCount);

                    // prepare the data structures for the result
                    result = new UniformDataArray(typeof(TriangleMesh), input.Count);
                    for (int i = 0; i < input.Count; i++)
                    {
                        TriangleMesh frame = new TriangleMesh();
                        frame.Points = new UniformDataArray(typeof(Point3D), pointCount);
                        result[i] = frame;
                    }

                    // traverse again, copy geometry to new indices
                    topoStream.Position = 0;
                    currentIndex = 0;
                    EdgebreakerDecompressor decompressor = new EdgebreakerDecompressor();
                    decompressor.InputStream = topoStream;
                    decompressor.firstTriangleEventHandler += new EventHandler(this.newTriangleDecompress);
                    decompressor.newVectorEventHandler += new EventHandler(this.newVertexDecompress);
                    UniformDataArray newTriangles = decompressor.Run();


                    // add the new connectivity to each frame
                    for (int f = 0; f < result.Count; f++)
                    {
                        TriangleMesh frame = (TriangleMesh)result[f];
                        frame.AddCells("triangles", newTriangles);
                    }
                }

                SetOutput("OutAnimation", result, false);
            }
        }

        void newTriangleCompress(object sender, EventArgs args)
        {
            int v1, v2, v3;
            TopologyEventData data = (TopologyEventData)args;
            v1 = data.LeftCornerVertexIndex;
            v2 = data.RightCornerVertexIndex;
            v3 = data.CornerVertexIndex;
            indices.Add(v1);
            indices.Add(v2);
            indices.Add(v3);
        }

        void newVertexCompress(object sender, EventArgs args)
        {
            int v4;
            TopologyEventData data = (TopologyEventData)args;            
            v4 = data.CornerVertexIndex;
            indices.Add(v4);
        }

        void newTriangleDecompress(object sender, EventArgs args)
        {
            int v1, v2, v3, ov1, ov2, ov3;
            TopologyEventData data = (TopologyEventData)args;
            v1 = data.LeftCornerVertexIndex;
            v2 = data.RightCornerVertexIndex;
            v3 = data.CornerVertexIndex;
            ov1 = indices[currentIndex++];            
            ov2 = indices[currentIndex++];            
            ov3 = indices[currentIndex++];            
            copyVertex(ov1, v1);
            copyVertex(ov2, v2);
            copyVertex(ov3, v3);
        }

        private void copyVertex(int ov1, int v1)
        {
            for (int i = 0; i < input.Count; i++)
            {
                UnstrGrid oldFrame = (UnstrGrid)input[i];
                UnstrGrid newFrame = (UnstrGrid)result[i];
                newFrame.Points[v1] = oldFrame.Points[ov1];
            }
        }

        void newVertexDecompress(object sender, EventArgs args)
        {
            int v4;
            TopologyEventData data = (TopologyEventData)args;
            v4 = data.CornerVertexIndex;
            int ov4 = indices[currentIndex++];            
            copyVertex(ov4, v4);
        }
    }
}
