﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using Zcu.Mve.Strings;
using System.IO;
using System.ComponentModel;

namespace Zcu.Mve.Compression
{
    public class EdgebreakerCompressor
    {
        #region Global variables
        public EventHandler firstTriangleEventHandler; //event handler for visiting the first triangle of the component
        public EventHandler newVectorEventHandler; //event handler for visiting new vertex
        public EventHandler progressEventHandler; //progress event handler

        private UniformDataArray triangles; // array of all triangles
        private int verticesCount; //number of vertices
        private int numOfComponents; //number of components

        private int[] V; // array of vertices of all triangles -integer reference into "vertices"
        private int[] O; // array of vertices of all triangles -integer reference into "V" na "oposite corners"
        private int[] M; // array of all vertices -">0" if not visited
        private int[] U; // array of all triangles -">0" if not visited
        private LinkedList<int> visitedTriangles; //indices of visited triangles in order they was visited

        private char[] CLERS;  //topology description

        private int lastTriangle; //number of visited triangles
        private int sCounter;     //number of 'S' symbols in CLERS

        private LinkedList<Integer> handles;  //array of handle pairs
        private LinkedList<Integer> holes;    //array of holes description
        private LinkedList<Integer> holeStartingComponents;    //array of component IDs, which are initializing by triangle touching a hole

        private int progressStep; //calculation progress
        private Stream outputStream; //output stream for topology
        private int invisibleTriangles;

        /// <summary>
        /// Stream with compressed topology
        /// </summary>
        public Stream OutputStream
        {
            get
            {
                return outputStream;
            }
            set
            {
                outputStream = value;
            }
        }

        #endregion


        /// <summary>
        /// Class constructor
        /// </summary>
        public EdgebreakerCompressor()
        {
        }

        /// <summary>
        /// Starts compression of topology
        /// </summary>
        /// <param name="inputTris">Input model -list of triangles</param>
        /// /// <param name="points">Number of model vertices</param>
        public void Run(UniformDataArray inputTris, int vertices)
        {
            this.triangles = inputTris;
            this.verticesCount = vertices;
            int startCorner = 0;
            numOfComponents = 0;

            //initialization of compression structures
            init();
            progressStep = CLERS.Length / 7;

            //compression start
            while (true)
            {
                startCompression(startCorner);
                numOfComponents++;

                //if not all vertices visited
                if (lastTriangle < (U.Length - numOfComponents - 1)) //-1 max index of array, -1 first triangle is not counted
                {
                    //looking for unvisited triangle
                    for (int i = 0; i < U.Length; i++)
                    {
                        if (U[i] == 0)
                        {
                            startCorner = i * 3;
                            break;
                        }
                    }
                }
                else break;
            }

            if (outputStream != null && outputStream.CanWrite)
            {
                BinaryWriter bw = new BinaryWriter(outputStream);

                //create ArithCoder
                ArithCoder coder = new ArithCoder(outputStream);

                //store number of components
                bw.Write(numOfComponents);

                //store handles
                bw.Write(handles.Count); //32bit
                foreach (int i in handles) bw.Write(i);

                //store holes
                bw.Write(holes.Count); //32bit
                foreach (int i in holes) bw.Write(i);

                //store holeStarts
                bw.Write(holeStartingComponents.Count); //32bit
                foreach (int i in holeStartingComponents) bw.Write(i);

                //array for CLERS string
                int[] bitCLERS = new int[lastTriangle];

                for (int i = 0; i < bitCLERS.Length; i++)
                {
                    switch (CLERS[i])
                    {
                        case 'C': bitCLERS[i] = 0;
                            break;
                        case 'L': bitCLERS[i] = 1;
                            break;
                        case 'E': bitCLERS[i] = 2;
                            break;
                        case 'R': bitCLERS[i] = 3;
                            break;
                        case 'S': bitCLERS[i] = 4;
                            break;
                        default:
                            Console.WriteLine("Error! Unknown CLERS symbol");
                            break;
                    }
                }

                //store CLERS
                coder.Encode(bitCLERS);
            }
        }


        /// <summary>
        /// Initialization of arrays and variables
        /// for compression of topology
        /// </summary>
        private void init()
        {
            //////////////////////
            // PREPARE TOPOLOGY //
            //////////////////////
            #region Initialization of topology
            V = new int[3 * triangles.Count];
            O = new int[3 * triangles.Count];
            M = new int[verticesCount];
            U = new int[triangles.Count];
            visitedTriangles = new LinkedList<int>();


            //-1 if oposite corner not exists
            for (int i = 0; i < O.Length; i++)
                O[i] = -1;
            //set to 0 = not visited
            for (int i = 0; i < M.Length; i++)
                M[i] = 0;
            for (int i = 0; i < U.Length; i++)
                U[i] = 0;

            //array for CLERS
            CLERS = new char[triangles.Count - 1];

            #region Inicializace hodnot v V a O tabulce
            //fill array V and O
            int numOfTriangles = 0;
            Dictionary<Line, int> hashOposite = new Dictionary<Line, int>(); //store vertex and oposite edge of the triangle

            foreach (Triangle t in triangles)
            {
                //store vertices of triangle -counterClockwise
                V[3 * numOfTriangles] = t.V1;
                V[3 * numOfTriangles + 1] = t.V2;
                V[3 * numOfTriangles + 2] = t.V3;

                //hash vertex of triangle and oposite edge
                //"l_" counterCloctwise
                Line l3 = new Line(t.V1, t.V2, true);
                Line l2 = new Line(t.V3, t.V1, true);
                Line l1 = new Line(t.V2, t.V3, true);

                if (hashOposite.ContainsKey(l1))
                    Console.WriteLine(hashOposite[l1]);

                hashOposite.Add(l1, 3 * numOfTriangles);
                hashOposite.Add(l2, 3 * numOfTriangles + 1);
                hashOposite.Add(l3, 3 * numOfTriangles + 2);

                numOfTriangles++;
            }


            //find all "oposite" vertices from O for vetices V
            for (int i = 0; i < numOfTriangles; i++)
            {
                Line l3 = new Line(V[3 * i + 1], V[3 * i], true);
                Line l2 = new Line(V[3 * i], V[3 * i + 2], true);
                Line l1 = new Line(V[3 * i + 2], V[3 * i + 1], true);

                if (hashOposite.ContainsKey(l1))
                {
                    O[3 * i] = hashOposite[l1];
                }

                if (hashOposite.ContainsKey(l2))
                {
                    O[3 * i + 1] = hashOposite[l2];
                }

                if (hashOposite.ContainsKey(l3))
                {
                    O[3 * i + 2] = hashOposite[l3];
                }
            }
            #endregion

            //if corner "i" don't has oposite corner,
            //corners next(i) and prev(i) are on the border of the hole
            for (int i = 0; i < O.Length; i++)
            {
                if (O[i] == -1)
                {
                    M[V[next(i)]] = -1;
                    M[V[prev(i)]] = -1;
                }
            }

            lastTriangle = 0;
            invisibleTriangles = 0;
            sCounter = 0;
            handles = new LinkedList<Integer>();
            holes = new LinkedList<Integer>();
            holeStartingComponents = new LinkedList<Integer>();
            #endregion
        }

        /// <summary>
        /// Start compression from corner "c"
        /// </summary>
        /// <param name="c">Start corner</param>
        private void startCompression(int c)
        {
            //find "corner" which has "oposite corner"
            int a = c;
            while (O[a] == -1)
            {
                a = next(a);
                if (a == c)
                {
                    #region Typ -9 lonely triangle
                    holeStartingComponents.AddLast(numOfComponents);
                    initFirstTriangle(a);
                    M[V[a]] = a + 1;
                    M[V[next(a)]] = next(a) + 1;
                    M[V[prev(a)]] = prev(a) + 1;
                    holes.AddLast(-9);
                    holes.AddLast(3);
                    #endregion
                    return;
                }
            }

            #region Holes
            //types of "first triangles"
            //a) triangle touches hole by 0 vertices
            //b) triangle touches hole by one vertex "-1"
            //c) triangle touches hole by 2 vertices and one edge "-2"
            //d) triangle touches hole by three vertices and two edges "-3"
            //e) triangle touches hole by two vertices and no edge "-4"
            //f) triangle touches hole by three vertices and one edge "-5"
            //g) triangle touches hole by three vertices and no edge "-6"
            //g) triangle touches hole by three vertices and two edges "-7"
            //h) triangle touches hole by three vertices and three edges => new komponent
            int touching1 = 0;
            int touching2 = 0;
            int touching3 = 0;
            if (M[V[a]] < 0 || O[next(a)] == -1 || O[prev(a)] == -1) touching1++;
            if (M[V[next(a)]] < 0 || O[a] == -1 || O[prev(a)] == -1) touching2++;
            if (M[V[prev(a)]] < 0 || O[next(a)] == -1 || O[a] == -1) touching3++;
            int lastHoleLength;

            switch (touching1 + touching2 + touching3)
            {
                case 1://hole touches 1 corner
                    #region Typ 1
                    while (M[V[a]] != -1)
                    {
                        a = next(a);
                    }

                    holeStartingComponents.AddLast(numOfComponents);

                    initFirstTriangle(a);
                    M[V[a]] = a + 1;
                    M[V[next(a)]] = next(a) + 1;
                    M[V[prev(a)]] = prev(a) + 1;

                    markHole(a);
                    lastHoleLength = holes.Last.Value;
                    holes.RemoveLast();
                    holes.RemoveLast();
                    holes.AddLast(-1);
                    holes.AddLast(lastHoleLength);
                    #endregion
                    break;

                case 2:// hole touches 2 corners
                    if (touching1 == 0) a = prev(a);
                    if (touching3 == 0) a = next(a);

                    initFirstTriangle(a);
                    markHole(a);

                    lastHoleLength = holes.Last.Value;
                    holes.RemoveLast();
                    holes.RemoveLast();

                    if (O[next(a)] != -1) //dva vrcholy se sice hole dotykaji, ale hrana mezi nimi ne
                    {

                        if (M[V[prev(a)]] == -1)//pokud jsou hole dve a ne jedna
                        {
                            #region Typ 7

                            holes.AddLast(-7);
                            holes.AddLast(lastHoleLength);

                            markHole(prev(a));

                            lastHoleLength = holes.Last.Value;
                            holes.RemoveLast();
                            holes.RemoveLast();
                            holes.AddLast(-7);
                            holes.AddLast(lastHoleLength);
                            M[V[a]] = a + 1;
                            M[V[next(a)]] = next(a) + 1;
                            M[V[prev(a)]] = prev(a) + 1;
                            #endregion
                            holeStartingComponents.AddLast(numOfComponents);
                        }
                        else//pokud je hole jedna, ale dotykaji se ji jen vrcholy a ne hrana
                        {
                            #region Typ 4
                            holes.AddLast(-4);
                            holes.AddLast(lastHoleLength);
                            //oznacit 3 vrcholy za prozkoumane
                            M[V[a]] = a + 1;
                            M[V[next(a)]] = next(a) + 1;
                            M[V[prev(a)]] = prev(a) + 1;
                            compress(O[next(a)]);
                            #endregion
                            holeStartingComponents.AddLast(numOfComponents);
                        }
                    }
                    else //pokud je hole jedna a dotykaji se ji i vrcholy i hrana
                    {
                        #region Typ 2
                        M[V[a]] = a + 1;
                        M[V[next(a)]] = next(a) + 1;
                        M[V[prev(a)]] = prev(a) + 1;

                        holes.AddLast(-2);
                        holes.AddLast(lastHoleLength);
                        #endregion
                        holeStartingComponents.AddLast(numOfComponents);
                    }
                    break;

                case 3:// vsechny tri se dotykaji hole - mozna zvlast komponenta
                    //pokud se dotykaji 3 hrany ->nova komponenta
                    //reseno vyse

                    //dotykaji se dve hrany
                    if ((O[next(a)] == -1) && (O[prev(a)] == -1))
                    {
                        #region Typ 3
                        initFirstTriangle(a);
                        holeStartingComponents.AddLast(numOfComponents);

                        markHole(next(a));
                        lastHoleLength = holes.Last.Value;
                        holes.RemoveLast();
                        holes.RemoveLast();
                        holes.AddLast(-3);
                        holes.AddLast(lastHoleLength);

                        M[V[a]] = a + 1;
                        M[V[next(a)]] = next(a) + 1;
                        M[V[prev(a)]] = prev(a) + 1;
                        break;
                        #endregion
                    }

                    //nedotyka se zadna hrana
                    if ((O[prev(a)] != -1) && (O[next(a)] != -1))
                    {
                        #region Typ 6
                        initFirstTriangle(a);
                        M[V[a]] = a + 1;
                        M[V[next(a)]] = next(a) + 1;
                        M[V[prev(a)]] = prev(a) + 1;

                        holeStartingComponents.AddLast(numOfComponents);

                        markHole(next(a));
                        compress(O[prev(a)]);
                        compress(O[next(a)]);
                        lastHoleLength = holes.Last.Value;
                        holes.RemoveLast();
                        holes.RemoveLast();
                        holes.AddLast(-6);
                        holes.AddLast(lastHoleLength);
                        break;
                        #endregion
                    }
                    else //dotyka jen jedna hrana
                        if ((O[next(a)] != -1) || (O[prev(a)] != -1))
                        {
                            if (O[prev(a)] != -1) a = prev(a);

                            initFirstTriangle(a);

                            markHole(next(a));
                            lastHoleLength = holes.Last.Value;
                            holes.RemoveLast();
                            holes.RemoveLast();

                            if (M[V[prev(a)]] == -1) //dve hole
                            {
                                #region Typ 8
                                holes.AddLast(-8);
                                holes.AddLast(lastHoleLength);

                                markHole(prev(a));

                                lastHoleLength = holes.Last.Value;
                                holes.RemoveLast();
                                holes.RemoveLast();
                                holes.AddLast(-8);
                                holes.AddLast(lastHoleLength);

                                M[V[a]] = a + 1;
                                M[V[next(a)]] = next(a) + 1;
                                M[V[prev(a)]] = prev(a) + 1;
                                #endregion
                                holeStartingComponents.AddLast(numOfComponents);
                            }
                            else //jedna hole
                            {
                                #region Typ 5
                                M[V[a]] = a + 1;
                                M[V[next(a)]] = next(a) + 1;
                                M[V[prev(a)]] = prev(a) + 1;

                                compress(O[next(a)]);

                                holes.AddLast(-5);
                                holes.AddLast(lastHoleLength);
                                #endregion
                                holeStartingComponents.AddLast(numOfComponents);
                            }
                            break;
                        }

                    break;

                default:
                    initFirstTriangle(a);
                    M[V[a]] = a + 1;
                    M[V[next(a)]] = next(a) + 1;
                    M[V[prev(a)]] = prev(a) + 1;
                    break;
            }
            #endregion

            //compress "oposite" vertex if not yet compressed
            if (U[tri(O[a])] == 0)
            {
                compress(O[a]);
            }
            else
            {
                Console.WriteLine("can't compress ... U[" + tri(O[a]) + "]");
            }
        }

        /// <summary>
        /// Inicializes first triangle of component
        /// </summary>
        /// <param name="firstTriangleCorner">Number of the first triangle corner</param>
        private void initFirstTriangle(int firstTriangleCorner)
        {
            //mark triangle as visited
            U[tri(firstTriangleCorner)] = 1;
            visitedTriangles.AddLast(tri(firstTriangleCorner));

            //event "Firts triangle of component"
            if (firstTriangleEventHandler != null) //if event handler is set
            {
                TopologyEventData data = new TopologyEventData();
                data.CornerVertexIndex = V[firstTriangleCorner];
                data.LeftCornerVertexIndex = V[next(firstTriangleCorner)];
                data.RightCornerVertexIndex = V[prev(firstTriangleCorner)];
                firstTriangleEventHandler(this, data);
            }
        }

        /// <summary>
        /// Geometric compression of new vertex
        /// </summary>
        /// <param name="c">Vertex corner</param>
        private void compressNewVector(int c)
        {
            //event new vector
            if (newVectorEventHandler != null)
            {
                TopologyEventData data = new TopologyEventData();
                data.CornerVertexIndex = V[c];
                data.OpositeCornerVertexIndex = V[O[c]];
                data.LeftCornerVertexIndex = V[prev(O[c])];
                data.RightCornerVertexIndex = V[next(O[c])];
                newVectorEventHandler(this, data);
            }
        }

        /// <summary>
        /// Traverse topology and compress it
        /// </summary>
        /// <param name="c">Start corner</param>
        private void compress(int c)
        {
            while (true)
            {
                #region Progress
                if (lastTriangle % progressStep == 0)
                    if (progressEventHandler != null)
                    {
                        ProgressEventData data = new ProgressEventData();
                        data.Progress = 0.01f;
                        progressEventHandler(this, data);
                    }
                #endregion

                U[tri(c)] = 1; //mark triangle as visited
                visitedTriangles.AddLast(tri(c));
                lastTriangle++;

                #region Check handles
                
                if (right(c) != -1)//-1 = border edge of some hole
                {
                    if (U[tri(right(c))] > 1) //if right triangle type is "S -handle"
                    {
                        //store into "handles" actual vertice and it's "oposite corner" stored in "U" 
                        //must add 3*numOfComponents, because Decompressor wil
                        //count "+ 1" triangle for each new component
                        handles.AddLast(U[tri(right(c))] + 3 * numOfComponents);
                        handles.AddLast((lastTriangle + numOfComponents + invisibleTriangles) * 3 + 1); //for decompresor it will be the same as next(c)
                    }
                    else if (left(c) != -1)//-1 = border edge of some hole
                        if (U[tri(left(c))] > 1) //if right triangle type is "S -handle"
                        {
                            //store into "handles" actual vertice and it's "oposite corner" stored in "U" 
                            //must add 3*numOfComponents, because Decompressor wil
                            //count "+ 1" triangle for each new component
                            handles.AddLast(U[tri(left(c))] + 3 * numOfComponents);
                            handles.AddLast((lastTriangle + numOfComponents + invisibleTriangles) * 3 + 2); //for decompresor it will be the same as next(c)
                        }
                }
                #endregion

                //store vertex
                if (M[V[c]] <= 0) compressNewVector(c);

                //if vertex unvisited -> new triangle
                if (M[V[c]] == 0)
                {
                    //store C into CLERS
                    CLERS[lastTriangle - 1] = 'C';
                    M[V[c]] = c + 1;
                    c = right(c);
                }
                else if (M[V[c]] == -1)
                {
                    //if vertex is part of unvisited hole
                    //mark all vertices as visited
                    //and store info into list of holes
                    sCounter++;
                    //store S into CLERS
                    CLERS[lastTriangle - 1] = 'S';
                    M[V[c]] = c + 1;
                    markHole(c);
                    c = right(c);
                }
                else //vertex visited
                {
                    //can't traverse right
                    if ((right(c) == -1) || //hole
                        ((right(c) != -1) && (U[tri(right(c))] > 0))) //visited triangle
                    {
                        //can't traverse left
                        if ((left(c) == -1) || //hole
                            ((left(c) != -1) && (U[tri(left(c))] > 0))) //visited triangle
                        {
                            CLERS[lastTriangle - 1] = 'E';
                            return; //can't go anywhere
                        }
                        else //right triangle visited, left not visited
                        {
                            CLERS[lastTriangle - 1] = 'R';
                            c = left(c);
                        }
                    }
                    else
                    {
                        //can't traverse left
                        if ((left(c) == -1) || //hole
                            ((left(c) != -1) && (U[tri(left(c))] > 0))) //left triangle visited, right not visited
                        {
                            CLERS[lastTriangle - 1] = 'L';
                            c = right(c);
                        }
                        else //left and right not visited
                        {
                            //find out if marking corner is reachable
                            if (checkAvailability(c) == true)
                            {
                                sCounter++; // store number of S in CLERS
                                U[tri(c)] = (lastTriangle + invisibleTriangles) * 3 + 2; //for decompressor the same as prev(c)
                                CLERS[lastTriangle - 1] = 'S';
                                compress(right(c));
                                c = left(c);

                                if (U[tri(c)] > 0) break;
                            }
                            else //if not, store C instead of S
                            {
                                compressNewVector(c);
                                CLERS[lastTriangle - 1] = 'C';
                                c = right(c);
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Finds hole and marks it as visited
        /// </summary>
        /// <param name="c">Start corner</param>
        private void markHole(int c)
        {
            int length = 0; // number of vertices on the border of the hole
            int c1, c2;
            int startCorner = c;


            //for all hole border vertices
            c1 = c;
            c2 = c;
            c = next(c);
            while (true)
            {
                while (O[c] != -1)
                {
                    c = prev(O[c]);//jump over the next edge

                    //hole traversed
                    if (prev(c) == startCorner)
                    {
                        //store info about hole
                        holes.AddLast(sCounter); //starting S szmbol of the hole
                        holes.AddLast(length);  //length of the hole
                        invisibleTriangles += length + 2;
                        return;
                    }
                }

                c = prev(c);

                M[V[c]] = -2;
                length++;

                c2 = c1;
                c1 = prev(c);
            }
        }

        /// <summary>
        /// Finds out if is possible to reach common vertex from other corner
        /// </summary>
        /// <param name="c">Tested corner</param>
        /// <returns>True, if vertex reachable</returns>
        private bool checkAvailability(int c)
        {
            int a = next(c);
            a = prev(O[a]);
            while (prev(a) != c)
            {
                //if hole reached, try it oposite direction
                if (O[a] == -1)
                {
                    return true;
                }
                else //else continue original direction
                {
                    a = prev(O[a]);
                    if (prev(a) == (M[V[c]] - 1)) return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Return next triangle corner
        /// </summary>
        /// <param name="c">Start corner</param>
        /// <returns>Next corner</returns>
        private int next(int c)
        {
            if (c % 3 == 2)
            {
                return (c - 2);
            }
            else
            {
                return (c + 1);
            }
        }

        /// <summary>
        /// Vrati predchozi corner trojuhelniku
        /// </summary>
        /// <param name="c">Startovaci corner</param>
        /// <returns>Predchozi corner</returns>
        private int prev(int c)
        {
            if (c % 3 == 0)
            {
                return (c + 2);
            }
            else
            {
                return (c - 1);
            }
        }

        /// <summary>
        /// Returns triangle of given corner
        /// </summary>
        /// <param name="c">Corner</param>
        /// <returns>Triangle</returns>
        private int tri(int c)
        {
            return c / 3;
        }

        /// <summary>
        /// Returns corner of the vertex on the left
        /// </summary>
        /// <param name="c">Given corner</param>
        /// <returns>Left corner</returns>
        private int left(int c)
        {
            return O[prev(c)];
        }

        /// <summary>
        /// Returns corner of the vertex on the right
        /// </summary>
        /// <param name="c">Given corner</param>
        /// <returns>Right corner</returns>
        private int right(int c)
        {
            return O[next(c)];
        }

        /// <summary>
        /// Returns number of components of the model
        /// </summary>
        /// <returns>Number of components</returns>
        public int getNumOfComponents()
        {
            return numOfComponents;
        }

        /// <summary>
        /// Returns CLERS description of topology
        /// </summary>
        /// <returns>CLERS</returns>
        public string getCLERS()
        {
            return new string(CLERS, 0, lastTriangle);
        }

        /// <summary>
        /// Returns list of handles
        /// </summary>
        /// <returns>List of handles</returns>
        public UniformDataArray getHandles()
        {
            UniformDataArray handlesArray = new UniformDataArray(typeof(Integer),
                                                                  handles.Count);
            int k = 0;
            foreach (Integer h in handles)
            {
                handlesArray[k] = h;
                k++;
            }

            return handlesArray;
        }

        /// <summary>
        /// Returns list of holes
        /// </summary>
        /// <returns>List of holes</returns>
        public UniformDataArray getHoles()
        {
            UniformDataArray holesArray = new UniformDataArray(typeof(Integer),
                                                                  holes.Count);
            int k = 0;
            foreach (Integer h in holes)
            {
                holesArray[k] = h;
                k++;
            }

            return holesArray;
        }

        /// <summary>
        /// Returns list of components starting by hole
        /// </summary>
        /// <returns>List of hole-starting components</returns>
        public UniformDataArray getHoleStarts()
        {
            UniformDataArray startsArray = new UniformDataArray(typeof(Integer),
                                                                  holeStartingComponents.Count);
            int k = 0;
            foreach (Integer s in holeStartingComponents)
            {
                startsArray[k] = s;
                k++;
            }

            return startsArray;
        }

        /// <summary>
        /// Returns list of triangles in order they were visited
        /// </summary>
        /// <returns>Visited triangles</returns>
        public int[] getVisitedTriangles()
        {
            int[] viTriangles = new int[visitedTriangles.Count];

            for (int i = 0; i < viTriangles.Length; i++)
            {
                viTriangles[i] = visitedTriangles.First.Value;
                visitedTriangles.RemoveFirst();
            }

            return viTriangles;
        }

        /// <summary>
        /// Returns Edgebreaker O-table
        /// </summary>
        /// <returns>O-table</returns>
        public int[] getO()
        {
            return O;
        }

        /// <summary>
        /// Returns Edgebreaker V-table
        /// </summary>
        /// <returns>V-table</returns>
        public int[] getV()
        {
            return V;
        }
    }
}
