﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using Zcu.Mve.Strings;
using System.IO;
using System.ComponentModel;
using Zcu.Mve.Compression;

namespace Zcu.Mve.Compression
{
    public class EdgebreakerDecompressor
    {
        #region Global variables
        private EdgebreakerTopologyDecompressor topDecompr; //topology decompressor using Edgebreaker
        private EdgebreakerCompressor topCompr;             //topology compressor using Edgebreaker 
        public EventHandler firstTriangleEventHandler;      //event handler for visiting the first triangle of component
        public EventHandler newVectorEventHandler;          //event handler for visiting new vertex
        public EventHandler progressEventHandler;           //progress event handler
        private Stream inputStream;                         //possible direct input of topology data from stream

        /// <summary>
        /// Stream with input data
        /// </summary>
        public Stream InputStream
        {
            get
            {
                return inputStream;
            }
            set
            {
                inputStream = value;
            }
        }

        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        public EdgebreakerDecompressor()
        {
            topDecompr = new EdgebreakerTopologyDecompressor();
            topCompr = new EdgebreakerCompressor();
        }

        /// <summary>
        /// Starting method, inputStream have to be set before running this
        /// </summary>
        public UniformDataArray Run()
        {
            topDecompr.InputStream = this.inputStream;
            topCompr.progressEventHandler = this.progressEventHandler;
            topCompr.newVectorEventHandler = this.newVectorEventHandler;
            topCompr.firstTriangleEventHandler = this.firstTriangleEventHandler;

            if (this.inputStream == null || !this.inputStream.CanRead)
            {
                throw new Exception("Can't read data! Input stream missing!");
                return null;
            }
            else
            {
                //run topology decompression
                topDecompr.Run();

                // now we have complete TOPOLOGY information

                #region Get number of vertices and V-table (triangles)
                int[] v = topDecompr.getV(); //v-table

                LinkedList<Triangle> triangles = new LinkedList<Triangle>(); //list of all triangles
                List<int> invisibles = new List<int>(topDecompr.getInvisibleTriangles()); //list of invisible triangles

                //create list of visible triangles
                int bigestIndex = 0;
                for (int i = 0; i < v.Length / 3; i++)
                {
                    if (!invisibles.Contains(i))
                    {
                        Triangle t = new Triangle(v[3 * i], v[3 * i + 1], v[3 * i + 2]);
                        triangles.AddLast(t);

                        if (v[3 * i] > bigestIndex)
                        {
                            bigestIndex = v[3 * i];
                        }

                        if (v[3 * i + 1] > bigestIndex)
                        {
                            bigestIndex = v[3 * i + 1];
                        }

                        if (v[3 * i + 2] > bigestIndex)
                        {
                            bigestIndex = v[3 * i + 2];
                        }
                    }
                }

                int numOfCorners = bigestIndex + 1;

                List<int> indices = new List<int>();

                foreach (Triangle t in triangles)
                {
                    if (!indices.Contains(t.V1)) indices.Add(t.V1);
                    if (!indices.Contains(t.V2)) indices.Add(t.V2);
                    if (!indices.Contains(t.V3)) indices.Add(t.V3);
                }

                List<Triangle> reindexedTriangles = new List<Triangle>();

                foreach (Triangle t in triangles)
                {
                    Triangle newTri = new Triangle();

                    newTri.V1 = indices.LastIndexOf(t.V1);
                    newTri.V2 = indices.LastIndexOf(t.V2);
                    newTri.V3 = indices.LastIndexOf(t.V3);

                    reindexedTriangles.Add(newTri);
                }

                //convert list to UniformDataArray
                UniformDataArray tmpTriangles = new UniformDataArray(typeof(Triangle), triangles.Count);

                for (int i = 0; i < triangles.Count; i++)
                {
                    tmpTriangles[i] = reindexedTriangles[i];
                }

                #endregion



                //start "compression" of topology -traverses mesh the same way as in EdgebreakerCompressor
                topCompr.Run(tmpTriangles, numOfCorners);


                return tmpTriangles;
            }

        }

        /// <summary>
        /// Starting method, inputStream don't have to be set
        /// </summary>
        /// <param name="clers"></param>
        /// <param name="numOfComponents"></param>
        /// <param name="holes"></param>
        /// <param name="holeStarts"></param>
        /// <param name="handles"></param>
        /// <returns></returns>
        public UniformDataArray Run(char[] clers, int numOfComponents, int[] holes, int[] holeStarts, int[] handles)
        {

            topDecompr.InputStream = this.inputStream;
            topCompr.progressEventHandler = this.progressEventHandler;
            topCompr.newVectorEventHandler = this.newVectorEventHandler;
            topCompr.firstTriangleEventHandler = this.firstTriangleEventHandler;

            //run topology decompression
            topDecompr.Run();

            // now we have complete TOPOLOGY information

            #region Get number of vertices and V-table (triangles)
            int[] v = topDecompr.getV(); //v-table

            LinkedList<Triangle> triangles = new LinkedList<Triangle>(); //list of all triangles
            List<int> invisibles = new List<int>(topDecompr.getInvisibleTriangles()); //list of invisible triangles

            //create list of visible triangles
            int bigestIndex = 0;
            for (int i = 0; i < v.Length / 3; i++)
            {
                if (!invisibles.Contains(i))
                {
                    Triangle t = new Triangle(v[3 * i], v[3 * i + 1], v[3 * i + 2]);
                    triangles.AddLast(t);

                    if (v[3 * i] > bigestIndex)
                    {
                        bigestIndex = v[3 * i];
                    }

                    if (v[3 * i + 1] > bigestIndex)
                    {
                        bigestIndex = v[3 * i + 1];
                    }

                    if (v[3 * i + 2] > bigestIndex)
                    {
                        bigestIndex = v[3 * i + 2];
                    }
                }
            }

            int numOfCorners = bigestIndex + 1;

            List<int> indices = new List<int>();

            foreach (Triangle t in triangles)
            {
                if (!indices.Contains(t.V1)) indices.Add(t.V1);
                if (!indices.Contains(t.V2)) indices.Add(t.V2);
                if (!indices.Contains(t.V3)) indices.Add(t.V3);
            }

            List<Triangle> reindexedTriangles = new List<Triangle>();

            foreach (Triangle t in triangles)
            {
                Triangle newTri = new Triangle();

                newTri.V1 = indices.LastIndexOf(t.V1);
                newTri.V2 = indices.LastIndexOf(t.V2);
                newTri.V3 = indices.LastIndexOf(t.V3);

                reindexedTriangles.Add(newTri);
            }

            //convert list to UniformDataArray
            UniformDataArray tmpTriangles = new UniformDataArray(typeof(Triangle), triangles.Count);

            for (int i = 0; i < triangles.Count; i++)
            {
                tmpTriangles[i] = reindexedTriangles[i];
            }

            #endregion



            //start "compression" of topology -traverses mesh the same way as in EdgebreakerCompressor
            topCompr.Run(tmpTriangles, numOfCorners);


            return tmpTriangles;
        }

        /// <summary>
        /// Returns array of invisible triangles
        /// </summary>
        /// <returns>Array of invisible trangles</returns>
        public int[] getInvisibleTriangles()
        {
            if (topDecompr != null) return topDecompr.getInvisibleTriangles();
            else return null;
        }

        /// <summary>
        /// Returns Edgebreaker V-table
        /// </summary>
        /// <returns>V-table</returns>
        public int[] getV()
        {
            return topCompr.getV();
        }
    }

    public class EdgebreakerTopologyDecompressor
    {
        #region Global variables

        private char[] CLERS; //description of topology 
        private int[] holes;  //list of component's holes 
        private int[] handles; //list of component's handles
        private int[] holeStarts; //list of components starting by hole
        private int[] invisibleTriangles; //list of invisible triangles -used just to create correct CornerTable

        private int[] V;     //array of vertices of all triangles -integer reference into "G"
        private int[] O;     //array of oposite vertices of all triangle corner vertices -integer reference into "V"
        private List<int> escapeCorners;      //corners causing cyclus break

        private int numOfComponents; //number of model components

        private int lastTriangle; //last decompressed triangle
        private int lastVertex;   //last visited vertex
        private int lastHandle;   //last visited handle
        private int lastCLERS;    //last symbol red from clers
        private int lastHole;     //last visited hole
        private int lastHoleStart;//last visited hole, starting some component
        private int sCounter;     //counter of 'S' symbols
        private int invisibleTriCounter; //number of invisible triangles

        //need for zip in "-4" hole
        private int lastCorner1 = 0;
        private int lastCorner2 = 0;

        List<int> usedHoles = new List<int>();

        private Stream inputStream; //possible direct input of topologz data from stream

        /// <summary>
        /// Stream with input data
        /// </summary>
        public Stream InputStream
        {
            get
            {
                return inputStream;
            }
            set
            {
                inputStream = value;
            }
        }

        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        public EdgebreakerTopologyDecompressor()
        {
        }

        /// <summary>
        /// Start topology decompression
        /// </summary>
        /// <param name="clers">CLERS</param>
        /// <param name="holes">Array of holes</param>
        /// <param name="hStarts">Array of components starting by hole</param>
        /// <param name="handles">Array of handles</param>
        public void Run(char[] clers, int numOfComps, int[] holes, int[] hStarts, int[] handles)
        {
            if (inputStream != null && inputStream.CanRead == true)
            {
                BinaryReader br = new BinaryReader(inputStream);

                //create ArithCoder
                ArithCoder coder = new ArithCoder(inputStream);

                //load number of components
                this.numOfComponents = br.ReadInt32();

                //load handles
                int handlesLength = (int)br.ReadUInt32(); //32bits
                this.handles = new int[handlesLength];

                for (int i = 0; i < handlesLength; i++)
                {
                    this.handles[i] = br.ReadInt32();
                }

                //load holes
                int holesLength = (int)br.ReadUInt32(); //32bits
                this.holes = new int[holesLength];

                for (int i = 0; i < holesLength; i++)
                {
                    this.holes[i] = br.ReadInt32();
                }

                //load holeStarts
                int holeStartsLength = (int)br.ReadUInt32(); //32bits
                this.holeStarts = new int[holeStartsLength];

                for (int i = 0; i < holeStartsLength; i++)
                {
                    this.holeStarts[i] = br.ReadInt32();
                }

                //decode clers
                int[] bitCLERS = coder.Decode();
                this.CLERS = new char[bitCLERS.Length];

                //convert to string
                for (int i = 0; i < bitCLERS.Length; i++)
                {
                    switch (bitCLERS[i])
                    {
                        case 0: this.CLERS[i] = 'C';
                            break;
                        case 1: this.CLERS[i] = 'L';
                            break;
                        case 2: this.CLERS[i] = 'E';
                            break;
                        case 3: this.CLERS[i] = 'R';
                            break;
                        case 4: this.CLERS[i] = 'S';
                            break;
                        default:
                            Console.WriteLine("Error! Unknown CLERS symbol");
                            break;
                    }
                }
            }
            else
            {
                this.CLERS = clers;
                this.numOfComponents = numOfComps;
                this.holes = holes;
                this.holeStarts = hStarts;
                this.handles = handles;
            }

            //create and initialize structures for decompression
            initTopology();
            numOfComponents = 0;
            Console.WriteLine("Hole types:");
            while (true)
            {
                //dekompress loaded data
                startTopologyDecompression();
                numOfComponents++;
                if (lastCLERS >= CLERS.Length) break;
            }
        }

        /// <summary>
        /// Start topology decompression, need inputStream to be set
        /// </summary>
        public void Run()
        {
            if (inputStream != null && inputStream.CanRead == true)
            {
                BinaryReader br = new BinaryReader(inputStream);

                //create ArithCoder
                ArithCoder coder = new ArithCoder(inputStream);

                //load number of components
                this.numOfComponents = br.ReadInt32();

                //load handles
                int handlesLength = (int)br.ReadUInt32(); //32bits
                this.handles = new int[handlesLength];

                for (int i = 0; i < handlesLength; i++)
                {
                    this.handles[i] = br.ReadInt32();
                }

                //load holes
                int holesLength = (int)br.ReadUInt32(); //32bits
                this.holes = new int[holesLength];

                for (int i = 0; i < holesLength; i++)
                {
                    this.holes[i] = br.ReadInt32();
                }

                //load holeStarts
                int holeStartsLength = (int)br.ReadUInt32(); //32bits
                this.holeStarts = new int[holeStartsLength];

                for (int i = 0; i < holeStartsLength; i++)
                {
                    this.holeStarts[i] = br.ReadInt32();
                }

                //decode clers
                int[] bitCLERS = coder.Decode();
                this.CLERS = new char[bitCLERS.Length];

                //convert to string
                for (int i = 0; i < bitCLERS.Length; i++)
                {
                    switch (bitCLERS[i])
                    {
                        case 0: this.CLERS[i] = 'C';
                            break;
                        case 1: this.CLERS[i] = 'L';
                            break;
                        case 2: this.CLERS[i] = 'E';
                            break;
                        case 3: this.CLERS[i] = 'R';
                            break;
                        case 4: this.CLERS[i] = 'S';
                            break;
                        default:
                            Console.WriteLine("Error! Unknown CLERS symbol");
                            break;
                    }
                }
            }
            else
            {
                throw new Exception("Can't read data! Input stream missing!");
                return;
            }

            //create and initialize structures for decompression
            initTopology();
            numOfComponents = 0;
            Console.WriteLine("Hole types:");
            while (true)
            {
                //dekompress loaded data
                startTopologyDecompression();
                numOfComponents++;
                if (lastCLERS >= CLERS.Length) break;
            }
        }

        /// <summary>
        /// Initialize structures for decompression
        /// </summary>
        private void initTopology()
        {
            //number of triangles
            int numOfTriangles = CLERS.Length + numOfComponents; //number of triangles without triangles creating holes
            for (int i = 0; i < (holes.Length / 2); i++)
            {
                if (holes[2 * i] != -9)
                    numOfTriangles += holes[2 * i + 1] + 2;
            }

            invisibleTriangles = new int[numOfTriangles - (CLERS.Length + numOfComponents)];


            //create V and O tables
            V = new int[3 * numOfTriangles];
            O = new int[3 * numOfTriangles];

            //init V and O
            for (int i = 0; i < V.Length; i++)
            {
                V[i] = -1; //possible values ">=0"
                O[i] = -3; //possible values ">= -2"
            }

            lastTriangle = -1;
            lastVertex = -1;
            lastHandle = 0;
            lastHole = 0;
            lastHoleStart = 0;
            lastCLERS = 0;
            sCounter = 0;
            invisibleTriCounter = 0;

            escapeCorners = new List<int>();
        }

        /// <summary>
        /// Create the first component triangle and start decompression of component
        /// </summary>
        private void startTopologyDecompression()
        {
            //1. is not stored in CLERS
            lastTriangle++;
            int startCorner = lastTriangle * 3;

            lastVertex++;
            V[startCorner] = lastVertex;

            lastVertex++;
            V[startCorner + 1] = lastVertex;

            lastVertex++;
            V[startCorner + 2] = lastVertex;

            //edges of triangle oposite 2. and 3. vertex
            //(prepare for "zip" => "-1")
            O[startCorner + 1] = -1;
            O[startCorner + 2] = -1;

            #region Holes
            //if some holes exists and component starts by it
            if ((holeStarts.Length != 0) && (lastHoleStart < holeStarts.Length) && (holeStarts[lastHoleStart] == numOfComponents))
            {
                lastHoleStart++;
                if ((holes.Length != 0) && (2 * lastHole < holes.Length))
                {
                    switch (holes[2 * lastHole])
                    {
                        case -1://1 vertex
                            createHoleTopology(startCorner + 1, holes[2 * lastHole + 1]);
                            lastHole++;
                            //start decompression
                            decompressTopology(startCorner);

                            if (!usedHoles.Contains(-1))
                            {
                                usedHoles.Add(-1);
                                Console.WriteLine("-1");
                            }

                            break;

                        case -2://2 vertices
                            createHoleTopology(startCorner + 1, holes[2 * lastHole + 1]);
                            lastHole++; // inkrementujeme pocet navstivenych hole
                            //connect right side of triangle with hole -> triangle touches it by 2 vertices
                            O[startCorner + 1] = startCorner + 10;
                            O[startCorner + 10] = startCorner + 1;

                            V[startCorner + 9] = V[startCorner + 2];
                            V[startCorner + 14] = V[startCorner + 2];
                            //start decompression
                            decompressTopology(startCorner);
                            if (!usedHoles.Contains(-2))
                            {
                                usedHoles.Add(-2);
                                Console.WriteLine("-2");
                            }
                            break;

                        case -3://3 vertices
                            createHoleTopology(startCorner + 2, holes[2 * lastHole + 1]);
                            lastHole++;
                            //connect right side of triangle with hole -> triangle touches it by 2 vertices
                            O[startCorner + 2] = startCorner + 10;
                            O[startCorner + 10] = startCorner + 2;
                            O[startCorner + 1] = startCorner + 13;
                            O[startCorner + 13] = startCorner + 1;

                            V[startCorner + 9] = V[startCorner];
                            V[startCorner + 14] = V[startCorner];
                            V[startCorner + 12] = V[startCorner + 2];
                            V[startCorner + 17] = V[startCorner + 2];

                            //start decompression
                            decompressTopology(startCorner + 5);
                            if (!usedHoles.Contains(-3))
                            {
                                usedHoles.Add(-3);
                                Console.WriteLine("-3");
                            }
                            break;

                        case -4://2 vertices
                            createHoleTopology(startCorner + 1, holes[2 * lastHole + 1]);
                            lastHole++;
                            decompressTopology(startCorner + 1);

                            V[prev(lastCorner1)] = V[next(lastCorner2)];
                            V[prev(O[next(lastCorner1)])] = V[next(lastCorner2)];

                            //start decompression
                            decompressTopology(startCorner);


                            if (!usedHoles.Contains(-4))
                            {
                                usedHoles.Add(-4);
                                Console.WriteLine("-4");
                            }

                            break;

                        case -5://3vertices, 1 edge
                            createHoleTopology(startCorner + 2, holes[2 * lastHole + 1]);
                            lastHole++;
                            //connect right side of triangle with hole -> triangle touches it by 2 vertices
                            O[startCorner + 2] = startCorner + 10;
                            O[startCorner + 10] = startCorner + 2;

                            V[startCorner + 9] = V[startCorner];
                            V[startCorner + 14] = V[startCorner];

                            decompressTopology(startCorner + 1);
                            
                            decompressTopology(startCorner + 5);

                            if (!usedHoles.Contains(-5))
                            {
                                usedHoles.Add(-5);
                                Console.WriteLine("-5");
                            }

                            break;

                        case -6://2vertices, one hole
                            createHoleTopology(startCorner + 2, holes[2 * lastHole + 1]);
                            lastHole++;

                            decompressTopology(startCorner + 2);
                            decompressTopology(startCorner + 1);
                            //start decompression
                            decompressTopology(startCorner + 5);

                            if (!usedHoles.Contains(-6))
                            {
                                usedHoles.Add(-6);
                                Console.WriteLine("-6");
                            }

                            break;

                        case -7://2 vertices, different holes
                            createHoleTopology(startCorner + 1, holes[2 * lastHole + 1]);
                            lastHole++;
                            createHoleTopology(startCorner, holes[2 * lastHole + 1]);
                            lastHole++;
                            decompressTopology(startCorner);

                            if (!usedHoles.Contains(-7))
                            {
                                usedHoles.Add(-7);
                                Console.WriteLine("-7");
                            }

                            break;

                        case -8://2 vertices, different holes
                            createHoleTopology(startCorner + 2, holes[2 * lastHole + 1]);
                            lastHole++;
                            O[startCorner + 2] = startCorner + 10;
                            O[startCorner + 10] = startCorner + 2;

                            V[startCorner + 9] = V[startCorner];
                            V[startCorner + 14] = V[startCorner];

                            createHoleTopology(startCorner, holes[2 * lastHole + 1]);
                            lastHole++;
                            decompressTopology(startCorner + 5);

                            if (!usedHoles.Contains(-8))
                            {
                                usedHoles.Add(-8);
                                Console.WriteLine("-8");
                            }

                            break;

                        case -9://lonely component
                            lastHole++;

                            if (!usedHoles.Contains(-9))
                            {
                                usedHoles.Add(-9);
                                Console.WriteLine("-9");
                            }

                            break;

                        default:
                            //start decompression
                            Console.WriteLine("Unknown type of starting component");
                            decompressTopology(startCorner);
                            break;
                    }
                }
            }
            else
            {
                decompressTopology(startCorner);
            }
            #endregion
        }

        /// <summary>
        /// Creates hole topology 
        /// </summary>
        /// <param name="c">Start corner</param>
        /// <param name="holeLength">´Hole length</param>
        private void createHoleTopology(int c, int holeLength)
        {
            /////////////////////////////////////////
            //creates "S'" triangle of '0' height -on the left side of "S'"
            /////////////////////////////////////////
            lastTriangle++;
            invisibleTriangles[invisibleTriCounter] = lastTriangle;
            invisibleTriCounter++;
            //oposite corners
            O[next(c)] = 3 * lastTriangle;
            O[3 * lastTriangle] = next(c);

            //new triangle, common edge with the last triangle
            V[3 * lastTriangle] = V[prev(c)];
            V[3 * lastTriangle + 1] = V[c];
            V[3 * lastTriangle + 2] = V[next(c)];

            c = O[next(c)];
            //prepare left side for zip
            O[prev(c)] = -1;
            c = next(c);

            //////////////////////////////////////////
            //create triangle of '0' heinght
            ///////////////////////////////////////////
            lastTriangle++;
            invisibleTriangles[invisibleTriCounter] = lastTriangle;
            invisibleTriCounter++;

            //set oposite corners
            O[c] = 3 * lastTriangle;
            O[3 * lastTriangle] = c;

            //new triangle, common edge with the last triangle
            V[3 * lastTriangle] = V[prev(c)];
            V[3 * lastTriangle + 1] = V[prev(c)];
            V[3 * lastTriangle + 2] = V[prev(c)];

            c = next(O[c]);
            int holeStartCorner = next(c);
            ////////////////////////////////////
            //create all triangles of hole
            /////////////////////////////////////
            int firstHoleVertex = lastVertex;
            //all triangles
            for (int i = 0; i < (holeLength - 1); i++)
            {
                //create triangle
                lastTriangle++; 
                invisibleTriangles[invisibleTriCounter] = lastTriangle;
                invisibleTriCounter++;
                O[c] = 3 * lastTriangle;
                O[3 * lastTriangle] = c;

                //load new vertex + store into 'G' table
                lastVertex++;
                V[3 * lastTriangle] = lastVertex;

                //2 common vetices with the last triangle
                V[3 * lastTriangle + 1] = V[prev(c)];
                V[3 * lastTriangle + 2] = V[next(c)];

                //right oposite set to -1
                c = prev(O[c]);
                O[prev(c)] = -1;
            }
            /////////////////////////////////////
            //last triangle of hole
            /////////////////////////////////////
            lastTriangle++;
            invisibleTriangles[invisibleTriCounter] = lastTriangle;
            invisibleTriCounter++;
            O[c] = 3 * lastTriangle;
            O[3 * lastTriangle] = c;

            //load new vertex
            V[3 * lastTriangle] = V[prev(c)];

            //2 common vetices with the last triangle
            V[3 * lastTriangle + 1] = V[prev(c)];
            V[3 * lastTriangle + 2] = V[next(c)];

            //right oposite set to -1
            c = prev(O[c]);
            O[prev(c)] = -1;
            //set oposite corner of the first and last triangle of hole
            O[c] = holeStartCorner;
            O[holeStartCorner] = c;
        }

        /// <summary>
        /// Decompress by CLERS
        /// </summary>
        /// <param name="c">Start corner</param>
        private void decompressTopology(int c)
        {
            while (true)
            {
                lastTriangle++;

                //indices of oposite corners
                O[c] = 3 * lastTriangle;
                O[3 * lastTriangle] = c;

                //set 2 vertices common with the last triangle
                V[3 * lastTriangle + 1] = V[prev(c)];
                V[3 * lastTriangle + 2] = V[next(c)];

                //go to oposite to the right edge of the triangle
                c = next(O[c]);

                //decode CLERS descrition
                switch (readNext(CLERS))
                {
                    case 'C':
                        O[next(c)] = -1;   //left edge prepare for "zip"
                        lastVertex++;      
                        V[3 * lastTriangle] = lastVertex;
                        break;
                    case 'L':
                        O[next(c)] = -2;   //left edge prepare for "zip" (-2, oposite corner unknown)
                        if (!checkHandle(next(c))) zip(next(c)); //run zip, if next(c) is not a part of handle
                        break;
                    case 'E':
                        O[c] = -2;       //right edge prepare for "zip"
                        O[next(c)] = -2; //left edge prepare for "zip"
                        checkHandle(c);  //if c (right edge) is part of handle, connect
                        if (!checkHandle(next(c))) zip(next(c)); //if left edge is part of handle, connect, else "zip"
                        //end of cyclus
                        return;
                    case 'R':
                        O[c] = -2;       //prepare right edge for "zip"
                        checkHandle(c);  //if c (right edge) is part of handle, connect
                        if (!checkHandle(c)) zip(c); //run zip, if next(c) is not a part of handle
                        c = next(c);     //go to next triangle corner -continue to the left
                        break;
                    case 'S':
                        sCounter++;
                        // if hole
                        if (checkHole() == true)
                        {
                            //create "S'" triangle -the same as "C"
                            lastVertex++;     
                            V[3 * lastTriangle] = lastVertex;

                            //create hole
                            createHoleTopology(c, holes[2 * lastHole + 1]);
                            lastHole++;
                        }
                        else // if not hole
                        {
                            decompressTopology(c);   //go recursive to the right
                            c = next(c);     //go to the next corner of the triangle -continue to the left
                            //if triangle visited (handle)
                            if (O[c] >= 0)
                            {
                                escapeCorners.Add(c);
                                return;
                            }
                        }
                        break;
                    default: Console.WriteLine("...unknown CLERS symbol!");
                        break;
                }
            }
        }

        /// <summary>
        /// Joins common edges of triangles (-2 edge with -1 edge)
        /// </summary>
        /// <param name="c">Start corner</param>
        private void zip(int c)
        {
            while (true)
            {
                //clockwise looking for free edge
                int b = next(c);
                while ((O[b] >= 0) && (O[b] != c))
                {
                    b = next(O[b]);
                }


                if (O[b] != -1) // if edge gelongs to "C" triangle -vertices known
                    return;

                O[c] = b;  //store oposite corners
                O[b] = c;

                lastCorner1 = b;
                lastCorner2 = c;

                int a = next(c);
                if (V[next(a)] < 0) //if vertice known (eg. in FirstHole)
                    V[next(a)] = V[next(b)]; //set common vertex

                while ((O[a] >= 0) && (a != b)) //set this common vertex also to others
                {
                    a = next(O[a]);
                    if (V[next(a)] < 0) //if vertice known (eg. in FirstHole)
                        V[next(a)] = V[next(b)];
                }

                //find the first unzipped edge on the right and try to zip it
                c = prev(c);
                while ((O[c] > 0) && (c != b))
                {
                    c = prev(O[c]);
                }

                if (O[c] != -2)
                {
                    break; //runs while unzipped edges exist
                }
            }
        }

        /// <summary>
        /// Finds out if hole reached
        /// </summary>
        /// <returns>True if hole reached</returns>
        private bool checkHole()
        {
            bool result = false;
            if (2 * lastHole < holes.Length)
                if (sCounter == holes[2 * lastHole])
                    result = true;

            return result;
        }

        /// <summary>
        /// If corner is a part of handle,
        /// joins and zips edges and returns true
        /// </summary>
        /// <param name="c">Tested corner</param>
        /// <returns>True, if handle reached</returns>
        private bool checkHandle(int c)
        {
            if ((lastHandle >= handles.Length)
                || (c != handles[lastHandle + 1]))
            {
                return false; //c is not handle
            }
            else //if "c" which belongs to "S" of handle-pair reached
            {
                //store oposite corners
                O[c] = handles[lastHandle];
                O[handles[lastHandle]] = c;

                //find the next closest unzipped edge
                int a = prev(c); //start with left edge
                while ((O[a] >= 0) && (a != handles[lastHandle]))
                {
                    a = prev(O[a]);
                }

                //if unzipped edge found and don't know it's vertices (it is not "C" from clers), zip it
                if (O[a] == -2) zip(a);

                //find the next closest unzipped edge
                a = prev(O[c]); //start in last triangle by edge on the right of common edge
                while ((O[a] >= 0) && (a != c))
                {
                    a = prev(O[a]);
                }

                //if unzipped edge found and don't know it's vertices (it is not "C" from clers), zip it
                if (O[a] == -2) zip(a);

                lastHandle += 2;

                return true;
            }
        }

        /// <summary>
        /// Read CLERS
        /// </summary>
        /// <param name="clers">CLERS</param>
        /// <returns>Red symbol</returns>
        private char readNext(char[] clers)
        {
            lastCLERS++;
            return clers[lastCLERS - 1];
        }

        /// <summary>
        /// Returns next corner of the triangle
        /// </summary>
        /// <param name="c">Start corner</param>
        /// <returns>Next corner</returns>
        private int next(int c)
        {
            if (c % 3 == 2)
            {
                return (c - 2);
            }
            else
            {
                return (c + 1);
            }
        }

        /// <summary>
        /// Returns previous corner of the triangle
        /// </summary>
        /// <param name="c">Start corner</param>
        /// <returns>Previous corner</returns>
        private int prev(int c)
        {
            if (c % 3 == 0)
            {
                return (c + 2);
            }
            else
            {
                return (c - 1);
            }
        }

        /// <summary>
        /// Returns Edgebreaker O-table
        /// </summary>
        /// <returns>O-table</returns>
        public int[] getO()
        {
            return O;
        }

        /// <summary>
        /// Returns Edgebreaker V-table
        /// </summary>
        /// <returns>V-table</returns>
        public int[] getV()
        {
            return V;
        }

        /// <summary>
        /// Returns array of invisible triangles
        /// </summary>
        /// <returns>Array of invisible triangles</returns>
        public int[] getInvisibleTriangles()
        {
            return invisibleTriangles;
        }

        /// <summary>
        /// Returns list escapeCorners
        /// </summary>
        /// <returns>List of escapeCorners</returns>
        public List<int> getEscapeCorners()
        {
            return escapeCorners;
        }
    }
}
