using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.IO.Compression;

namespace Zcu.Mve.Compression
{
    public class DirectCoder
    {
        public static void encode(int[] data, Stream target, int contextCount)
        {
            int min = int.MaxValue;
            int max = int.MinValue;
            for (int i = 0; i < data.Length; i++)
            {
                if (data[i] < min)
                    min = data[i];
                if (data[i] > max)
                    max = data[i];
            }

            BinaryWriter bw = new BinaryWriter(target);
            bw.Write(data.Length);
            bw.Write(min);
            if ((max - min) < 255)
            {
                for (int i = 0; i < data.Length; i++)
                {

                    bw.Write((byte)(data[i]-min));
                }
            }
            else
            {
                for (int i = 0; i < data.Length; i++)
                {

                    bw.Write((ushort)(data[i]-min));
                }
            }
            bw.Flush();
        }

        public static int[] decode(Stream source)
        {
            BinaryReader br = new BinaryReader(source);
            int length = br.ReadInt32();
            int[] result = new int[length];
            for (int i = 0; i < length; i++)
            {
                result[i] = br.ReadInt16();
            }
            return (result);
        }
    }

    public class DeltaHuffmannCoder
    {
        public static void encode(int[] data, Stream target, int contextCount)
        {
            int[] reordered = new int[data.Length];
            int cnt = data.Length / contextCount;
            int pos = 0;
            for (int comp = 0; comp < contextCount; comp++)
            {
                for (int i = 0; i < cnt; i++)
                {
                    reordered[pos] = data[contextCount * i + comp];
                    pos++;
                }
            }
            for (int i = reordered.Length - 1; i > 0; i--)
            {
                reordered[i] = reordered[i] - reordered[i - 1];
            }
            BinaryWriter sw = new BinaryWriter(target);
            sw.Write(contextCount);
            sw.Flush();
            HuffmanCoder.encode(reordered, target, 1, false);
        }

        public static int[] decode(Stream source)
        {
            BinaryReader br = new BinaryReader(source);
            int contextCount = br.ReadInt32();
            int[] data = HuffmanCoder.decode(source);

            for (int i = 1; i < data.Length; i++)
            {
                data[i] += data[i - 1];
            }

            int cnt = data.Length / contextCount;

            int[] result = new int[data.Length];

            for (int v = 0; v < cnt; v++)
            {
                for (int c = 0; c < contextCount; c++)
                {
                    result[v * contextCount + c] = data[c * cnt + v];
                }
            }
            return (result);
        }
    }

    class DeltaHuffmannCoderDirect
    {
        public static void encode(int[] data, Stream target, int contextCount)
        {            
            for (int i = data.Length - 1; i > 0; i--)
            {
                data[i] = data[i] - data[i - 1];
            }
            BinaryWriter sw = new BinaryWriter(target);
            sw.Write(contextCount);
            sw.Flush();
            HuffmanCoder.encode(data, target, 1, true);
        }

        public static int[] decode(Stream source)
        {
            BinaryReader br = new BinaryReader(source);
            int contextCount = br.ReadInt32();
            int[] data = HuffmanCoder.decode(source);

            for (int i = 1; i < data.Length; i++)
            {
                data[i] += data[i - 1];
            }
            
            return (data);
        }
    }

    public class ZipCoder
    {

        public static void encode(int[] data, Stream target, int contextCount)
        {
            GZipStream stream = new GZipStream(target, CompressionMode.Compress,true);

            using (MemoryStream temp = new MemoryStream())
            {
                BinaryWriter writer = new BinaryWriter(temp);
                writer.Write(data.Length);
                writer.Write(contextCount);
                int cnt = data.Length / contextCount;
                for (int component = 0; component < contextCount; component++)
                {
                    for (int j = 0; j < cnt; j++)
                    {
                        writer.Write(data[j * contextCount + component]);
                    }
                }
                writer.Flush();

                temp.Position = 0;
                byte[] bytes = temp.GetBuffer();


                stream.Write(bytes, 0, bytes.Length);

                stream.Flush();
                stream.Close();
            }
        }

        public static int[] decode(Stream source)
        {
            GZipStream stream = new GZipStream(source, CompressionMode.Decompress);            
            BinaryReader reader = new BinaryReader(stream);
            int length = reader.ReadInt32();
            int compCount = reader.ReadInt32();
            int cnt = length/compCount;
            int[] output = new int[length];
            for (int i = 0;i<compCount;i++)
                for (int j = 0; j < cnt; j++)
                {
                    output[j * compCount + i] = reader.ReadInt32();
                }
            return (output);
        }
    }

    public class DeflateCoder
    {

        public static void encode(int[] data, Stream target, int contextCount)
        {
            DeflateStream stream = new DeflateStream(target, CompressionMode.Compress, true);
            BinaryWriter writer = new BinaryWriter(stream);
            writer.Write(data.Length);
            writer.Write(contextCount);
            int cnt = data.Length / contextCount;
            for (int component = 0; component < contextCount; component++)
            {
                for (int j = 0; j < cnt; j++)
                {
                    writer.Write(data[j * contextCount + component]);
                }
            }
            writer.Flush();
            stream.Flush();
            //stream.Close();
        }

        public static int[] decode(Stream source)
        {
            DeflateStream stream = new DeflateStream(source, CompressionMode.Decompress);
            BinaryReader reader = new BinaryReader(stream);
            int length = reader.ReadInt32();
            int compCount = reader.ReadInt32();
            int cnt = length / compCount;
            int[] output = new int[length];
            for (int i = 0; i < compCount; i++)
                for (int j = 0; j < cnt; j++)
                {
                    output[j * compCount + i] = reader.ReadInt32();
                }
            return (output);
        }
    }

    /// <summary>
    /// Huffman encoder
    /// </summary>
    public class HuffmanCoder
    {
        public HuffmanCoder()
        {
        }

        public static void encode(int[] data, Stream target, int contextCount, bool reorder)
        {
            position = 0;
            buffer = 0;

            HuffmanNode[] roots = new HuffmanNode[contextCount];
            Dictionary<int, bool[]>[] codetables = new Dictionary<int, bool[]>[contextCount];
            for (int context = 0; context < contextCount; context++)
            {
                Dictionary<int, HuffmanNode> cases = new Dictionary<int, HuffmanNode>();
                for (int i = 0; i < data.Length/contextCount; i++)
                {
                    int v = 0;
                    if (reorder)
                        v = data[i * contextCount + context];
                    else
                        v = data[context * (data.Length / contextCount) + i];
                    if (cases.ContainsKey(v))
                    {
                        cases[v].weight++;
                    }
                    else
                    {
                        cases[v] = new HuffmanNode(1, v);
                    }
                }
                HuffmanNode[] nodeArray = new HuffmanNode[cases.Values.Count];
                int j = 0;
                foreach (HuffmanNode node in cases.Values)
                {
                    nodeArray[j] = node;
                    j++;
                }
                Array.Sort(nodeArray);
                if (nodeArray.Length > 1)
                {
                    List<HuffmanNode> list1 = new List<HuffmanNode>(nodeArray);
                    List<HuffmanNode> list2 = new List<HuffmanNode>();
                    while (list1.Count + list2.Count > 1)
                    {
                        List<HuffmanNode> smallest = new List<HuffmanNode>();
                        if (list1.Count > 1)
                        {
                            smallest.Add(list1[1]);
                        }
                        if (list1.Count > 0)
                        {
                            smallest.Add(list1[0]);
                        }
                        if (list2.Count > 1)
                        {
                            smallest.Add(list2[1]);
                        }
                        if (list2.Count > 0)
                        {
                            smallest.Add(list2[0]);
                        }
                        HuffmanNode[] sm = smallest.ToArray();
                        Array.Sort(sm);
                        HuffmanNode sm1 = sm[0];
                        HuffmanNode sm2 = sm[1];                        
                        if (sm1.isLeaf)
                        {
                            if (list1[0].value == sm1.value)
                                list1.RemoveAt(0);
                            else
                                list1.RemoveAt(1);
                        }
                        else
                        {
                            if (list2[0].value == sm1.value)
                                list2.RemoveAt(0);
                            else
                                list2.RemoveAt(1);
                        }
                        if (sm2.isLeaf)
                        {
                            if (list1[0].value == sm2.value)
                                list1.RemoveAt(0);
                            else
                                list1.RemoveAt(1);
                        }
                        else
                        {
                            if (list2[0].value == sm2.value)
                                list2.RemoveAt(0);
                            else
                                list2.RemoveAt(1);
                        }
                        list2.Add(sm1.Join(sm2));
                    }

                    roots[context] = list2[0];
                }
                else
                {
                    roots[context] = nodeArray[0];
                }
                codetables[context] = new Dictionary<int, bool[]>();
                fillCodeTable(codetables[context], new List<bool>(), roots[context]);
            }
            writeInt((uint)contextCount, 10, target);
            for (int i = 0; i < contextCount; i++)
            {
                encodeTable(roots[i], target);
            }

            writeInt((uint)data.Length, 30, target);
            int contextLength = data.Length/contextCount;
            for (int i = 0; i < data.Length; i++)
            {
                bool[] code = null;
                if (reorder)
                    code = codetables[i % contextCount][data[i]];
                else
                    code = codetables[i / contextLength][data[i]];
                for (int k = 0; k < code.Length; k++)
                {
                    writeBit(target, code[k]);
                }
            }
            FlushStream(target);
        }

        public static int[] decode(Stream source)
        {
            position = 0;
            buffer = 0;
            
            int contextCount = (int)readInt(10, source);

            HuffmanNode[] roots = new HuffmanNode[contextCount];

            for (int context = 0; context < contextCount; context++)
            {
                roots[context] = new HuffmanNode(0, 0);
                decodeTable(roots[context], source);
            }

            int count = (int)readInt(30, source);
            int[] result = new int[count];
            int pos = 0;
            int currentContext = 0;
            HuffmanNode current = roots[currentContext];
            while (pos < count)
            {
                if (current.isLeaf)
                {
                    result[pos] = current.value;
                    pos++;
                    currentContext++;
                    if (currentContext == contextCount)
                        currentContext = 0;
                    current = roots[currentContext];
                }
                else
                {
                    if (readBit(source))
                    {
                        current = current.right;
                    }
                    else
                    {
                        current = current.left;
                    }
                }
            }
            return (result);
        }

        private static void encodeTable(HuffmanNode root, Stream s)
        {
            int min = root.GetMinValue();
            int max = root.GetMaxValue();
            int bitCount = (int)Math.Ceiling(Math.Log(max - min+1, 2));
            if (min<0)
            {
                writeBit(s, true);
                writeInt((uint)-min, 30, s);
            }
            else
            {
                writeBit(s, false);
                writeInt((uint)min, 30, s);
            }
            writeInt((uint)bitCount, 6, s);
            
            encodeNode(root, s, min, bitCount);
        }

        private static void decodeTable(HuffmanNode root, Stream s)
        {
            bool revert = readBit(s);
            int min = (int)readInt(30, s);
            if (revert)
                min = -min;
            int bitCount = (int)readInt(6, s);

            decodeNode(root, s, min, bitCount);
        }


        private static void encodeNode(HuffmanNode node, Stream s, int min, int bitCount)
        {
            if (node.isLeaf)
            {
                writeBit(s,true);
                writeInt((uint)(node.value - min), bitCount, s);
            }
            else
            {
                writeBit(s, false);
                encodeNode(node.left, s, min, bitCount);
                encodeNode(node.right, s, min, bitCount);
            }
        }

        private static void decodeNode(HuffmanNode node, Stream s, int min, int bitCount)
        {
            if (readBit(s))
            {
                node.value = (int)readInt(bitCount, s) + min;
                node.isLeaf = true;
            }
            else
            {
                node.isLeaf = false;
                node.left = new HuffmanNode(0, 0);
                decodeNode(node.left, s, min, bitCount);
                node.right = new HuffmanNode(0, 0);
                decodeNode(node.right, s, min, bitCount);
            }
        }

        private static void fillCodeTable(Dictionary<int, bool[]> codetable, List<bool> code, HuffmanNode node)
        {
            if (node.isLeaf)
            {
                if (!codetable.ContainsKey(node.value))
                {
                    codetable.Add(node.value, code.ToArray());
                }
                else
                {
                    Console.WriteLine("Huffmann failure");
                }
            }
            else
            {
                code.Add(false);
                fillCodeTable(codetable, code, node.left);
                code[code.Count - 1] = true;
                fillCodeTable(codetable, code, node.right);
                code.RemoveAt(code.Count - 1);
            }
        }

        private static byte buffer;
        private static int position;

        private static void writeBit(Stream s, bool bit)
        {
            if (bit)
            {
                switch (position)
                {
                    case 0: buffer += 1;
                        break;
                    case 1: buffer += 2;
                        break;
                    case 2: buffer += 4;
                        break;
                    case 3: buffer += 8;
                        break;
                    case 4: buffer += 16;
                        break;
                    case 5: buffer += 32;
                        break;
                    case 6: buffer += 64;
                        break;
                    case 7: buffer += 128;
                        break;
                }
            }
            position++;
            if (position == 8)
            {
                position = 0;
                s.WriteByte(buffer);
                buffer = 0;
            }
        }

        private static void writeInt(uint n, int bitCount, Stream s)
        {
            uint m = 1;
            for (int i = 0; i < bitCount; i++)
            {
                if ((m & n) > 0)
                    writeBit(s, true);
                else
                    writeBit(s, false);
                m *= 2;
            }
        }

        private static uint readInt(int bitCount, Stream s)
        {
            uint r = 0;
            uint m = 1;
            for (int i = 0; i < bitCount; i++)
            {
                if (readBit(s))
                {
                    r += m;
                }
                m*=2;
            }
            return (r);
        }


        private static void FlushStream(Stream s)
        {
            if (position != 0)
            {
                s.WriteByte(buffer);
                position = 0;
            }
        }

        private static bool readBit(Stream s)
        {
            if (position == 0)
            {
                buffer = (Byte)s.ReadByte();
            }
            
            bool result = false;
            switch (position)
            {
                case 0:
                    result = ((buffer & 1)>0);
                    break;
                case 1:
                    result = ((buffer & 2) > 0);
                    break;
                case 2:
                    result = ((buffer & 4) > 0);
                    break;
                case 3:
                    result = ((buffer & 8) > 0);
                    break;
                case 4:
                    result = ((buffer & 16) > 0);
                    break;
                case 5:
                    result = ((buffer & 32) > 0);
                    break;
                case 6:
                    result = ((buffer & 64) > 0);
                    break;
                case 7:
                    result = ((buffer & 128) > 0);
                    break;
            }            
            position++;
            if (position == 8)
                position = 0;
            return (result);
        }
    }    

    class HuffmanNode : IComparable<HuffmanNode>
    {
        public int value;
        public int weight;
        public HuffmanNode left, right, parent;
        public bool isLeaf = true;

        public static int index = 0;

        public HuffmanNode(int weight, int value)
        {
            this.weight = weight;
            this.value = value;
        }

        public HuffmanNode Join(HuffmanNode target)
        {
            HuffmanNode result = new HuffmanNode(this.weight + target.weight, 0);
            result.left = this;
            result.right = target;
            this.parent = result;
            target.parent = result;
            result.isLeaf = false;
            result.value = index;
            index++;
            return (result);
        }

        public int GetMinValue()
        {
            if (isLeaf)
                return(this.value);
            else return(Math.Min(this.left.GetMinValue(), this.right.GetMinValue()));
        }

        public int GetMaxValue()
        {
            if (isLeaf)
                return(this.value);
            else return(Math.Max(this.left.GetMaxValue(), this.right.GetMaxValue()));
        }        

        #region IComparable<HuffmanNode> Members

        public int CompareTo(HuffmanNode other)
        {
            return (this.weight.CompareTo(other.weight));
        }

        #endregion
    }
}
