﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Compares original and distorted version of a mesh using mean square error method.
    /// </summary>
    [ModuleInfo("Oldřich Petřík", "Compares original and distorted version of a mesh using mean square error method.", CategoryType.Filter, CategoryData.Mesh)]
    [PortInfo("InMeshOriginal", "Original version of the input mesh.")]
    [PortInfo("InMeshDistorted", "Distorted version of the input mesh.")]
    [PortInfo("OutMSE", "Mean square error.")]
    class MeshComparerMSE : Module
    {
        private Scalar _res;

        private bool _done = false;


        public MeshComparerMSE()
        {
            AddInPort("InMeshOriginal", typeof(UnstrGrid));
            AddInPort("InMeshDistorted", typeof(UnstrGrid));
            AddOutPort("OutMSE", typeof(Scalar));
        }

        bool normalize = true;

        public bool Normalize
        {
            get { return normalize; }
            set { normalize = value; }
        }

        public override void Execute()
        {
            if (_done && IsAllDataSame() && IsAllUserSettingsSame())
            {
                SetOutput("OutMSE", _res, true);
            }
            else
            {
                UnstrGrid original = (UnstrGrid)GetInput("InMeshOriginal");
                UnstrGrid distorted = (UnstrGrid)GetInput("InMeshDistorted");

                _res = CalculateError(original, distorted);

                WriteLog("MeanSquaredError", (double)_res);

                SetOutput("OutMSE", _res, false);
            }
        }

        private double CalculateError(UnstrGrid original, UnstrGrid distorted)
        {
            Point3D[] pOrig = (Point3D[])original.Points.Array;
            Point3D[] pDist = (Point3D[])distorted.Points.Array;
            Point3D pO, pD;
            double edgeSum = 0;
            double sum = 0;

            if (normalize)
            {
                foreach (Triangle t in (original.GetCells(typeof(Triangle))[0]))
                {
                    edgeSum += pOrig[t.V1].SquareDistanceTo(pOrig[t.V2]);
                    edgeSum += pOrig[t.V1].SquareDistanceTo(pOrig[t.V3]);
                    edgeSum += pOrig[t.V2].SquareDistanceTo(pOrig[t.V3]);
                }
                edgeSum *= 0.5;
            }

            for (int i = 0; i < pOrig.Length; i++)
            {
                pO = pOrig[i];
                pD = pDist[i];

                sum += pO.SquareDistanceTo(pD);

                ProgressInfo((i + 1) / pOrig.Length);
            }

            if (normalize)
                return sum / edgeSum;
            else
                return sum / original.Points.Count;
        }
    }
}
