﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Compares original and distorted version of a mesh using the Visual Error metric by Karni and Gotsman.
    /// </summary>
    [ModuleInfo("Oldřich Petřík", "Compares original and distorted version of a mesh using the Visual Error metric by Karni and Gotsman.", CategoryType.Filter, CategoryData.Mesh)]
    [PortInfo("InMeshOriginal", "Original version of the input mesh.")]
    [PortInfo("InMeshDistorted", "Distorted version of the input mesh.")]
    [PortInfo("OutVisualKGError", "Visual error.")]
    public class MeshComparerVisualKG : Module
    {
        private Scalar _res;

        private bool _done = false;

        /// <summary>
        /// MeshComparerVisualKG constructor.
        /// </summary>
        public MeshComparerVisualKG()
        {
            AddInPort("InMeshOriginal", typeof(UnstrGrid));
            AddInPort("InMeshDistorted", typeof(UnstrGrid));
            AddOutPort("OutVisualKGError", typeof(Scalar));
        }

        ///<inheritdoc />
        public override void Execute()
        {
            if (_done && IsAllDataSame() && IsAllUserSettingsSame())
            {
                SetOutput("VisualKGError", _res, true);
            }
            else
            {
                UnstrGrid original = (UnstrGrid)GetInput("InMeshOriginal");
                UnstrGrid distorted = (UnstrGrid)GetInput("InMeshDistorted");

                _res = CalculateError(original, distorted);

                WriteLog("VisualError", (double)_res);

                SetOutput("OutVisualKGError", _res, false);
            }
        }

        private List<int>[] GetIncidence(UnstrGrid mesh)
        {
            Triangle[] triangles = (Triangle[])(mesh.GetCells(typeof(Triangle))[0]).Array;
            int count = mesh.Points.Count;
            List<int>[] res = new List<int>[count];
            for (int i = 0; i < count; i++) res[i] = new List<int>();
            int v1, v2, v3;

            foreach (Triangle t in triangles)
            {
                v1 = t.V1;
                v2 = t.V2;
                v3 = t.V3;

                if (!res[v1].Contains(v2))
                {
                    res[v1].Add(v2);
                    res[v2].Add(v1);
                }
                if (!res[v1].Contains(v3))
                {
                    res[v1].Add(v3);
                    res[v3].Add(v1);
                }
                if (!res[v2].Contains(v3))
                {
                    res[v2].Add(v3);
                    res[v3].Add(v2);
                }
            }
            return res;
        }

        private double CalculateError(UnstrGrid original, UnstrGrid distorted)
        {
            List<int>[] incidence = GetIncidence(original);
            Point3D[] pOrig = (Point3D[])original.Points.Array;
            Point3D[] pDist = (Point3D[])distorted.Points.Array;
            Point3D glO, glD;
            double ilenO, ilenD, lenO, lenD;
            double sumLO, sumLD;
            Point3D sumVO, sumVD;
            Point3D pO, pD, qO, qD;
            List<int> nbh;
            double sum = 0;

            for (int i = 0; i < pOrig.Length; i++)
            {
                nbh = incidence[i];
                pO = pOrig[i];
                pD = pDist[i];
                sumLO = sumLD = 0;
                sumVO = new Point3D(0, 0, 0);
                sumVD = new Point3D(0, 0, 0);

                foreach (int n in nbh)
                {
                    qO = pOrig[n];
                    qD = pDist[n];
                    lenO = pO.DistanceTo(qO);
                    lenD = pD.DistanceTo(qD);
                    ilenO = (lenO == 0) ? 0 : (1.0 / lenO);
                    ilenD = (lenD == 0) ? 0 : (1.0 / lenD);
                    sumVO += qO * ilenO;
                    sumVD += qD * ilenD;
                    sumLO += ilenO;
                    sumLD += ilenD;
                }

                glO = pO - (sumVO / sumLO);
                glD = pD - (sumVD / sumLD);

                sum += pO.DistanceTo(pD) + glO.DistanceTo(glD);

                ProgressInfo((i + 1) / pOrig.Length);
            }

            return sum / (2 * pOrig.Length);
        }
    }
}
