﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Module for controlled distortion of static triangular meshes
    /// </summary>
    [ModuleInfo("Libor Váša", "Module for controlled distortion of static triangular meshes", CategoryType.Filter, CategoryData.Mesh)]
    [PortInfo("InMesh", "input original mesh")]
    [PortInfo("OutMesh", "distorted output mesh")]
    class MeshDistorter:Module
    {
        public MeshDistorter()
        {
            AddInPort("InMesh", typeof(UnstrGrid));
            AddInPort("Amount", typeof(Scalar), false);
            AddOutPort("OutMesh", typeof(UnstrGrid));
        }

        double amount;

        public double Amount
        {
            get { return amount; }
            set { amount = value; }
        }

        double frequency = 1;

        public double Frequency
        {
            get { return frequency; }
            set { frequency = value; }
        }

        int axis = 1;

        public int Axis
        {
            get { return axis; }
            set { axis = value; }
        }

        public enum DistortionType { Gauss, Uniform, Sine, UniformSeparateAxis};

        private DistortionType distortionType = DistortionType.Gauss;

        public DistortionType DistortionType1
        {
            get { return distortionType; }
            set { distortionType = value; }
        }

        private static Random r = new Random();

        public override void Execute()
        {
            UnstrGrid input = (UnstrGrid)GetInput("InMesh");
            if (GetInPort("Amount").Link != null)
            {
                amount = ((Scalar)GetInput("Amount")).Value;
            }
            double diagonal = input.BoundingBox.MaxDiag;
            double delta = diagonal / Math.Pow(2, amount);
            UnstrGrid output = (UnstrGrid)input.DeepCopy();
            double mX = 1;
            double mY = 1;
            double mZ = 1;
            if (distortionType == DistortionType.UniformSeparateAxis)
            {
                BoundingBox bb = input.BoundingBox;
                IPoint maxPoint = (IPoint)bb.MaxPoint;
                IPoint minPoint = (IPoint)bb.MinPoint;
                mX = maxPoint[0] - minPoint[0];
                mY = maxPoint[1] - minPoint[1];
                mZ = maxPoint[2] - minPoint[2];
                double max = mX;
                if (mY > max)
                    max = mY;
                if (mZ > max)
                    max = mZ;
                mX = mX / max;
                mY = mY / max;
                mZ = mZ / max;
            }
            if (distortionType == DistortionType.Sine)
            {

                BoundingBox bb = input.BoundingBox;
                IPoint maxPoint = (IPoint)bb.MaxPoint;
                IPoint minPoint = (IPoint)bb.MinPoint;
                mX = (frequency * Math.PI * 2) / (maxPoint[0] - minPoint[0]);
                mY = (frequency * Math.PI * 2) / (maxPoint[1] - minPoint[1]);
                mZ = (frequency * Math.PI * 2) / (maxPoint[2] - minPoint[2]);
            }
            for (int i = 0; i < input.Points.Count; i++)
            {
                Point3D point = (Point3D)input.Points[i];
                switch (distortionType)
                {
                    case DistortionType.Gauss:
                        point.X = point.X + AnimationDistorter.Gauss(amount);
                        point.Y = point.Y + AnimationDistorter.Gauss(amount);
                        point.Z = point.Z + AnimationDistorter.Gauss(amount);
                        break;
                    case DistortionType.Sine:
                        switch (axis)
                        {
                            case 0:
                                point.X = point.X + amount * Math.Sin(point.X * mX);
                                point.Y = point.Y + amount * Math.Sin(point.X * mX);
                                point.Z = point.Z + amount * Math.Sin(point.X * mX);
                                break;
                            case 1:
                                point.X = point.X + amount * Math.Sin(point.Y * mY);
                                point.Y = point.Y + amount * Math.Sin(point.Y * mY);
                                point.Z = point.Z + amount * Math.Sin(point.Y * mY);
                                break;
                            case 2:
                                point.X = point.X + amount * Math.Sin(point.Z * mZ);
                                point.Y = point.Y + amount * Math.Sin(point.Z * mZ);
                                point.Z = point.Z + amount * Math.Sin(point.Z * mZ);
                                break;
                        }                        
                        break;
                    case DistortionType.Uniform:
                        point.X = point.X + amount * r.NextDouble();
                        point.Y = point.Y + amount * r.NextDouble();
                        point.Z = point.Z + amount * r.NextDouble();
                        break;
                    case DistortionType.UniformSeparateAxis:
                        point.X = point.X + amount * r.NextDouble()*mX;
                        point.Y = point.Y + amount * r.NextDouble()*mY;
                        point.Z = point.Z + amount * r.NextDouble()*mZ;
                        break;
                }                
                output.Points[i] = point;
            }
            SetOutput("OutMesh", output);
        }

    }
}
