using System;
using System.Collections.Generic;
using MathNet.Numerics.LinearAlgebra;
using System.IO;
using Zcu.Mve.SVD;
using Zcu.Mve.Compression;
using Zcu.Mve.Core;
using System.Runtime.InteropServices;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Class representing encapsulated PCA basis. Can be used to compute PCA of given data, get PCA coefficients for data vectors and recosntruct
    /// data vectors from the coefficients.
    /// </summary>
    public class PCA
    {
        /// <summary>
        /// determines a method for computation of PCA. Direct is based on eigenvalue decomposition, and is used when number of samples is larger
        /// than the dimension of the space. SVD is used when the dimension of the space is larger tahn the number of samples.
        /// </summary>
        public enum PCAMethod
        {
            /// <summary>
            /// direct PCA using eigenvalue decomposition
            /// </summary>
            direct,
            /// <summary>
            /// PCA using singular value decomposition
            /// </summary>
            SVD
        }

        /// <summary>
        /// Determines the method used by the PCA instance.
        /// </summary>
        public PCAMethod method = PCAMethod.direct;

        bool subtractMean = true;

        /// <summary>
        /// determines whether or not to subtract mean value. True for most cases.
        /// </summary>
        public bool SubtractMean
        {
            get { return subtractMean; }
            set { subtractMean = value; }
        }

        private double[][] basisBackup;
        private double[] meanBackup;

        /// <summary>
        /// saves PCA basis into an internal structure for future retrieval by RestoreBasis()
        /// </summary>
        public void BackupBasis()
        {
            basisBackup = new double[basis.Length][];
            for (int i = 0; i < basis.Length; i++)
            {
                basisBackup[i] = new double[basis[i].Length];
                for (int j = 0; j < basis[i].Length; j++)
                {
                    basisBackup[i][j] = basis[i][j];
                }
            }
            meanBackup = new double[means.Length];
            for (int i = 0; i < means.Length; i++)
            {
                meanBackup[i] = means[i];
            }
        }

        /// <summary>
        /// restores basis (previously stored by calling BackupBasis) from the internal structure
        /// </summary>
        public void RestoreBasis()
        {
            for (int i = 0; i < basis.Length; i++)
            {
                for (int j = 0; j < basis[i].Length; j++)
                {
                    basis[i][j] = basisBackup[i][j];
                }
            }
            for (int i = 0; i < means.Length; i++)
            {
                means[i] = meanBackup[i];
            }
        }

        private int sampleCount;

        /// <summary>
        /// number of components of the used samples
        /// </summary>
        public int sampleLength;

        private double[][] data;

        /// <summary>
        /// Data to perform PCA upon. Each vector should have the same length.
        /// </summary>
        public double[][] Data
        {
            get { return data; }
            set
            {
                data = value;
                this.sampleCount = value.Length;
                this.sampleLength = value[0].Length;
            }

        }

        /// <summary>
        /// PCA basis, each element is one basis vector
        /// </summary>
        public double[][] basis;

        /// <summary>
        /// creates a volume data file with prediction residuals for a given basis vector
        /// </summary>
        /// <param name="index">Index of the basis vector to be processed</param>
        public void AnalyzeBasisVector(int index)
        {
            double span = 1;
            double spanstep = span / 64;
            double[] residuals = new double[basis[index].Length - 9];
            double q = 0.0001;
            using (BinaryWriter bw = new BinaryWriter(File.Open("out.dat", FileMode.Create)), bw2 = new BinaryWriter(File.Open("out2.dat", FileMode.Create)))
            {
                for (int i = 0; i < 128; i++)
                {
                    for (int j = 0; j < 128; j++)
                    {
                        for (int k = 0; k < 128; k++)
                        {
                            double l1 = 3 - span + i * spanstep;
                            double l2 = -3 - span + j * spanstep;
                            double l3 = 1 - span + k * spanstep;
                            for (int s = 9; s < basis[index].Length; s++)
                            {
                                double prediction = l1 * basis[index][s - 3] + l2 * basis[index][s - 6] + l3 * basis[index][s - 9];
                                residuals[s - 9] = prediction - basis[index][s];
                            }

                            Array.Sort(residuals);
                            double entropy = 0;

                            Dictionary<int, int> cases = new Dictionary<int, int>();
                            for (int l = 1; l < residuals.Length; l++)
                            {
                                entropy += -Math.Log(1 / (residuals[l] - residuals[l - 1]), 2);
                                int quantized = (int)Math.Round(residuals[l] / q);
                                if (cases.ContainsKey(quantized))
                                {
                                    cases[quantized] = cases[quantized] + 1;
                                }
                                else
                                {
                                    cases[quantized] = 1;
                                }
                            }

                            double e2 = 0;
                            foreach (int key in cases.Keys)
                            {
                                double p = (double)cases[key] / residuals.Length;
                                e2 += -p * Math.Log(p, 2);
                            }

                            bw.Write((int)(e2 * 1000));
                            bw2.Write((int)(-entropy));
                        }
                    }
                }
            }
        }

        /// <summary>
        /// writes the PCA basis into a text file
        /// </summary>
        /// <param name="count">Number of basis vectors to be output</param>
        /// <param name="fileName">Name of the output file</param>
        public void ExportBasis(int count, string fileName)
        {
            using (StreamWriter sw = new StreamWriter(fileName))
            {
                sw.WriteLine(basis[0].Length / 3);
                sw.WriteLine(count);

                for (int f = 0; f < (basis[0].Length / 3); f++)
                {
                    int index = 3 * f;
                    string line = means[index].ToString() + " ";
                    for (int i = 0; i < count; i++)
                    {
                        line += basis[i][index].ToString() + " ";
                    }
                    line += data[2223][index];
                    sw.WriteLine(line);
                }
                sw.WriteLine();
            }
        }

        /// <summary>
        /// Constructs an empty PCA instance
        /// </summary>
        public PCA()
        {
            this.method = PCAMethod.direct;
            this.subtractMean = true;
        }

        /// <summary>
        /// Constructs a new PCA instance
        /// </summary>
        /// <param name="method">Method of computing PCA</param>
        /// <param name="subtractMean">Determines whether or not to subtact means (value true is suggested in most situations)</param>
        public PCA(PCAMethod method, bool subtractMean)
        {
            this.method = method;
            this.subtractMean = subtractMean;
        }

        private double[] means;

        /// <summary>
        /// mean values.
        /// </summary>
        public double[] Means
        {
            get { return means; }
            set { means = value; }
        }

        /// <summary>
        /// Creates an instance of PCA by loading its basis vectors and mean vectors from 
        /// a stream previously created by calling the EncodeBasisCobra() method.
        /// </summary>
        /// <param name="source">Source to load the basis from.</param>
        public PCA(Stream source, bool approximateQuantizations)
        {
            this.subtractMean = true;
            this.method = PCA.PCAMethod.direct;
            this.decodeCobra(source, approximateQuantizations);
        }

        /// <summary>
        /// Creates an instance of PCA by loading its basis vectors and mean vectors from 
        /// a stream previously created by calling the EncodeBasisCobra() method.
        /// </summary>
        /// <param name="source">Source to load the basis from.</param>
        public PCA(Stream source)
        {
            this.subtractMean = true;
            this.method = PCA.PCAMethod.direct;
            this.decodeCobra(source, true);
        }

        public void EncodeBasisDouble(Stream target, int vectorCount)
        {
            BinaryWriter bw = new BinaryWriter(target);
            bw.Write(vectorCount);
            bw.Write(basis[0].Length);
            for (int i = 0; i < vectorCount; i++)
            {
                for (int j = 0; j < basis[0].Length; j++)
                    bw.Write(basis[i][j]);
            }
            for (int j = 0; j < basis[0].Length; j++)
                bw.Write(means[j]);
        }

        public void DecodeBasisDouble(Stream source)
        {
            BinaryReader br = new BinaryReader(source);
            int vectorCount = br.ReadInt32();
            int length = br.ReadInt32();
            this.basis = new double[vectorCount][];
            for (int i = 0; i < vectorCount; i++)
            {
                basis[i] = new double[length];
                for (int j = 0; j < length; j++)
                    basis[i][j] = br.ReadDouble();
            }
            this.means = new double[length];
            for (int j = 0; j < length; j++)
                means[j] = br.ReadDouble();

            this.sampleLength = length;
        }

        /// <summary>
        /// writes the basis into a file, using uniform quantization
        /// </summary>
        /// <param name="count">number of basis vectors to be used</param>
        /// <param name="order">order of prediction to be used</param>
        /// <param name="delta">quantization constant</param>
        /// <param name="target">target stream to write to</param>
        /// <returns></returns>
        public double EncodeBasisUni(int count, int order, double delta, Stream target)
        {
            // this thing performs LPC upon the basis vectors and returns its size in bits

            Dictionary<int, int> cases = new Dictionary<int, int>();
            int valueCount = 0;
            int rowCount = basis[0].Length - order;

            MathNet.Numerics.LinearAlgebra.Matrix m = new MathNet.Numerics.LinearAlgebra.Matrix(rowCount * count, order);
            MathNet.Numerics.LinearAlgebra.Matrix r = new MathNet.Numerics.LinearAlgebra.Matrix(rowCount * count, 1);

            for (int i = 0; i < count; i++)
            {

                for (int row = 0; row < rowCount; row++)
                {
                    r[row + rowCount * i, 0] = basis[i][row + order];
                    for (int c = 0; c < order; c++)
                        m[row + rowCount * i, c] = basis[i][row + c];
                }
            }

            MathNet.Numerics.LinearAlgebra.Matrix s = m.Solve(r);

            int origLength = 0;
            BinaryWriter writer = null;

            if (target != null)
            {
                origLength = (int)target.Length;
                writer = new BinaryWriter(target);
                writer.Write((byte)order);
                writer.Write((ushort)count);
                writer.Write((ushort)basis[0].Length);
                for (int i = 0; i < basis[0].Length; i++)
                {
                    writer.Write((float)means[i]);
                }
                writer.Write(delta);
                for (int i = 0; i < order; i++)
                {
                    writer.Write(s[i, 0]);
                }
            }

            int[] data = new int[count * basis[0].Length];

            for (int i = 0; i < count; i++)
            {
                for (int p = 0; p < order; p++)
                {
                    data[i * basis[0].Length + p] = (int)Math.Round(basis[i][p] / delta);
                    basis[i][p] = delta * data[i * basis[0].Length + p];
                }
                for (int p = order; p < basis[0].Length - 1; p++)
                {
                    double prediction = 0;
                    for (int j = 1; j <= order; j++)
                    {
                        prediction += s[order - j, 0] * basis[i][p - j];

                    }
                    double residual = basis[i][p] - prediction;

                    int quantized = (int)(residual / delta);

                    data[i * basis[0].Length + p] = quantized;

                    if (cases.ContainsKey(quantized))
                    {
                        cases[quantized] = cases[quantized] + 1;
                    }
                    else cases[quantized] = 1;
                    valueCount++;

                    basis[i][p] = prediction + quantized * delta;
                }
            }

            double entropy = 0;
            foreach (int key in cases.Keys)
            {
                double probability = (double)cases[key] / valueCount;
                entropy += probability * Math.Log(probability, 2);
            }
            entropy = -entropy;

            double size = valueCount * entropy + (count * order * 64) + order * 64;

            if (target != null)
            {
                HuffmanCoder.encode(data, target, 1, true);
                return (target.Length - origLength);
            }
            else
                return (size);
        }

        /// <summary>
        /// eigen values received by eigenvalue decomposition during the PCA
        /// </summary>
        public double[] eigenValues = null;


        /// <summary>
        /// encodes the basis into a stream using non-uniform quantization and data-independent prediction
        /// </summary>
        /// <param name="target">target stream to write to</param>
        /// <param name="count">number of basis vectors to be used</param>
        /// <param name="delta">quantization constant (will be divided according to importance of basis vectors)</param>
        /// <returns></returns>
        public double EncodeBasisCobra(Stream target, int count, double delta)
        {
            double[] sums = new double[count];
            for (int i = 0; i < data.Length; i++)
            {
                double[] coefs = this.GetCoefs(data[i], count);
                for (int j = 0; j < count; j++)
                    sums[j] += Math.Abs(coefs[j]);
            }
            for (int i = 0; i < sums.Length; i++)
                sums[i] /= data.Length;

            double result = EncodeBasisCobra(count, 3, delta, target, sums, data.Length, 1, false);
            return (result);
        }

        void decodeCobra(Stream source, bool approximateQuantizations)
        {
            BinaryReader reader = new BinaryReader(source);
            int order = reader.ReadByte();
            int count = reader.ReadUInt16(); //number of basis vectors
            int basisLength = reader.ReadUInt16();

            double[] qw = new double[count];
            if ((count > 5) && (approximateQuantizations))
            {
                double cc = reader.ReadDouble();
                double b = reader.ReadDouble();
                double ac = reader.ReadDouble();
                for (int i = 0; i < count; i++)
                {
                    qw[i] = Math.Pow(i + 1, cc) * b + ac; //compute the exact qw                                
                }
            }
            else
            {
                for (int i = 0; i < count; i++)
                {
                    qw[i] = reader.ReadDouble();
                }
            }
            double meanqw = reader.ReadDouble();

            this.sampleLength = basisLength;
            this.means = new double[basisLength];
            this.basis = new double[count][];
            for (int i = 0; i < count; i++)
            {
                basis[i] = new double[basisLength];
            }

            int[] data = ArithCoder.decode(source);

            int pos = 0;
            for (int p = 0; p < order * 3; p++)
            {
                means[p] = data[pos] * meanqw;
                pos++;
            }

            double prediction = 0;

            // now write the rest of the means, this time using an appropriate predictor
            for (int p = order * 3; p < basis[0].Length; p++)
            {
                if (order == 2)
                {
                    prediction = means[p - 3] * 2 - means[p - 6];
                }
                else if (order == 3)
                    prediction = means[p - 3] * 3 - means[p - 6] * 3 + means[p - 9];
                else if (order == 4)
                    prediction = means[p - 3] * 4 - means[p - 6] * 6 + means[p - 9] * 4 - means[p - 12];
                else if (order == 1)
                    prediction = means[p - 3];

                int quantized = data[pos];
                pos++;

                means[p] = prediction + quantized * meanqw;
            }

            // now encode the basis vectors
            for (int i = 0; i < count; i++)
            {
                // first 2 or 3 simply quantized
                for (int p = 0; p < order * 3; p++)
                {
                    basis[i][p] = data[pos] * qw[i];
                    pos++;
                }

                // rest predicted
                for (int p = order * 3; p < basis[0].Length; p++)
                {

                    if (order == 2)
                    {
                        prediction = basis[i][p - 3] * 2 - basis[i][p - 6];
                    }
                    else if (order == 3)
                    {
                        prediction = basis[i][p - 3] * 3 - basis[i][p - 6] * 3 + basis[i][p - 9];
                    }
                    else if (order == 4)
                    {
                        prediction = basis[i][p - 3] * 4 - basis[i][p - 6] * 6 + basis[i][p - 9] * 4 - basis[i][p - 12];
                    }
                    else if (order == 1)
                    {
                        prediction = basis[i][p - 3];
                    }

                    int quantized = data[pos];
                    pos++;

                    basis[i][p] = prediction + quantized * qw[i];
                }
            }


        }

        private void normalize(double[] vector)
        {
            double sumsq = 0;
            for (int i = 0; i < vector.Length; i++)
            {
                sumsq += vector[i] * vector[i];
            }
            double length = Math.Sqrt(sumsq);
            double m = 1 / length;
            for (int i = 0; i < vector.Length; i++)
            {
                vector[i] *= m;
            }
        }

        private double dotProduct(double[] v1, double[] v2)
        {
            double result = 0;
            for (int i = 0; i < v1.Length; i++)
                result += v1[i] * v2[i];
            return (result);
        }

        private void subtractProjection(double[] v1, double[] v2)
        {
            double dp = dotProduct(v1, v2);
            for (int i = 0; i < v1.Length; i++)
            {
                v2[i] = v2[i] - dp * v1[i];
            }
        }

        /// <summary>
        /// Orthogonaslises basis using the Gramm-schmidt orthogonalisation process. May be used after compression of basis
        /// </summary>
        /// <param name="count">Number of basis vectors to process</param>
        public void RenormalizeBasis(int count)
        {
            normalize(basis[0]);
            for (int i = 1; i < count; i++)
            {
                for (int j = 0; j < i; j++)
                {
                    subtractProjection(basis[j], basis[i]);
                }
                normalize(basis[i]);
            }
        }

        /// <summary>
        /// encodes the basis into a stream using non-uniform quantization and data-independent prediction
        /// </summary>
        /// <param name="count">number of basis vectors to be encoded</param>
        /// <param name="order">order of prediction</param>
        /// <param name="delta">quantization constant</param>
        /// <param name="target">target data stream</param>
        /// <param name="sums">sums of absolute values of coefficients for each basis vector</param>
        /// <param name="pointCount">number of vertices</param>
        /// <param name="predictor">predictor to be used</param>
        /// <param name="approximateQuantizations">sets whether the quantizations should be approximated, or used exaclty</param>
        /// <returns>number of bytes needed for encoding the basis</returns>
        public double EncodeBasisCobra(int count, int order, double delta, Stream target, double[] sums, int pointCount, int predictor, bool approximateQuantizations)
        {
            return (EncodeBasisCobra(count, order, delta, target, sums, pointCount, predictor, delta, approximateQuantizations));
        }

        /// <summary>
        /// encodes the basis into a stream using non-uniform quantization and data-independent prediction
        /// </summary>
        /// <param name="count">number of basis vectors to be encoded</param>
        /// <param name="order">order of prediction</param>
        /// <param name="delta">quantization constant</param>
        /// <param name="target">target data stream</param>
        /// <param name="sums">sums of absolute values of coefficients for each basis vector</param>
        /// <param name="pointCount">number of vertices</param>
        /// <param name="predictor">predictor to be used</param>
        /// <param name="approximateQuantizations">sets whether the quantizations should be approximated, or used exaclty</param>
        /// <returns>number of bytes needed for encoding the basis></returns>
        public double EncodeBasisCobra(int count, int order, double delta, Stream target, double[] sums, int pointCount, int predictor)
        {
            return (EncodeBasisCobra(count, order, delta, target, sums, pointCount, predictor, delta, true));
        }

        /// <summary>
        /// encodes the basis into a stream using non-uniform quantization and data-independent prediction
        /// </summary>
        /// <param name="count">number of basis vectors to be encoded</param>
        /// <param name="order">order of prediction</param>
        /// <param name="delta">quantization constant</param>
        /// <param name="target">target data stream</param>
        /// <param name="sums">sums of absolute values of coefficients for each basis vector</param>
        /// <param name="pointCount">number of vertices</param>
        /// <param name="predictor">predictor to be used</param>        
        /// <param name="coefDelta">quantization constant used to quantize coefficients</param>
        /// <param name="approximateQuantizations">sets whether the quantizations should be approximated, or used exaclty</param>
        /// <returns>number of bytes needed for encoding the basis</returns>
        public double EncodeBasisCobra(int count, int order, double delta, Stream target, double[] sums, int pointCount, int predictor, double coefDelta, bool approximateQuantizations)
        {
            /*
             * predictors:
             * 
             * 0 - LPC
             * 1 - Cobra
             * 2 - quadratic
             * 3 - none
             * 
             *          
             * */

            // Expected Basis Element Length
            double ebel = Math.Sqrt(1.0 / basis[0].Length);

            // initialization
            double totSum = 0;
            for (int i = 0; i < sums.Length; i++)
            {
                totSum += sums[i];
            }

            double[] qw = new double[count];
            double norm = delta; //error per component
            for (int i = 0; i < count; i++)
            {

                qw[i] = norm / (sums[i] / totSum);
            }

            double b = 0, cc = 0, ac = 0;

            if (approximateQuantizations)
            {

                int frac = 2; //fraction of important basis components                

                if (count > 5)
                {

                    MathNet.Numerics.LinearAlgebra.Matrix mc = new MathNet.Numerics.LinearAlgebra.Matrix(count / frac, 2);
                    MathNet.Numerics.LinearAlgebra.Matrix rc = new MathNet.Numerics.LinearAlgebra.Matrix(count / frac, 1);
                    for (int i = 0; i < count / frac; i++)
                    {
                        mc[i, 0] = 1.0;
                        mc[i, 1] = Math.Log(i + 1); // we have to avoid zero
                        rc[i, 0] = Math.Log(qw[i]);
                    }
                    MathNet.Numerics.LinearAlgebra.Matrix sc = mc.Solve(rc);
                    b = Math.Exp(sc[0, 0]); //multiplier
                    cc = sc[1, 0]; //exponent
                    ac = qw[0] - b; //shift, so that the first qw is exact
                    for (int i = 0; i < count; i++)
                    {
                        if (predictor == 3)
                        {
                            qw[i] = norm / sums[3];
                        }
                        else
                        {
                            qw[i] = Math.Pow(i + 1, cc) * b + ac; //compute the exact qw                
                        }
                    }
                }
            }

            double maximumQWReduction = 10000;
            for (int i = 0; i < count; i++)
            {
                if (qw[i] > (maximumQWReduction * qw[0]))
                    qw[i] = maximumQWReduction * qw[0];
            }

            double meanqw = (ebel * coefDelta + qw[0] * sums[0] + coefDelta * qw[0]) / 10; //qw for the mean values            
            Dictionary<int, int> cases = new Dictionary<int, int>();
            int origLength = 0;

            MathNet.Numerics.LinearAlgebra.Matrix s = null;

            if (predictor == 0)
            {
                int compCount = (basis[0].Length - order * 3);
                int rowCount = (basis[0].Length - order * 3) * count;

                MathNet.Numerics.LinearAlgebra.Matrix m = new MathNet.Numerics.LinearAlgebra.Matrix(rowCount, order);
                MathNet.Numerics.LinearAlgebra.Matrix r = new MathNet.Numerics.LinearAlgebra.Matrix(rowCount, 1);

                for (int i = 0; i < count; i++)
                {
                    for (int comp = 0; comp < compCount; comp++)
                    {
                        int row = i * compCount + comp;
                        r[row, 0] = basis[i][comp + order * 3];
                        for (int c = 0; c < order; c++)
                            m[row, c] = basis[i][comp + order * 3 - (c + 1) * 3];
                    }
                }
                s = m.Solve(r);
                MathNet.Numerics.LinearAlgebra.Matrix res = m * s - r;
                double resum = res.Norm2();
            }

            BinaryWriter writer = null;

            if (target != null)
            {
                // write the initialization values
                origLength = (int)target.Length;
                writer = new BinaryWriter(target);
                writer.Write((byte)order);
                writer.Write((ushort)count);
                writer.Write((ushort)basis[0].Length);
                if ((count > 5) && (approximateQuantizations))
                {
                    writer.Write((double)cc);
                    writer.Write((double)b);
                    writer.Write((double)ac);
                }
                else
                {
                    for (int i = 0; i < qw.Length; i++)
                    {
                        writer.Write((double)qw[i]);
                    }
                }
                writer.Write((double)meanqw);
            }

            List<int> dataList = new List<int>(); //quantized values to be written                        

            // write the initial mean values needed for the predictor, simply quantized
            for (int p = 0; p < order * 3; p++)
            {
                dataList.Add((int)Math.Round(means[p] / meanqw));
                means[p] = dataList[dataList.Count - 1] * meanqw;
            }

            // now write the rest of the means, this time using an appropriate predictor
            for (int p = order * 3; p < basis[0].Length; p++)
            {
                double prediction = 0;

                if (predictor == 0)
                {
                    for (int j = 1; j <= order; j++)
                    {
                        prediction += s[j - 1, 0] * means[p - j * 3];
                    }
                }
                else if (predictor == 1)
                {
                    if (order == 2)
                    {
                        prediction = means[p - 3] * 2 - means[p - 6];
                    }
                    else if (order == 3)
                        prediction = means[p - 3] * 3 - means[p - 6] * 3 + means[p - 9];
                    else if (order == 4)
                        prediction = means[p - 3] * 4 - means[p - 6] * 6 + means[p - 9] * 4 - means[p - 12];
                    else if (order == 1)
                        prediction = means[p - 3];
                }


                double residual = means[p] - prediction;
                int quantized = (int)(residual / meanqw);

                dataList.Add(quantized);
                means[p] = prediction + quantized * meanqw;
            }

            // now encode the basis vectors
            for (int i = 0; i < count; i++)
            {
                // first 2 or 3 simply quantized
                for (int p = 0; p < order * 3; p++)
                {
                    dataList.Add((int)Math.Round(basis[i][p] / qw[i]));
                    basis[i][p] = dataList[dataList.Count - 1] * qw[i];
                }

                // rest predicted
                for (int p = order * 3; p < basis[0].Length; p++)
                {
                    double prediction = 0;
                    if (predictor == 0)
                    {
                        for (int j = 1; j <= order; j++)
                        {
                            prediction += s[j - 1, 0] * basis[i][p - j * 3];
                        }
                    }
                    else if (predictor == 1)
                    {
                        if (order == 2)
                        {
                            prediction = basis[i][p - 3] * 2 - basis[i][p - 6];
                        }
                        else if (order == 3)
                        {
                            prediction = basis[i][p - 3] * 3 - basis[i][p - 6] * 3 + basis[i][p - 9];
                        }
                        else if (order == 4)
                        {
                            prediction = basis[i][p - 3] * 4 - basis[i][p - 6] * 6 + basis[i][p - 9] * 4 - basis[i][p - 12];
                        }
                        else if (order == 1)
                        {
                            prediction = basis[i][p - 3];
                        }
                    }
                    else if (predictor == 2)
                    {
                        MathNet.Numerics.LinearAlgebra.Matrix m = new MathNet.Numerics.LinearAlgebra.Matrix(order, 3);
                        MathNet.Numerics.LinearAlgebra.Matrix r = new MathNet.Numerics.LinearAlgebra.Matrix(order, 1);
                        for (int j = 0; j < order; j++)
                        {
                            m[j, 0] = j * j;
                            m[j, 1] = j;
                            m[j, 2] = 1;
                            r[j, 0] = basis[i][p - (order - j) * 3];
                        }
                        MathNet.Numerics.LinearAlgebra.Matrix ls = m.Solve(r);
                        prediction = order * order * ls[0, 0] + order * ls[1, 0] + ls[2, 0];
                    }

                    double residual = basis[i][p] - prediction;
                    int quantized = (int)(residual / qw[i]);

                    dataList.Add(quantized);
                    basis[i][p] = prediction + quantized * qw[i];
                }
            }

            // compute the case count
            foreach (int i in dataList)
            {
                if (cases.ContainsKey(i))
                {
                    cases[i] = cases[i] + 1;
                }
                else
                {
                    cases[i] = 1;
                }
            }

            // compute the entropy
            int valueCount = dataList.Count;
            double entropy = 0;
            foreach (int key in cases.Keys)
            {
                double probability = (double)cases[key] / valueCount;
                entropy += probability * Math.Log(probability, 2);
            }
            entropy = -entropy;

            double size = valueCount * entropy + (count * order * 64) + order * 64;

            // finally, write the data
            if (target != null)
            {
                ArithCoder.encode(dataList.ToArray(), target);
                return (target.Length - origLength);
            }
            else
                return (size);
        }

        double[] TestBasisPredictors(int count, int order, double delta, Stream target, double[] sums, int pointCount, int predictor)
        {
            double[] qw = new double[count];
            double norm = delta / (count + 1);
            for (int i = 0; i < count; i++)
            {
                qw[i] = norm / sums[i];
            }

            double[] lmeans = (double[])means.Clone();
            double[][] lbasis = new double[basis.Length][];
            for (int i = 0; i < lbasis.Length; i++)
            {
                lbasis[i] = (double[])basis[i].Clone();
            }

            int frac = 2;

            MathNet.Numerics.LinearAlgebra.Matrix mc = new MathNet.Numerics.LinearAlgebra.Matrix(count / frac, 2);
            MathNet.Numerics.LinearAlgebra.Matrix rc = new MathNet.Numerics.LinearAlgebra.Matrix(count / frac, 1);
            for (int i = 0; i < count / frac; i++)
            {
                mc[i, 0] = 1.0;
                mc[i, 1] = Math.Log(i + 1);
                rc[i, 0] = Math.Log(qw[i]);
            }
            MathNet.Numerics.LinearAlgebra.Matrix sc = mc.Solve(rc);
            double b = Math.Exp(sc[0, 0]);
            double cc = sc[1, 0];
            double ac = qw[0] - b;
            for (int i = 0; i < count; i++)
            {
                qw[i] = Math.Pow(i + 1, cc) * b + ac;
            }

            double meanqw = delta / (count + 1);

            Dictionary<int, int> cases = new Dictionary<int, int>();
            MathNet.Numerics.LinearAlgebra.Matrix s = null;

            if (predictor == 0)
            {
                int compCount = (lbasis[0].Length - order * 3);
                int rowCount = (lbasis[0].Length - order * 3) * count;

                MathNet.Numerics.LinearAlgebra.Matrix m = new MathNet.Numerics.LinearAlgebra.Matrix(rowCount, order);
                MathNet.Numerics.LinearAlgebra.Matrix r = new MathNet.Numerics.LinearAlgebra.Matrix(rowCount, 1);

                for (int i = 0; i < count; i++)
                {
                    for (int comp = 0; comp < compCount; comp++)
                    {
                        int row = i * compCount + comp;
                        r[row, 0] = lbasis[i][comp + order * 3];
                        for (int c = 0; c < order; c++)
                            m[row, c] = lbasis[i][comp + order * 3 - (c + 1) * 3];
                    }
                }
                s = m.Solve(r);
                MathNet.Numerics.LinearAlgebra.Matrix res = m * s - r;
                double resum = res.Norm2();
            }
            int origLength = 0;

            BinaryWriter writer = null;

            if (target != null)
            {
                origLength = (int)target.Length;
                writer = new BinaryWriter(target);
                writer.Write((byte)order);
                writer.Write((ushort)count);
                writer.Write((ushort)lbasis[0].Length);
                writer.Write((float)cc);
                writer.Write((float)b);
                writer.Write((float)ac);
            }

            double residualSum = 0;

            List<int> dataList = new List<int>();
            for (int p = 0; p < order * 3; p++)
            {
                int quantized = (int)Math.Round(lmeans[p] / meanqw);
                lmeans[p] = quantized * meanqw;
            }

            for (int p = order * 3; p < lbasis[0].Length; p++)
            {
                double prediction = 0;

                if (predictor == 0)
                {
                    for (int j = 1; j <= order; j++)
                    {
                        prediction += s[j - 1, 0] * lmeans[p - j * 3];
                    }
                }
                else if (predictor == 1)
                {
                    if (order == 2)
                    {
                        prediction = lmeans[p - 3] * 2 - lmeans[p - 6];
                    }
                    else if (order == 3)
                        prediction = lmeans[p - 3] * 3 - lmeans[p - 6] * 3 + lmeans[p - 9];
                    else if (order == 4)
                        prediction = lmeans[p - 3] * 4 - lmeans[p - 6] * 6 + lmeans[p - 9] * 4 - lmeans[p - 12];
                    else if (order == 1)
                        prediction = lmeans[p - 3];
                }


                double residual = lmeans[p] - prediction;
                int quantized = (int)(residual / meanqw);
                lmeans[p] = prediction + quantized * meanqw;
            }

            for (int i = 0; i < count; i++)
            {
                for (int p = 0; p < order * 3; p++)
                {
                    dataList.Add((int)Math.Round(lbasis[i][p] / qw[i]));
                    lbasis[i][p] = dataList[dataList.Count - 1] * qw[i];
                }
                for (int p = order * 3; p < lbasis[0].Length; p++)
                {
                    double prediction = 0;
                    if (predictor == 0)
                    {
                        for (int j = 1; j <= order; j++)
                        {
                            prediction += s[j - 1, 0] * lbasis[i][p - j * 3];
                        }
                    }
                    else if (predictor == 1)
                    {
                        if (order == 2)
                        {
                            prediction = lbasis[i][p - 3] * 2 - lbasis[i][p - 6];
                        }
                        else if (order == 3)
                        {
                            prediction = lbasis[i][p - 3] * 3 - lbasis[i][p - 6] * 3 + lbasis[i][p - 9];
                        }
                        else if (order == 4)
                        {
                            prediction = lbasis[i][p - 3] * 4 - lbasis[i][p - 6] * 6 + lbasis[i][p - 9] * 4 - lbasis[i][p - 12];
                        }
                        else if (order == 1)
                        {
                            prediction = lbasis[i][p - 3];
                        }
                    }
                    else if (predictor == 2)
                    {
                        MathNet.Numerics.LinearAlgebra.Matrix m = new MathNet.Numerics.LinearAlgebra.Matrix(order, 3);
                        MathNet.Numerics.LinearAlgebra.Matrix r = new MathNet.Numerics.LinearAlgebra.Matrix(order, 1);
                        for (int j = 0; j < order; j++)
                        {
                            m[j, 0] = j * j;
                            m[j, 1] = j;
                            m[j, 2] = 1;
                            r[j, 0] = lbasis[i][p - (order - j) * 3];
                        }
                        MathNet.Numerics.LinearAlgebra.Matrix ls = m.Solve(r);
                        prediction = order * order * ls[0, 0] + order * ls[1, 0] + ls[2, 0];
                    }


                    double residual = lbasis[i][p] - prediction;
                    residualSum += residual * residual;
                    int quantized = (int)(residual / qw[i]);
                    dataList.Add(quantized);
                    lbasis[i][p] = prediction + quantized * qw[i];
                }
            }

            // compute the case count
            foreach (int i in dataList)
            {
                if (cases.ContainsKey(i))
                {
                    cases[i] = cases[i] + 1;
                }
                else
                {
                    cases[i] = 1;
                }
            }

            double entropy = 0;
            int valueCount = dataList.Count;
            foreach (int key in cases.Keys)
            {
                double probability = (double)cases[key] / valueCount;
                entropy += probability * Math.Log(probability, 2);
            }
            entropy = -entropy;

            double size = valueCount * entropy;

            return (new double[] { entropy, residualSum });
        }

        private double[] computeRow(object iO)
        {
            int i = (int)iO;
            double[] result = new double[sampleLength];
            for (int j = i; j < sampleLength; j++)
            {
                double element = 0;
                for (int k = 0; k < sampleCount; k++)
                {

                    element += dB[i][k] * dB[j][k];
                }
                result[j] = element;
            }
            return (result);
        }

        bool useDoubles = true;

        double[][] dB = null;

        /// <summary>
        /// performs the actual PCA step. Before calling you should assign data using the Data property.
        /// </summary>
        public void ComputeBasis()
        {
            if (subtractMean)
            {
                // compute means
                means = new double[sampleLength];
                for (int i = 0; i < sampleLength; i++)
                {

                    for (int j = 0; j < sampleCount; j++)
                    {
                        means[i] += data[j][i];
                    }
                    means[i] /= sampleCount;
                }
            }

            // construct matrix
            MathNet.Numerics.LinearAlgebra.Matrix B = null;
            if (useDoubles)
            {
                dB = new double[data[0].Length][];
                for (int i = 0; i < data[0].Length; i++)
                {
                    dB[i] = new double[data.Length];
                }
                if (subtractMean)
                {
                    for (int i = 0; i < sampleCount; i++)
                        for (int j = 0; j < sampleLength; j++)
                        {
                            dB[j][i] = data[i][j] - means[j];
                        }
                }
                else
                {
                    for (int i = 0; i < sampleCount; i++)
                        for (int j = 0; j < sampleLength; j++)
                        {
                            dB[j][i] = data[i][j];
                        }
                }
                //B = new Matrix(dB);
            }
            else
            {
                B = new MathNet.Numerics.LinearAlgebra.Matrix(sampleLength, sampleCount);
                if (subtractMean)
                {
                    for (int i = 0; i < sampleCount; i++)
                        for (int j = 0; j < sampleLength; j++)
                        {
                            B[j, i] = data[i][j] - means[j];
                        }
                }
                else
                {
                    for (int i = 0; i < sampleCount; i++)
                        for (int j = 0; j < sampleLength; j++)
                        {
                            B[j, i] = data[i][j];
                        }
                }
            }


            if (this.method == PCAMethod.direct)
            {
                // compute the covariance matrix                
                MathNet.Numerics.LinearAlgebra.Matrix covarianceMatrix;
                if (useDoubles)
                {
                    object[] inputArray = new object[sampleLength];
                    for (int i = 0; i < sampleLength; i++)
                    {
                        inputArray[i] = (object)i;
                    }
                    ThreadedExecution<object, double[]> task = new ThreadedExecution<object, double[]>(inputArray, computeRow);
                    double[][] dCovarianceMatrix = task.Execute();
                    for (int i = 0; i < sampleLength; i++)
                    {
                        for (int j = i; j < sampleLength; j++)
                        {
                            dCovarianceMatrix[j][i] = dCovarianceMatrix[i][j];
                        }
                    }
                    covarianceMatrix = new Matrix(dCovarianceMatrix);
                }
                else
                {
                    covarianceMatrix = new MathNet.Numerics.LinearAlgebra.Matrix(sampleLength, sampleLength);
                    for (int i = 0; i < sampleLength; i++)
                    {
                        for (int j = i; j < sampleLength; j++)
                        {
                            double element = 0;
                            for (int k = 0; k < sampleCount; k++)
                            {

                                element += B[i, k] * B[j, k];
                            }
                            covarianceMatrix[i, j] = covarianceMatrix[j, i] = element;
                        }
                    }
                }

                // compute the eigenvalue decomposition
                EigenvalueDecomposition evd = new EigenvalueDecomposition(covarianceMatrix, true);

                for (int i = (evd.RealEigenvalues.Length - 1); i > 0; i--)
                {
                    //Console.WriteLine(evd.RealEigenvalues[i]);
                }
                this.eigenValues = evd.RealEigenvalues;

                this.basis = new double[sampleLength][];
                for (int i = 0; i < sampleLength; i++)
                {
                    this.basis[i] = new double[sampleLength];
                    for (int j = 0; j < sampleLength; j++)
                    {
                        this.basis[i][j] = evd.EigenVectors[j, sampleLength - i - 1];
                    }
                }
            }
            else
            {
                // this one hangs on some matrices
                //SingularValueDecomposition svd = B.SVD();

                // alternate SVD algorithm...
                double[,] svdA = B.CopyToArray();
                int svdM = B.RowCount;
                int svdN = B.ColumnCount;
                int svdUNeeded = 1; // just the left singular vectors (U cropped to largest square matrix)
                int svdVNeeded = 0; // we don't need the right singular vectors
                int svdExtMemory = 2; // use more memory for the fastest algorithm
                double[] svdW = null; // singular value vector (output parameter)
                double[,] svdU = null; // left singular vector matrix (output parameter)
                double[,] svdVt = null; // right singular vector matrix (output parameter)
                NewSVD.RMatrixSVD(svdA, svdM, svdN, svdUNeeded, svdVNeeded, svdExtMemory, ref svdW, ref svdU, ref svdVt);

                this.basis = new double[sampleCount][];
                for (int i = 0; i < sampleCount; i++)
                {
                    this.basis[i] = new double[sampleLength];
                    for (int j = 0; j < sampleLength; j++)
                    {
                        //this.basis[i][j] = svd.LeftSingularVectors[j, i];

                        this.basis[i][j] = svdU[j, i];
                    }
                }
            }
        }

        /// <summary>
        /// gets pca coefficients for the given data vector
        /// </summary>
        /// <param name="vector">input data vector</param>
        /// <param name="numCoefs">number of coefficients to be retrieved</param>
        /// <returns>vector of PCA coefficients</returns>
        public double[] GetCoefs(double[] vector, int numCoefs)
        {
            double[] result = new double[numCoefs];
            int l = vector.Length;
            if (subtractMean)
            {
                for (int i = 0; i < numCoefs; i++)
                    for (int j = 0; j < l; j++)
                        result[i] += basis[i][j] * (vector[j] - means[j]);
            }
            else
            {
                for (int i = 0; i < numCoefs; i++)
                    for (int j = 0; j < l; j++)
                        result[i] += basis[i][j] * vector[j];
            }
            return result;
        }

        /// <summary>
        /// gets all pca coefficients for the given data vector
        /// </summary>
        /// <param name="vector">input data vector</param>
        /// <returns>vector of PCA coefficients</returns>
        public double[] GetCoefs(double[] vector)
        {
            return (this.GetCoefs(vector, vector.Length));
        }

        /// <summary>
        /// Retrieves PCA coefficients, using projection of basis vectors into the original space. Should provide exactly identical results as the GetCoefs method.
        /// </summary>
        /// <param name="vector">Data vector to be processed.</param>
        /// <param name="numCoefs">Number of coefficients to retrieve.</param>
        /// <returns>vector of PCA coefficients</returns>
        public double[] GetCoefsExact(double[] vector, int numCoefs)
        {
            MathNet.Numerics.LinearAlgebra.Matrix m = new MathNet.Numerics.LinearAlgebra.Matrix(vector.Length, numCoefs);
            MathNet.Numerics.LinearAlgebra.Matrix r = new MathNet.Numerics.LinearAlgebra.Matrix(vector.Length, 1);
            for (int i = 0; i < vector.Length; i++)
            {
                for (int j = 0; j < numCoefs; j++)
                {
                    m[i, j] = basis[j][i];
                }
                if (subtractMean)
                    r[i, 0] = vector[i] - means[i];
                else
                    r[i, 0] = vector[i];
            }
            MathNet.Numerics.LinearAlgebra.Matrix solution = m.Solve(r);
            double[] result = new double[numCoefs];
            for (int i = 0; i < numCoefs; i++)
            {
                result[i] = solution[i, 0];
            }
            return (result);
        }

        /// <summary>
        /// retrieves one PCA coefficient for the given data vector
        /// </summary>
        /// <param name="vector">Data vector to be processed.</param>
        /// <param name="coefIndex">Index of the coefficient to be used.</param>
        /// <returns>single PCA coefficient</returns>
        public double GetCoef(double[] vector, int coefIndex)
        {
            double result = 0;
            if (subtractMean)
            {
                for (int i = 0; i < vector.Length; i++)
                {
                    result += basis[coefIndex][i] * (vector[i] - means[i]);
                }
            }
            return (result);
        }

        /// <summary>
        /// transforms a vector of PCA coefficients onto the original space
        /// </summary>
        /// <param name="encoded">vector of PCA coefficients</param>
        /// <returns>vector in original space</returns>
        public double[] Decode(double[] encoded)
        {
            double[] decoded = new double[sampleLength];
            for (int i = 0; i < encoded.Length; i++)
            {
                for (int j = 0; j < sampleLength; j++)
                {
                    decoded[j] += basis[i][j] * encoded[i];
                }
            }

            if (subtractMean)
            {
                for (int i = 0; i < sampleLength; i++)
                    decoded[i] += means[i];
            }

            return (decoded);
        }

        /// <summary>
        /// deterimes the error caused by PCA representation of a given vector, using a given number of PCA coefficients
        /// </summary>
        /// <param name="vector">Vector to be processed</param>
        /// <param name="numCoefs">Number of PCA coefficients to be used</param>
        /// <returns>Error - root mean squared</returns>
        public double encodingError(double[] vector, int numCoefs)
        {
            double[] decoded = Decode(GetCoefs(vector, numCoefs));
            double distance = 0;
            double difference = 0;
            for (int i = 0; i < vector.Length; i++)
            {
                difference = vector[i] - decoded[i];
                distance += difference * difference;
            }
            distance = Math.Sqrt(distance);
            return (distance);
        }
    }
}