﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Zcu.Mve.Compression
{
    /// <summary>
    /// Uniform quantization class
    /// </summary>
    public class UniformQuantizer : Zcu.Mve.Compression.IQuantizer
    {
        /// <summary>
        /// size of the quantization step
        /// </summary>
        public double delta = 1;

        double invDelta;

        /// <summary>
        /// Constructs a new uniform quantizer using the given size of quantization step.
        /// </summary>
        /// <param name="delta">size of quantization step</param>
        public UniformQuantizer(Double delta)
        {
            this.delta = delta;
            this.invDelta = 1 / delta;
        }

        /// <summary>
        /// saves the quantization step into the stream
        /// </summary>
        /// <param name="s">stream to write to</param>
        public void Serialize(Stream s)
        {
            BinaryWriter bw = new BinaryWriter(s);
            bw.Write(delta);
        }

        /// <summary>
        /// reads the quantization step from a file
        /// </summary>
        /// <param name="s">stream to deserialize from</param>
        public void Deserialize(Stream s)
        {
            BinaryReader br = new BinaryReader(s);
            delta = br.ReadDouble();
        }

        /// <summary>
        /// Quantizes a scalar value into an integer
        /// </summary>
        /// <param name="v">Value to quantize</param>
        /// <returns>quantized value</returns>
        public int Quantize(double v)
        {
            int result = (int)Math.Round(v*invDelta);
            return (result);
        }

        /// <summary>
        /// Restores a value from integer
        /// </summary>
        /// <param name="q">Quantized integer input</param>
        /// <returns>Restored value</returns>
        public double Restore(int q)
        {
            double result = q * delta;
            return (result);
        }

        /// <summary>
        /// Returns a quantized and restored value.
        /// </summary>
        /// <param name="v">value to quantoze and resotore</param>
        /// <returns>result of subsequent quantization and restoration</returns>
        public double QR(double v)
        {
            return(Restore(Quantize(v)));
        }
    }

    public class LloydMax : Zcu.Mve.Compression.IQuantizer
    {

        #region IQuantizer Members

        private double[] ys, xs;

        private int zeropos;

        public void Train(double[] samples, double delta)
        {
            Array.Sort(samples);
            double min = samples[0];
            double max = samples[samples.Length-1];
            double span = max - min;
            ys = new double[(int)Math.Ceiling(span / delta)];
            xs = new double[ys.Length - 1];
            for (int i = 0; i < xs.Length; i++)
            {
                xs[i] = min + (i + 1) * delta;
                ys[i + 1] = xs[i] + delta / 2;
            }
            double biggestchange = delta;
            while (biggestchange > delta / 5000000)
            {
                biggestchange = 0;
                int index = 0;

                //update ys
                for (int i = 0; i < ys.Length; i++)
                {
                    double sum = 0;
                    int count = 0;
                    double ceil = i<xs.Length?xs[i]:double.MaxValue;
                    while ((index < samples.Length) && (samples[index] < ceil))
                    {
                        sum += samples[index];
                        count++;
                        index++;
                    }
                    if (count != 0)                        
                    ys[i] = sum / count;
                }

                //update xs
                for (int i = 0; i < xs.Length; i++)
                {
                    double nxs = (ys[i] + ys[i + 1]) / 2;
                    double change = Math.Abs(xs[i] - nxs);
                    if (change > biggestchange)
                    {
                        biggestchange = change;
                    }
                    xs[i] = nxs;
                }
                //Console.WriteLine(biggestchange);
            }
            zeropos = 0;
            while (xs[zeropos] < 0)
                zeropos++;
        }

        public void Deserialize(Stream s)
        {
            throw new NotImplementedException();
        }

        public int Quantize(double v)
        {
            for (int i = 0; i < xs.Length; i++)
            {
                if (v < xs[i])
                {
                    return (i-zeropos);
                }
            }
            return (xs.Length-zeropos);
        }

        public double Restore(int q)
        {
            return (ys[q+zeropos]);
        }

        public void Serialize(Stream s)
        {
            throw new NotImplementedException();
        }

        #endregion
    }

    public class CollectingUniformQuantizer : Zcu.Mve.Compression.IQuantizer
    {
        private double delta = 1;
        private List<double> values = new List<double>();

        public List<double> Values
        {
            get { return values; }
            set { values = value; }
        }

        public CollectingUniformQuantizer(Double delta)
        {
            this.delta = delta/100;
        }

        public void Serialize(Stream s)
        {
            BinaryWriter bw = new BinaryWriter(s);
            bw.Write(delta);
        }

        public void Deserialize(Stream s)
        {
            BinaryReader br = new BinaryReader(s);
            delta = br.ReadDouble();
        }

        public int Quantize(double v)
        {
            int result = (int)Math.Round(v / delta);
            values.Add(v);
            return (result);
        }

        public double Restore(int q)
        {
            double result = q * delta;
            return (result);
        }
    }

    public class SqrQuantizer : IQuantizer
    {
        private double delta = 1;

        public SqrQuantizer(double d)
        {
            delta = d;
        }

        #region IQuantizer Members

        public void Deserialize(Stream s)
        {
            throw new NotImplementedException();
        }

        public int Quantize(double v)
        {
            double abs = Math.Sqrt(Math.Abs(v));
            int result = (int)Math.Round(abs / delta);
            if (v < 0) result = -result;
            return (result);
        }

        public double Restore(int q)
        {
            double result = q * delta;
            result *= result;
            if (q < 0)
                result = -result;
            return (result);
        }

        public void Serialize(Stream s)
        {
            throw new NotImplementedException();
        }

        #endregion
    }
}
