﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

using Zcu.Mve.Core;

namespace Zcu.Mve.Compression
{
    public class ThreadedExecution<TInput, TOutput>
        where TInput: class
        /*where TOutput: class*/
    {
        public delegate TOutput ExecBody(TInput input);

        
        private TInput[] _in;
        private TOutput[] _out;
        private ExecBody _body;
        private int _numDone, _numAssigned;
        private Thread[] _threads;
        private ProgressInfoDelegate _progressInfo;
        private bool _aborted;


        public ThreadedExecution(TInput[] input, ExecBody execBody, int threadCount, ProgressInfoDelegate progressInfo)
        {
            _in = input;
            _out = new TOutput[input.Length];
            _body = execBody;

            _progressInfo = progressInfo;
            _numDone = 0;
            _numAssigned = 0;
            _aborted = false;

            _threads = new Thread[threadCount];
            for (int i = 0; i < _threads.Length; i++)
            {
                _threads[i] = new Thread(ThreadBody);
            }
        }

        public ThreadedExecution(TInput[] input, ExecBody execBody, ProgressInfoDelegate progressInfo)
            : this(input, execBody, Environment.ProcessorCount, progressInfo)
        { }

        private TInput NextTask(TOutput prevResult, ref int taskId)
        {
            lock (_out)
            {
                if (_aborted) return default(TInput);

                if (taskId >= 0)
                {
                    _out[taskId] = prevResult;
                    _numDone++;
                    _progressInfo((double)_numDone / _out.Length);
                }

                if (_numAssigned < _in.Length)
                {
                    taskId = _numAssigned++;
                    return _in[taskId];
                }

                return default(TInput);
            }
        }

        private void ThreadBody()
        {
            int taskId = -1;
            TInput task;
            TOutput result = default(TOutput);

            while (true)
            {
                task = NextTask(result, ref taskId);
                if (task == null) break;
                result = _body(task);
            }
        }

        public TOutput[] Execute()
        {
            for (int i = 0; i < _threads.Length; i++)
            {
                _threads[i].Start();
            }

            for (int i = 0; i < _threads.Length; i++)
            {
                try
                {
                    _threads[i].Join();
                }
                catch (Exception)
                { }
            }

            if (_aborted) return null;

            return _out;
        }

        public void Abort()
        {
            lock (_out)
            {
                _aborted = true;
            }

            for (int i = 0; i < _threads.Length; i++)
            {
                try
                {
                    _threads[i].Join();
                }
                catch (Exception)
                { }
            }
        }
    }
}
